/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package algo

import scala.Array.canBuildFrom
import scala.collection.JavaConversions._
import scala.collection.mutable.ArrayBuffer
import scala.collection.mutable.BitSet
import scala.collection.mutable.HashMap
import scala.collection.mutable.HashSet
import scala.util.Random
import measures.GraphMeasures
import utils.Stat
import utils.Timer
import xps.PlotDegreeDistrib
import guidyn.components.ProgressBar
import guidyn.components.graph.VertexEvents
import algo.structures.MicroCommunityUpdate
import algo.structures.Graph
import scala.collection.mutable.Stack

case class MicroProba(probaMicro : Double,  
        var addVertices : Double, var removeVertices: Double, var updateAttributes: Double, 
        var removeInternalEdges: Double, var addInternalEdges: Double,
        var removeExternalEdges: Double, var addExternalEdges: Double, 
        var probaMerge : Double, var probaSplit : Double, var probaChange : Double)

class Generator1(val k: Int, val nbVertices: Int, val dev: Array[Double], val theta: Double,
			     val nbRep: Int, val maxInternalEdges: Int, val maxExternalEdges: Int, val minimumTotalEdges: Int, val seed: Option[Long] = None,
			     val microProba: Option[MicroProba] = None, useAverageDegree : Boolean = false, val probaCloseTriangle : Double, 
			     skipGraphGeneration : Boolean = false) {
	
    // Random seed
    val selectedSeed = seed.getOrElse(System.nanoTime())
    Random.setSeed(selectedSeed)

    // The graph
    var graph = new Graph()
    
    // Generate power law distributions for a given mean and maximum degree
    val powerLawExponent = 2D
    val withinPowerLawDistribution = new PowerLawDistribution(maxInternalEdges.toDouble / 2)
    val betweenPowerLawDistribution = new PowerLawDistribution(maxExternalEdges.toDouble / 2)
    
    val startTime = System.currentTimeMillis()
	if(!skipGraphGeneration) {
        // Build the set of candidate vertices
	    val candVertices = new BitSet(nbVertices)
	    for(i <- 0 until nbVertices) candVertices.add(i)
	
	    // Generate attributes and vertices
	    for (_ <- 0 until nbVertices) { graph.generateVertex(dev) }
	    
	    // Initialize the communities
	    ProgressBar.startTask("Communities Initialisation")
	    Timer.startTimer("Communities Init")
	    for ((_, community) <- KMedoids.cluster((0 until (nbRep * k)).toArray, k, graph.attr)) {
	        val cIdx = graph.createCommunity(community)
	        for(v1 <- community) {
	            candVertices.remove(v1)
	            for (v2 <- Stat.sample(community, Stat.randInt(1, math.min(community.size, maxInternalEdges)), v1)) { graph.addEdge(v1, v2, cIdx, cIdx) }
	        }
	        
	    }
	    graph.updateRepresentatives(nbRep)
	    Timer.stopTimer("Communities Init")
	    
	    
	    // Main loop
	    ProgressBar.startTask("Adding vertices", Option(candVertices.size))
	    Timer.startTimer("Main")
	    while (!candVertices.isEmpty) {
	        for (v <- candVertices.take(Stat.randInt(1, candVertices.size))) {
	            
			    
	            val cIdx = selectVertexCommunity(v)
	            addEdges(v, cIdx)
	            graph.setVertexCommunity(v, cIdx)
	            candVertices.remove(v)
	            ProgressBar.taskAdvance()
	        }
	        graph.updateRepresentatives(nbRep)
	    }
	    
	    
	    addRemainingEdges()
	    
	    Timer.stopTimer("Main")
    }
    val runTimeSec = (System.currentTimeMillis() - startTime).toDouble / 1000
    /**
     * Given  a vertex v, returns the index of the community to which it will be associated
     */
    def selectVertexCommunity(v : Int) : Int = {
        if (Random.nextDouble < theta) { Stat.pickAtRandom(graph.partition.keys.toArray) }
        else {
            // Select community being less distant from v
            var (minDist, bestCom) = (Double.MaxValue, -1)
            for((cIdx, _) <- graph.representatives) {
                val d = graph.distanceWithCommunity(v, cIdx)
                if(d < minDist) { bestCom = cIdx; minDist = d }
            }
            bestCom
        }
    }

    /**
     * Add a random number of internal and external edges connecting a vertex v
     * The added edges might be already present in the graph
     */
    def addEdges(v: Int, cIdx: Int) {
        Timer.startTimer("Add Edges")
        val nbWithinEdges = 
            if(useAverageDegree) { withinPowerLawDistribution.pickValue(graph.partition(cIdx).size - 1) } 
        	else { Stat.randIntPowerLaw(math.min(maxInternalEdges, graph.partition(cIdx).size), powerLawExponent) }
        var sumDegreeAlreadyConnected = 0
        val candNeighbors = graph.partition(cIdx)
        val withinNeighborsAdded = Stack[Int](v)
        candNeighbors.remove(v)
        while (graph.degree(v) != nbWithinEdges) {
            Timer.startTimer("Select Within edges")
        	val v2 = Stat.weightedProbabilityWithSum(candNeighbors, (a : Int) => graph.degree(a), graph.communitiesSumDegree(cIdx) - graph.degree(v) - sumDegreeAlreadyConnected)
        	Timer.stopTimer("Select Within edges")
            if(graph.addEdge(v, v2, cIdx, cIdx)) {
                sumDegreeAlreadyConnected += graph.degree(v2)
                withinNeighborsAdded.push(v2)
                candNeighbors.remove(v2)
            }
        }
        for(v2 <- withinNeighborsAdded) { candNeighbors.add(v2) }
        
        Timer.startTimer("Add between")
        if(graph.nbCommunities > 1 && maxExternalEdges > 0) {
            val maxBetweenEdges = math.min(graph.representativesNotInCommunity(cIdx).size, maxExternalEdges)
	        val nbBetweenEdges = 
	            if(useAverageDegree) { betweenPowerLawDistribution.pickValue(maxBetweenEdges)  } 
				else { Stat.randIntPowerLaw(maxBetweenEdges + 1, powerLawExponent) - 1 }
            
            addBetweenEdges(v, cIdx, nbBetweenEdges)
            
            /*
            val verticesAdded = new Stack[Int]()
            val candNeighbors = graph.representativesNotInCommunity(cIdx)
	        while (graph.degree(v) != nbWithinEdges + nbBetweenEdges) {
	            Timer.startTimer("Select Between edges")
	            val v2 = Stat.inverseWeightedProbability[Int](candNeighbors, n => graph.distance(v, n))
	            Timer.stopTimer("Select Between edges")
	            graph.addEdge(v, v2, cIdx, graph.vertexCommunities(v2))
	            candNeighbors.remove(v2)
	            verticesAdded.push(v2)
	        }
            for(x <- verticesAdded) { candNeighbors.add(x) }
            */
            
        }
        Timer.stopTimer("Add between")
        Timer.stopTimer("Add Edges")
    }
    
    def addBetweenEdges(v : Int, cIdx : Int, nbBetweenEdges : Int) = {
        val candNeighbors = graph.representativesNotInCommunity(cIdx)
        var nbCandNeighbors = candNeighbors.size
        
        var sumNorm = 0D
        var i = 0
        var cacheValues = new Array[Double](nbCandNeighbors)
        for(v2 <- candNeighbors) {
            val d = graph.distance(v, v2)
            cacheValues(i) = if(d == 0) { 100000 } else { 1D / d }
            sumNorm += cacheValues(i)
            i += 1
        }

        var nbEdgesAdded = 0
        while(nbEdgesAdded != nbBetweenEdges) {
            val idx = Random.nextDouble * sumNorm
		    var acc = 0D
		    var j = 0
		    val v2 = (candNeighbors.find { i =>
		        acc += cacheValues(j)
		        j += 1
		        acc >= idx
		    }).get
		    
		    if(!graph.edgeExists(v, v2)) {
		        nbEdgesAdded += 1
		        sumNorm -= cacheValues(j - 1)
		        cacheValues(j - 1) = 0
		        graph.addEdge(v, v2, cIdx, graph.vertexCommunities(v2))
		    }
        }
    }
    
    /**
     * Add edges closing triangle formed by vertices in the same community
     * Edges are added until there is at least minimumTotalEdges edges in the graph
     * The first vertex forming the triangle is selected uniformly at random among all vertices
     */
    def addRemainingEdges() : Unit = {
        Timer.startTimer("MTE")
        // Update the maximum reachable number of edges
        val edgesInCliques = graph.partition.values.map(c => (c.size * (c.size - 1)) / 2).sum
        val updatedMte = math.min(edgesInCliques, minimumTotalEdges)
        ProgressBar.startTask("MTE", Option(math.max(0, updatedMte - graph.nbEdges)))
        
        // Add the edges
        while (graph.nbEdges < updatedMte) {
            if(microProba.isDefined) {
                val v1 = Stat.weightedProbability(graph.vertices, (x : Int) => graph.degree(x).toDouble)
                val cIdx = graph.vertexCommunities(v1)
                val withinNeighborhood = graph(v1).filter(n => graph.sameCommunity(v1, n))
                if(!withinNeighborhood.isEmpty) {
                    val v = Stat.pickAtRandom(withinNeighborhood)
                    val candNeighbors = graph(v).filter(n => !graph.edgeExists(n, v1) && graph.sameCommunity(v1, n) && n != v1)
                    if(!candNeighbors.isEmpty) {
                        val v2 = Stat.pickAtRandom(candNeighbors)
                        graph.addEdge(v1, v2, cIdx, cIdx)
                        ProgressBar.taskAdvance()
                    }
                }
            } else {
                val candEdges = new HashSet[(Int, Int, Int)]()
	            for((cIdx, community) <- graph.partition; v1 <- community; v2 <- community; if (!graph.edgeExists(v1,v2) && v1 != v2)) {
	                if(graph(v1).exists(n => graph.edgeExists(v2, n))) { candEdges.add( if(v1 < v2) (v1, v2, cIdx) else (v2, v1, cIdx) ) }
	            }
	            for ((v1, v2, cIdx) <- candEdges.take(updatedMte - graph.nbEdges)) { 
	                graph.addEdge(v1, v2, cIdx, cIdx)
	                ProgressBar.taskAdvance()
	            }
            }
        }
        Timer.stopTimer("MTE")
    }
    
    /*********************************************************************************************************************
     * Micro dynamics
     *********************************************************************************************************************/
    def doMicroDynamics() = {
        val microSteps : List[() => Unit] = List(microUpdateRemoveVertices, microUpdateAddVertices, microUpdateChangeAttributes, 
                microUpdateRemoveWithinEdges, microUpdateAddWithinEdges, microUpdateRemoveBetweenEdges, microUpdateAddBetweenEdges)
        for(f <- microSteps) {
            f()
//            graph.testMinMaxAttributeValues
        }
    }
    
    /*
    def testGraphStruct(message : String) = {
        for(v <- graph.vertices) {
            if(graph.graphTab(v).size != graph.withinDegree(v) + graph.betweenDegree(v)) {
                println(message + " " + v)
                println(graph.graphTab(v).size)
                println(graph.withinDegree(v))
                println(graph.betweenDegree(v))
                System.exit(0)
            }
        }
    }
    */
    
    def microUpdateRemoveVertices() = {
        ProgressBar.startTask("Micro - Remove vertices")
        Timer.startTimer("Micro - Remove vertices")
//        graph.testGraphStruct("rem Vert In")
        if(microProba.get.removeVertices > 0) {
            for((cIdx, com) <- graph.partition; if(Random.nextDouble <= microProba.get.probaMicro)) {
	            val candVertices = {
	                val verticesWithDegreeOne = com.filter(v => graph.degree(v) == 1 && !graph.representatives(cIdx).contains(v))
	                if(verticesWithDegreeOne.isEmpty) {
	                    val verticesWithDegreeTwo = com.filter(v => graph.degree(v) == 2 && !graph.representatives(cIdx).contains(v))
	                    Random.shuffle(verticesWithDegreeTwo.filter(v => {
	                        val n = graph(v).toArray
	                        graph.edgeExists(n(0), n(1))
	                    })).toArray
	                } else { Random.shuffle(verticesWithDegreeOne).toArray }
	            }
	            val nbVerticesToRemove = math.min(math.ceil(com.size * microProba.get.removeVertices), candVertices.size).toInt
	            for(v <- candVertices.take(nbVerticesToRemove)) {
	                graph.removeVertex(v) 
                }
//	            graph.testWithinAndBetweenDegrees()
	        }
        }
//        testGraphStruct("Remove Vert")
        Timer.stopTimer("Micro - Remove vertices")
    }
    
    def microUpdateAddVertices() = {
        ProgressBar.startTask("Micro - Add vertices")
        Timer.startTimer("Micro - Add vertices")
//        graph.testWithinAndBetweenDegrees()
        for((cIdx, com) <- graph.partition; if(Random.nextDouble < microProba.get.probaMicro)) {
            val nbVerticesToAdd = math.ceil((com.size * microProba.get.addVertices)).toInt
            for(_ <- 0 until nbVerticesToAdd) {
                val attributes = graph.minMaxAttributeValues(cIdx).map { a => (Random.nextDouble * (a._2 - a._1)) + a._1 }
                val v = graph.generateVertexWithAttributes(attributes)
	            
	            // Assign the vertex to its community
	            graph.microCommunitiesUpdates.getOrElseUpdate(cIdx, new MicroCommunityUpdate()).nbVerticesAdded += 1
	            Timer.startTimer("Set vertex community")
	            addEdges(v, cIdx)
	            graph.setVertexCommunity(v, cIdx)
	            Timer.stopTimer("Set vertex community")
            }
        }
//        testGraphStruct("Add Vert")
        Timer.stopTimer("Micro - Add vertices")
//        graph.testWithinAndBetweenDegrees()
    }
    
    def microUpdateChangeAttributes() = {
        ProgressBar.startTask("Micro - Update attributes")
        Timer.startTimer("Micro - Update attributes")
        val nbAttributes = graph.attr.find(_ => true).get._2.length
        for((cIdx, com) <- graph.partition; if(Random.nextDouble <= microProba.get.probaMicro)) {
            for(a <- 0 until nbAttributes; _ <- 0 until math.ceil(com.size * microProba.get.updateAttributes).toInt) {
	            val v = Stat.pickAtRandom(com)
	            graph.vertexEvents(v).add(VertexEvents.attributesUpdated)
	            
	            if(graph.withinDegree(v) != 0) {
	                val neighbAttrs = for(v2 <- graph.withinNeighborhood(v).toArray) yield graph.attr(v2)(a)
	//	            val neighbAttrs = for(v2 <- graph(v).toArray) yield graph.attr(v2)(a)
		            graph.attr(v)(a) = neighbAttrs.min + (Random.nextDouble * (neighbAttrs.max - neighbAttrs.min))
	            }
            }
            
            // Update min and max attribute values
            for(attrId <- 0 until nbAttributes) {
	            graph.minMaxAttributeValues(cIdx)(attrId) = com.foldLeft((Double.MaxValue, Double.MinValue))((b, cur) => 
		            (math.min(graph.attr(cur)(attrId), b._1), math.max(graph.attr(cur)(attrId), b._2))
		        )
	        }
        }
//        testGraphStruct("Update")
        Timer.stopTimer("Micro - Update attributes")
    }
    
    
    def microUpdateRemoveWithinEdges() = {
        ProgressBar.startTask("Micro - Remove Within Edges")
        Timer.startTimer("Micro - Remove Within Edges")
//        graph.testWithinAndBetweenDegrees()
        if(microProba.get.removeInternalEdges > 0) {
            for((cIdx, com) <- graph.partition; if(Random.nextDouble <= microProba.get.probaMicro)) {
                val withinEdges = HashSet[(Int, Int)]()
                for(v1 <- com; v2 <- graph(v1); if(v1 < v2 && graph.sameCommunity(v1, v2))) { withinEdges.add((v1, v2)) }
                var i = 0
                var maxNbEdgesToRemove = math.ceil(com.size * microProba.get.removeInternalEdges).toInt
                var candEdgesToRemove = withinEdges.filter(a => !graph(a._1).intersect(graph(a._2)).isEmpty)
	            while(!candEdgesToRemove.isEmpty && i != maxNbEdgesToRemove) {
	                val (v1, v2) = Stat.pickAtRandom(candEdgesToRemove.toArray)
	                graph.removeEdge(v1, v2, cIdx, cIdx)
//	                Timer.startTimer("Update Cand Edges To Remove Before")
//	                withinEdges.remove((v1, v2))
//	                val candEdgesToRemoveBefore = withinEdges.filter(a => !graph(a._1).intersect(graph(a._2)).isEmpty)
//	                Timer.stopTimer("Update Cand Edges To Remove Before")
	                // Update the set of candidate edges to remove
	                Timer.startTimer("Update Cand Edges To Remove Now")
	                candEdgesToRemove.remove((v1, v2))
	                
	                candEdgesToRemove = candEdgesToRemove.filter { case (x1, x2) => {
	                    // This optimization avoids recomputing the whole set of candidate edges
	                    // The idea is to remove an edge between a vertex v and v1 (or v2) iff. v is connected to v2 (or v1) and v and v1 (or v2) does not have a common neighbor
	                    val (isCandForRemoval, otherVertex, vertexInRemovedEdge, vertexToTestForNeighborhood) = if(x1 == v2) { (true, x2, v2, v1) }
	                    else if(x1 == v1) { (true, x2, v1, v2) }
	                    else if(x2 == v2) { (true, x1, v2, v1) }
	                    else if(x2 == v1) { (true, x1, v1, v2) }
	                    else { (false, -1, -1, -1) }
	                    if(isCandForRemoval) {
	                        if(graph.edgeExists(otherVertex, vertexToTestForNeighborhood)) {
	                            !graph(otherVertex).intersect(graph(vertexInRemovedEdge)).isEmpty
	                        } else { true }
	                    } else { true }
	                }}
	                Timer.stopTimer("Update Cand Edges To Remove Now")
	                i += 1
	            }
	        }
        }
//        testGraphStruct("Rem Wth")
        Timer.stopTimer("Micro - Remove Within Edges")
//        graph.testWithinAndBetweenDegrees()
    }
    
    
    def microUpdateAddWithinEdges() = {
        ProgressBar.startTask("Micro - Add Within Edges")
        Timer.startTimer("Micro - Add Within Edges")
//        graph.testWithinAndBetweenDegrees()
        for((cIdx, com) <- graph.partition; if(Random.nextDouble <= microProba.get.probaMicro)) {
            val comToArray = com.toArray
            for(_ <- 0 until math.ceil(com.size * microProba.get.addInternalEdges).toInt) {
                val v1 = comToArray(Random.nextInt(comToArray.length))
                
                if(graph.withinDegree(v1) < com.size - 1) {
	                var sum = 0D
	                val items = for(v2Cand <- comToArray; if(!graph.edgeExists(v1, v2Cand) && v2Cand != v1)) yield {
	                    sum += graph.degree(v2Cand)
	                    v2Cand
	                }
			        val idx = Random.nextDouble * sum
			        var acc = 0D
			        val v2 = (items.find { i => acc += graph.degree(i); acc >= idx }).get
	                
	                graph.addEdge(v1, v2, cIdx, cIdx)
                }
                /*
                val vCand = com.filter(v => graph.withinDegree(v) < com.size - 1)
                if(!vCand.isEmpty) {
                    val v1 = Stat.pickAtRandom(vCand)
//                    Timer.startTimer("weightedProba 1")
//	                Stat.weightedProbability[Int](com.filter(v => !graph.edgeExists(v1, v) && v != v1), x => graph.degree(x))
//	                Timer.stopTimer("weightedProba 1")
	                
	                Timer.startTimer("weightedProba sum")
	                var sum = 0D
	                val items = for(v2Cand <- com; if(!graph.edgeExists(v1, v2Cand) && v2Cand != v1)) yield {
	                    sum += graph.degree(v2Cand)
	                    v2Cand
	                }
                    Timer.stopTimer("weightedProba sum")
                    Timer.startTimer("weightedProba pick")
			        val idx = Random.nextDouble * sum
			        var acc = 0D
			        val v2 = (items.find { i => acc += graph.degree(i); acc >= idx }).get
	                Timer.stopTimer("weightedProba pick")
	                
	                graph.addEdge(v1, v2, cIdx, cIdx)
                }
                */
            }
        }
//        testGraphStruct("Add Wth")
        Timer.stopTimer("Micro - Add Within Edges")
//        graph.testWithinAndBetweenDegrees()
    }
    
    def microUpdateRemoveBetweenEdges() = {
//        graph.testWithinAndBetweenDegrees()
        ProgressBar.startTask("Micro - Remove Between Edges")
        Timer.startTimer("Micro - Remove Between Edges")
        if(Random.nextDouble <= microProba.get.probaMicro && graph.partition.size > 1 && microProba.get.removeExternalEdges > 0) {
	        val nbExternalEdges = graph.betweenDegree.values.sum / 2
	        for(_ <- 0 until math.ceil(microProba.get.removeExternalEdges * nbExternalEdges).toInt) {
	            Timer.startTimer("Select Community")
	            val communityCand = graph.partition.filter(a => { a._2.exists(v => graph.betweenDegree(v) != 0) })
	            val (cIdx, com) = Stat.pickAtRandom(communityCand.toArray)
	            Timer.stopTimer("Select Community")
	            Timer.startTimer("Select first vertex between")
	            var cands = Array.fill(com.size)(0)
	            var nbCand = 0
	            for(v <- com; if(graph.betweenDegree(v) != 0)) {
	                cands(nbCand) = v
	                nbCand += 1 
                }
	            val v1 = cands(Random.nextInt(nbCand))
//	            val vCand = com.filter(v => graph.betweenDegree(v) != 0)
//	            val v1 = Stat.pickAtRandom(vCand)
                Timer.stopTimer("Select first vertex between")
                Timer.startTimer("Compute distance btw")
	            val (v2, _) = (for(v2 <- graph.betweenNeighborhood(v1)) yield { (v2, graph.distance(v1, v2)) }).maxBy(_._2)
	            Timer.stopTimer("Compute distance btw")
	            Timer.startTimer("Remove Edge btw")
        		graph.removeEdge(v1, v2, graph.vertexCommunities(v1), graph.vertexCommunities(v2))
        		Timer.stopTimer("Remove Edge btw")
	        }
        }
//        testGraphStruct("Rem Btw")
        Timer.stopTimer("Micro - Remove Between Edges")
//        graph.testWithinAndBetweenDegrees()
    }
    
    def microUpdateRemoveBetweenEdgesFast() = {
//        graph.testWithinAndBetweenDegrees()
        ProgressBar.startTask("Micro - Remove Between Edges")
        Timer.startTimer("Micro - Remove Between Edges")
        if(Random.nextDouble <= microProba.get.probaMicro && graph.partition.size > 1 && microProba.get.removeExternalEdges > 0) {
	        val nbExternalEdges = graph.betweenDegree.values.sum / 2
	        val nbEdgesToRemove = math.ceil(microProba.get.removeExternalEdges * nbExternalEdges).toInt
	        var nbEdgesRemoved = 0
	        val communityCand = graph.partition.filter(a => { a._2.exists(v => graph.betweenDegree(v) != 0) }).toArray
	        val cIdxToComArray = HashMap[Int, ArrayBuffer[Int]]()
	        while(nbEdgesRemoved != nbEdgesToRemove) {
	            Timer.startTimer("Select Community")
//	            val communityCand = graph.partition.filter(a => { a._2.exists(v => graph.betweenDegree(v) != 0) })
	            val (cIdx, com) = communityCand(Random.nextInt(communityCand.length))
	            
//	            Timer.startTimer("First approach")
//	            val cands2 = cIdxToComArray.getOrElseUpdate(cIdx, {
//	                val tmp = new ArrayBuffer[Int]()
//		            for(v <- com; if(graph.betweenDegree(v) != 0)) {
//		                tmp += v
//	                }
//	                tmp
//	            })
//	            if(!cands2.isEmpty) {
//	                val idxToPick = Random.nextInt(cands2.length)
//	                val v1 = cands2(idxToPick)
//	                if(graph.betweenDegree(v1) == 1) { cands2.remove(idxToPick) }
//	                if(graph.betweenDegree(v1) != 0) {
//    		            val (v2, _) = (for(v2 <- graph.betweenNeighborhood(v1)) yield { (v2, graph.distance(v1, v2)) }).maxBy(_._2)
//			            val cIdx2 = graph.vertexCommunities(v2)
//		        		graph.removeEdge(v1, v2, graph.vertexCommunities(v1), cIdx2)
//		        		if(cIdxToComArray.contains(cIdx2) && graph.betweenDegree(v2) == 0) {
//		        		    cIdxToComArray.remove(cIdx2)
//		        		}
//		        		nbEdgesRemoved += 1
//	                }
//	                
//	            }
//	        	Timer.stopTimer("First approach")
	            
	        	Timer.startTimer("Second approach")
	            var cands = Array.fill(com.size)(0)
	            var nbCand = 0
	            for(v <- com; if(graph.betweenDegree(v) != 0)) {
	                cands(nbCand) = v
	                nbCand += 1 
                }
	        	if(nbCand != 0) {
	        	    val v1 = cands(Random.nextInt(nbCand))
		            val (v2, _) = (for(v2 <- graph.betweenNeighborhood(v1)) yield { (v2, graph.distance(v1, v2)) }).maxBy(_._2)
	        		graph.removeEdge(v1, v2, graph.vertexCommunities(v1), graph.vertexCommunities(v2))
	        		nbEdgesRemoved += 1
	        	}
	        	Timer.stopTimer("Second approach")
	        }
        }
//        testGraphStruct("Rem Btw")
        Timer.stopTimer("Micro - Remove Between Edges")
//        graph.testWithinAndBetweenDegrees()
    }
    
    def microUpdateAddBetweenEdges() = {
//        graph.testWithinAndBetweenDegrees()
        ProgressBar.startTask("Micro - Add Between Edges")
        Timer.startTimer("Micro - Add Between Edges")
        if(Random.nextDouble <= microProba.get.probaMicro && microProba.get.addExternalEdges > 0 && graph.partition.size > 1) {
            val vCand = (graph.vertices.filter { v => val (wth, btw) = graph(v).partition(v2 => graph.sameCommunity(v, v2)); btw.size < wth.size }).toArray
	        for(_ <- 0 until math.ceil(microProba.get.addExternalEdges * vCand.size).toInt) {
	            val v1 = Stat.pickAtRandom(vCand)
	            val v2 = Stat.inverseWeightedProbability[Int](graph.representativesNotInCommunity(graph.vertexCommunities(v1)), a => graph.distance(v1, a))
	            graph.addEdge(v1, v2, graph.vertexCommunities(v1), graph.vertexCommunities(v2))
	        }
        }
//        testGraphStruct("Add Btw")
        Timer.stopTimer("Micro - Add Between Edges")
//        graph.testWithinAndBetweenDegrees()
    }
}

object Test {
    def main(args: Array[String]): Unit = {
        val microDynamicsProba = MicroProba(
                probaMicro = 0.8D,
                removeVertices = 0.1D, 
                updateAttributes = 0.1D, 
                removeInternalEdges = 0.1D, 
                addInternalEdges = 0.1D, 
                removeExternalEdges = 0.1D, 
                addExternalEdges = 0.1D,
                addVertices = 0.1D,
                probaMerge = 0.1D,
                probaSplit = 0.1D,
                probaChange = 0.1D)
        
        Timer.startTimer("Total")
        val gen = new Generator1(
                k = 3, 
                nbVertices = 1200, 
                dev = Array[Double](15D, 15D), 
                theta = 0.20D, 
                nbRep = 300,
                probaCloseTriangle = 0,
                maxInternalEdges = 20, 
                maxExternalEdges = 7, 
                minimumTotalEdges = 0, 
                seed = Option(1))
        Timer.stopTimer("Total")
        
//        PlotDegreeDistrib.plotGen(gen, "DegreeDistribution.eps", true)

        val measures = new GraphMeasures(gen.graph, gen.dev)
        println("Avg. Degree: " + measures.avgDegree)
        println("Nb Edges: " + measures.nbEdges)
        println("Nb Within Edges: " + measures.nbInternalExternalEdges._1)
        println("Nb Between Edges: " + measures.nbInternalExternalEdges._2)
        println("Modularity: " + measures.modularity)
        Timer.printTimers
        
        //power low fonction verification
        /*
        var e=0D;
        for(max<-2 until 100){
        		val pldist=new PowerLawDistribution(max/2);
        		val exp=pldist.getExponentGivenMean(max/2, max);
        		val v=pldist.pickValue(max);
        		println("mean:"+max/2+" max:"+max+" exp:"+exp +" value:"+v);
        		e=(v.toDouble-(max.toDouble/2))*(v.toDouble-(max.toDouble/2));
        }
        println("ecart:"+e/98);
        */
        
    }
}
