/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package algo

import scala.util.Random
import guidyn.components.ProgressBar
import scala.collection.mutable.HashSet
import guidyn.components.graph.VertexEvent
import algo.structures.DynamicGraph
import algo.structures.MicroCommunityUpdate
import algo.structures.Graph
import utils.Timer
import measures.GraphMeasures
import java.io.PrintStream
import java.io.FileOutputStream
import java.io.File

object TestGeneratorDyn {
    def main(args : Array[String]) : Unit = {
        System.setOut(new PrintStream(new FileOutputStream(new File("runtimes"))))
        
        val microProba = MicroProba(probaMicro = 1, 
                addVertices  = 0.2, 
                removeVertices = 0.2, 
                updateAttributes = 0.2, 
                addExternalEdges = 0.2,
                removeExternalEdges = 0.2, 
        		addInternalEdges = 0.2,
        		removeInternalEdges = 0.2, 
                probaMerge = 0.2D,
                probaSplit = 0.2D,
                probaChange = 0.2D)
        
                
                
        for(seed <- 0 until 2) {
            println("Seed " + seed)
            Timer.startTimer("Total")
            val gen = new GeneratorDyn(k = 1000, nbVertices = 50000, dev = Array(1,1), theta = 0D,
			     nbRep = 10, avgWthDegree = 6, avgBtwDegree = 3, minimumTotalEdges = 0,
			     microProba = microProba, nbTimestamps = 20, 
			     seed = Option(seed), useAverageDegree = false, probaCloseTriangle = 0)
            val g = gen.generateDynamicGraph(computeDynamicMeasure = false)
            Timer.stopTimer("Total")
            Timer.printTimers
//            g.graphs(3).writeGraphCsv("checkMeasures.csv")
//            val m = new GraphMeasures(g.graphs(19), Array(1,1))
//            Timer.printTimers
//            println(m.clusteringCoefficient)
//            println(m.modularity)
//            println(m.diameter)
//            println(m.avgShortestPath)
//            println(m.nbVertices)
//            println(m.nbEdges)
        }
    }
}

class GeneratorDyn(val k: Int, val nbVertices: Int, val dev: Array[Double], val theta: Double,
			     val nbRep: Int, val avgWthDegree: Int, val avgBtwDegree: Int, val minimumTotalEdges: Int,
			     val microProba: MicroProba, val nbTimestamps : Int, 
			     seed : Option[Long] = None, val useAverageDegree : Boolean = false, probaCloseTriangle : Double) {
    Timer.clearAllTimers
    
    // Random seed
    val selectedSeed = seed.getOrElse(System.nanoTime())
    Random.setSeed(selectedSeed)
    
	
    
    def generateDynamicGraph(startingGraph : Option[Graph] = None, computeDynamicMeasure : Boolean = true) : DynamicGraph = {
        
        Timer.startTimer("Dynamic graph generation")
        val startTime = System.currentTimeMillis()
	    val gen = new Generator1(k = k, nbVertices = nbVertices, dev = dev, theta = theta, nbRep = nbRep, 
    		maxInternalEdges = avgWthDegree, maxExternalEdges = avgBtwDegree, minimumTotalEdges = minimumTotalEdges, seed = Option(selectedSeed), 
    		microProba = Option(microProba), useAverageDegree = useAverageDegree, probaCloseTriangle = probaCloseTriangle, skipGraphGeneration = startingGraph.isDefined)
	    
	    gen.graph = startingGraph.getOrElse(gen.graph)
	    val graph = gen.graph
	    val macroDynamic = new MacroDynamic(graph, nbRep, microProba)
		val dynamicGraph = new DynamicGraph()
		
//        println("Seed: " + selectedSeed)
        
	    for(t <- 0 until nbTimestamps) yield {
	        if(t > 0) {
//	            graph.testGraphStruct("A")
	            gen.doMicroDynamics
//	            graph.testGraphStruct("B")
	            macroDynamic.doMacroDynamics
//	            graph.testGraphStruct("C")
	            Timer.startTimer("Update representatives")
	            graph.updateRepresentatives(nbRep)
	            Timer.stopTimer("Update representatives")
	        }
	        dynamicGraph.addGraph(graph.clone)
	        for((cIdx, _) <- graph.microCommunitiesUpdates) { graph.microCommunitiesUpdates(cIdx) = new MicroCommunityUpdate() }
	        for(v <- graph.vertices) { graph.vertexEvents(v) = new HashSet[VertexEvent]() }
	        
	        val runtime = ((System.currentTimeMillis() - startTime).toDouble / 1000)
//	        println(List(t + 1, runtime, graph.nbVertices, graph.nbEdges).mkString("t", "\t", ""))
//	        Timer.printTimers
	        
	        ProgressBar.nextTimestamp()
	    }
        
        if(computeDynamicMeasure) {
            Timer.stopTimer("Dynamic graph generation")
		    dynamicGraph.updateVertexEvents()
		    Timer.printTimers
		    ProgressBar.startTask("Computing Community Measures")
		    dynamicGraph.computeCommunityProperties()
        }
        dynamicGraph
    }
}