/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package algo

import utils.FileUtils
import scala.collection.mutable.HashMap
import scala.io.Source

case class GeneratorParameters(
        seed : Long,
        k : Int,
        nbVertices : Int,
        nbRep : Int,
        theta : Double,
        nbAttributes : Int,
        devs : Array[Double],
        averageDegree : Boolean,
        maxEdgesWithin : Int,
        maxEdgesBetween : Int,
        mte : Int,
        nbTimestamps : Int,
        dynamicParameters : Option[MicroProba]) {
    
    override def toString() = {
        val thetaStr = f"${theta}%.2f"
        val devsStr = devs.take(nbAttributes).map(a => f"${a}%.2f").mkString(",")
        var dynamicParametersStr = ""
        for(d <- dynamicParameters) {
            dynamicParametersStr = s"pMicro${d.probaMicro}_pSplit${d.probaSplit}_pMerge${d.probaMerge}_pMigrate${d.probaChange}_ratioAddVert${d.addVertices}"
        }
//        w.println("addBetweenEdges:" + d.addExternalEdges)
//                w.println("addWithinEdges:" + d.addInternalEdges)
//                w.println("addVertices:" + d.addVertices)
//                w.println("removeBetweenEdges:" + d.removeExternalEdges)
//                w.println("removeWithinEdges:" + d.removeInternalEdges)
//                w.println("removeVertices:" + d.removeVertices)
//                w.println("updateAttributes:" + d.updateAttributes)
            
//        s"nbCom${k}_nbVert${nbVertices}_nbTimestamps${nbTimestamps}_nbRep${nbRep}_theta${thetaStr}_dev${devsStr}_edgeWth${maxEdgesWithin}_edgeBtw${maxEdgesBetween}_mte${mte}_avgDeg${averageDegree}_seed${seed}_${dynamicParametersStr}"
        s"k${k}_n${nbVertices}_t${nbTimestamps}_nRep${nbRep}_th${thetaStr}_dev${devsStr}_eWth${maxEdgesWithin}_eBtw${maxEdgesBetween}_mte${mte}_seed${seed}"
    }
    
    def save(fileName : String) = {
        FileUtils.using(fileName) { w =>
            w.println("seed:" + seed)
            w.println("k:" + k)
            w.println("nbVertices:" + nbVertices)
            w.println("nbRep:" + nbRep)
            w.println("theta:" + theta)
            w.println("nbAttributes:" + nbAttributes)
            w.println("devs:" + devs.mkString(";"))
            w.println("averageDegree:" + averageDegree)
            w.println("maxEdgesWithin:" + maxEdgesWithin)
            w.println("maxEdgesBetween:" + maxEdgesBetween)
            w.println("mte:" + mte)
            w.println("nbTimestamps:" + nbTimestamps)
            
            for(d <- dynamicParameters) {
                w.println("probaMicro:" + d.probaMicro)
                w.println("probaMerge:" + d.probaMerge)
                w.println("probaSplit:" + d.probaSplit)
                w.println("probaChange:" + d.probaChange)
                w.println("addBetweenEdges:" + d.addExternalEdges)
                w.println("addWithinEdges:" + d.addInternalEdges)
                w.println("addVertices:" + d.addVertices)
                w.println("removeBetweenEdges:" + d.removeExternalEdges)
                w.println("removeWithinEdges:" + d.removeInternalEdges)
                w.println("removeVertices:" + d.removeVertices)
                w.println("updateAttributes:" + d.updateAttributes)
            }
        }
    }
}

object GeneratorParameters {
    def read(fileName : String) = {
        val p = HashMap[String, String]()
        for(line <- Source.fromFile(fileName).getLines) { p(line.split(":")(0)) = line.split(":")(1) }
        
        val microProba = MicroProba(
                probaMicro = p("probaMicro").toDouble, 
                addVertices = p("addVertices").toDouble,
                removeVertices = p("removeVertices").toDouble, 
                updateAttributes = p("updateAttributes").toDouble,
                removeInternalEdges = p("removeWithinEdges").toDouble,
                addInternalEdges = p("addWithinEdges").toDouble,
                removeExternalEdges = p("removeBetweenEdges").toDouble,
                addExternalEdges = p("addBetweenEdges").toDouble,
                probaMerge = p("probaMerge").toDouble,
                probaSplit = p("probaSplit").toDouble,
                probaChange = p("probaChange").toDouble)
        
        GeneratorParameters(
                seed = p("seed").toLong,
		        k = p("k").toInt,
		        nbVertices = p("nbVertices").toInt,
		        nbRep = p("nbRep").toInt,
		        theta = p("theta").toDouble,
		        nbAttributes = p("nbAttributes").toInt,
		        devs = p("devs").split(";").map(_.toDouble),
		        averageDegree = p("averageDegree").toBoolean,
		        maxEdgesWithin = p("maxEdgesWithin").toInt,
		        maxEdgesBetween = p("maxEdgesBetween").toInt,
		        mte = p("mte").toInt,
		        nbTimestamps = p("nbTimestamps").toInt,
		        dynamicParameters = Option(microProba))
        
    }
    
    def read(gen : GeneratorDyn) = {
        GeneratorParameters(
                seed = gen.selectedSeed,
		        k = gen.k,
		        nbVertices = gen.nbVertices,
		        nbRep = gen.nbRep,
		        theta = gen.theta,
		        nbAttributes = gen.dev.size,
		        devs = gen.dev,
		        averageDegree = false,
		        maxEdgesWithin = gen.avgWthDegree,
		        maxEdgesBetween = gen.avgBtwDegree,
		        mte = gen.minimumTotalEdges,
		        nbTimestamps = gen.nbTimestamps,
		        dynamicParameters = Option(gen.microProba))
    }
}