/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package algo

import scala.collection.mutable.HashSet
import utils.Stat
import scala.collection.mutable.HashMap

/**
 * Implementation of the K-Means algorithm
 */
object KMeans {
    def cluster(data: Seq[(Int, Array[Double])], k: Int) : (Array[HashSet[Int]], Array[Array[Double]]) = {
        var means = data.take(k).map(_._2)
        var clusters = new Array[HashSet[Int]](k)
        var converged = false
        while (!converged) {
            val newClusters = data.groupBy { x => (means.zipWithIndex.map( c => (distance(c._1, x), c._2)).minBy(_._1))._2 }
            val newMeans = new Array[Array[Double]](k)
            for((cIdx, points) <- newClusters) { newMeans(cIdx) = Stat.gravityCenterOfPoints((points.map(_._2).toArray)) }
            converged = (newMeans zip means).forall { case (a,b) => Stat.distance(a,b) < 1E-4 }
            if(converged) { for((cIdx, values) <- newClusters) { clusters(cIdx) = HashSet[Int](values.map(_._1) : _*) } }
            means = newMeans
        }
        (clusters, means.toArray)
    }
    
    def distance(a: Array[Double], b: (Int, Array[Double])) : Double = Stat.distance(a, b._2)
    
}

object KMedoids {
    def cluster(data: Array[Int], k: Int, attr : HashMap[Int, Array[Double]]) : HashMap[Int, HashSet[Int]] = {
        var medoids = data.take(k)
        var clusters = Array.fill(k)(HashSet[Int]())
        var converged = false
        while (!converged) {
            converged = true
            val prevCluster = clusters
            clusters = Array.fill(k)(HashSet[Int]())
            for(p <- data) {
                // Find closest community
                var bestCIdx = -1
                var bestDistance = Double.MaxValue
                for((medoid, i) <- medoids.zipWithIndex) { 
                    val d = Stat.distance(attr(p), attr(medoid))
                    if(d < bestDistance) {
                        bestDistance = d
                        bestCIdx = i
                    }
                }
                clusters(bestCIdx).add(p)
                if(!prevCluster(bestCIdx).contains(p)) {
                    converged = false
                }
            }
            
            // Compute new medoids
            medoids = for(cluster <- clusters) yield {
                var newMedoid = -1
                var bestDistance = Double.MaxValue
                for(p1 <- cluster) {
                    var sumDist = 0D
                    for(p2 <- cluster; if(p1 != p2)) { sumDist += Stat.distance(attr(p1), attr(p2)) }
                    if(sumDist < bestDistance) {
                        bestDistance = sumDist
                        newMedoid = p1
                    }
                }
                newMedoid
            }
        }
        
        // Resize the clusters
        val minClusterSize = clusters.map(_.size).min
		val newClusters = HashMap[Int, HashSet[Int]]()
		    for ((p, cIdx) <- clusters.zipWithIndex) {
            val g = Stat.gravityCenterOfPoints(p.toArray.map(attr))
            newClusters(cIdx) = HashSet((for (v <- p.toArray) yield { (v, Stat.distance(attr(v), g)) }).sortBy(_._2).take(minClusterSize).map(_._1) : _*)
        }
        newClusters
    }
    
    def clusterOld(data: Array[Int], k: Int, attr : Array[Array[Double]]) : (Array[HashSet[Int]], Array[Int]) = {
        var medoids = data.take(k)
        var clusters = Array.fill(k)(HashSet[Int]())
        var converged = false
        while (!converged) {
            converged = true
            val prevCluster = clusters
            clusters = Array.fill(k)(HashSet[Int]())
            for(p <- data) {
                // Find closest community
                var bestCIdx = -1
                var bestDistance = Double.MaxValue
                for((medoid, i) <- medoids.zipWithIndex) { 
                    val d = Stat.distance(attr(p), attr(medoid))
                    if(d < bestDistance) {
                        bestDistance = d
                        bestCIdx = i
                    }
                }
                clusters(bestCIdx).add(p)
                if(!prevCluster(bestCIdx).contains(p)) {
                    converged = false
                }
            }
            
            // Compute new medoids
            medoids = for(cluster <- clusters) yield {
                var newMedoid = -1
                var bestDistance = Double.MaxValue
                for(p1 <- cluster) {
                    var sumDist = 0D
                    for(p2 <- cluster; if(p1 != p2)) { sumDist += Stat.distance(attr(p1), attr(p2)) }
                    if(sumDist < bestDistance) {
                        bestDistance = sumDist
                        newMedoid = p1
                    }
                }
                newMedoid
            }
        }
        
        (clusters, medoids)
    }
    
}
