/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package algo

import scala.Array.canBuildFrom
import scala.collection.mutable.HashMap
import scala.collection.mutable.HashSet
import guidyn.components.ProgressBar
import guidyn.components.graph.VertexEvents
import utils.Stat
import algo.structures.Graph
import scala.util.Random
import utils.Timer
import utils.MBitSet

/**
 * This class implement the macro operations
 */
class MacroDynamic(graph : Graph, nbRep : Int, proba : MicroProba) {
    def doMacroDynamics() = {
        if(!Stat.doWithProbability(proba.probaChange, migrateVertices)) { 
	        if(graph.partition.size > 1) { Stat.doWithProbability(proba.probaMerge, mergeCommunities) }
	        Stat.doWithProbability(proba.probaSplit, splitCommunities)
        }
    }
    
    def migrateVertices() = {
//        graph.testGraphStruct("migrateVertices In")
        Timer.startTimer("Macro - Migrate")
        ProgressBar.startTask("Macro - Migrate vertices")

        // Select source and destination communities
        val communitiesIdx = HashSet[Int](graph.partition.keys.filter(i => graph.partition(i).size >= 5).toSeq : _*)
        val cFromIdx = Stat.pickAtRandom(communitiesIdx)
        val (cToIdx, vSource) = if(Random.nextDouble <= proba.probaMerge && communitiesIdx.size > 1) {
            communitiesIdx.remove(cFromIdx)
            (Stat.pickAtRandom(communitiesIdx), Stat.pickAtRandom(graph.partition(cFromIdx)))
        } else {
            // Destination is a new community
            val cToIdx = graph.createCommunity(new HashSet[Int])
            val potentialNewLeader = graph.partition(cFromIdx).filter(x => graph(x).exists(y => graph.degree(y) == 1))
            val vSource = Stat.pickAtRandom(if(potentialNewLeader.isEmpty) { graph.partition(cFromIdx) } else { potentialNewLeader })  
            (cToIdx, vSource)
        }
        
        graph.vertexEvents(vSource).add(VertexEvents.leaderForCommunityMigrate)
        
        // Build the set of migrating vertices
        var vMigrate = HashSet[Int](vSource)
        val vMigrateAll = HashSet[Int](vSource)
        do {
            // Update cTo and cFrom
            for(vUpdated <- vMigrate) { graph.changeVertexCommunity(vUpdated, cFromIdx, cToIdx) }
            
            // Update the set of migrating vertices
            for(x <- vMigrate) { graph.vertexEvents(x).add(VertexEvents.migratingVertices) }
            vMigrate = for(n <- graph(vSource); if(graph.vertexCommunities(n) == cFromIdx && 
                (graph.degreeSet(n, graph.partition(cToIdx)) > graph.degreeSet(n, graph.partition(cFromIdx))
                		))) yield { n }
            for(x <- vMigrate) { vMigrateAll.add(x) }
        } while(!vMigrate.isEmpty)
        
        graph.updateWithinBetweenDegrees(graph.partition(cFromIdx))
        graph.updateWithinBetweenDegrees(graph.partition(cToIdx))
            
        if(graph.partition(cFromIdx).isEmpty) {
            graph.removeCommunity(cFromIdx)
        } else {
            // Ensure connectivity in cFrom
	        val vCand = HashSet[Int]()
	        for(v1 <- vMigrateAll; v2 <- graph(v1); if(graph.partition(cFromIdx).contains(v2))) { vCand.add(v2) }
	        if(!vCand.isEmpty) {
	            var v1 = Stat.pickAtRandom(vCand)
		        vCand.remove(v1)
		        while(!vCand.isEmpty) {
		            val verticesAlreadyConnectedToV1 = vCand.filter(n => graph(v1).contains(n))
		            v1 = if(verticesAlreadyConnectedToV1.isEmpty) {
		                val v2 = Stat.pickAtRandom(vCand)
		                graph.addEdge(v1, v2, cFromIdx, cFromIdx)
		                v2
		            } else {
		                Stat.pickAtRandom(verticesAlreadyConnectedToV1)
		            }
		            vCand.remove(v1)
		        }
	        }
        }
        
        Timer.stopTimer("Macro - Migrate")
        
        // Update the representatives
        graph.updateRepresentatives(nbRep)
//        graph.testGraphStruct("migrateVertices Out")
    }
    
    def mergeCommunities() : Unit = {
//        graph.testGraphStruct("merge In")
        ProgressBar.startTask("Macro - Merge Communities")
        Timer.startTimer("Macro - Merge")
        // Compute the average distance between the communities
        val communityDistances = HashMap[Int, HashMap[Int, Double]]()
        val representatives = for((cIdx, vertices) <- graph.partition) yield (cIdx, Stat.sample(vertices, math.min(nbRep, vertices.size)))
        
        var avgCommunityDistance = 0D
        var counter = 0
        for((i, com1) <- graph.partition; (j, com2) <- graph.partition; if(i != j)) {
            val minDist = (for(v1 <- representatives(i); v2 <- representatives(j)) yield { graph.distance(v1, v2) }).min
            communityDistances.getOrElseUpdate(i, HashMap[Int, Double]())(j) = minDist
            communityDistances.getOrElseUpdate(j, HashMap[Int, Double]())(i) = minDist
            avgCommunityDistance += minDist
            counter += 1
        }
        avgCommunityDistance /= counter
        
        // Select the two communities to merge as the ones having the highest number of between edges
        var bestCommunityMerge = (-1, -1)
        var maxBetweenEdges = -1
        for((i, com1) <- graph.partition; (j, com2) <- graph.partition; if(i != j && communityDistances(i)(j) <= avgCommunityDistance)) {
            var nbBetweenEdges = 0 
            for(v1 <- com1; v2 <- graph(v1); if(graph.vertexCommunities(v2) == j)) { nbBetweenEdges += 1 }
            if(nbBetweenEdges >= maxBetweenEdges) {
                bestCommunityMerge = (i, j)
                maxBetweenEdges = nbBetweenEdges
            }
        }
        
        val (com1, com2) = bestCommunityMerge
        val com1Vertices = graph.partition(com1)
        val com2Vertices = graph.partition(com2)
        
        // Build the new community
        val newCom = com1Vertices ++ com2Vertices
        
        var nbBetweenEdges = 0
        for(v1 <- com1Vertices; v2 <- graph(v1); if(com2Vertices.contains(v2))) { nbBetweenEdges += 1 }
        
        // Compute the within density of com1 and com2
        var minDensity = (for(com <- List(com1Vertices, com2Vertices)) yield {
            val communitySize = com.size
            var nbWithinEdges = 0D
            for(v1 <- com; v2 <- graph(v1); if(v1 > v2 && com.contains(v2))) { nbWithinEdges += 1 }
            2 * nbWithinEdges / (communitySize * (communitySize - 1))
        }).min
        if(com1Vertices.size == 1 || com2Vertices.size == 1) {
            minDensity = 0D
        }
        
        // Update the structures
        val newComIdx = graph.createCommunity(newCom)
        graph.removeCommunity(com1)
        graph.removeCommunity(com2)
        
        // Update the representatives
        graph.updateRepresentatives(nbRep)
        
        // Osmar
//        val nbEdgesToAdd = math.max(0, (minDensity * graph.partition(com1).size * graph.partition(com2).size).toInt - nbBetweenEdges)
        
        // Christine
        val nbEdgesToAdd = if(minDensity == 0) 0 else { math.floor((1+minDensity) / (2*minDensity)) }
        var nbEdgesAdded = 0
        while(nbEdgesAdded != nbEdgesToAdd) {
            val v1 = Stat.weightedProbability(com1Vertices, { v : Int => graph.degree(v) } )
            val v2 = Stat.weightedProbability(com2Vertices, { v : Int => graph.degree(v) } )
            graph.addEdge(v1, v2, newComIdx, newComIdx)
            nbEdgesAdded += 1
        }
        Timer.stopTimer("Macro - Merge")
//        graph.testGraphStruct("merge Out")
    }
    
    def splitCommunities() = {
//        graph.testGraphStruct("split In")
        ProgressBar.startTask("Macro - Split Communities")
        Timer.startTimer("Macro - Split")
        val comToSplit = getCommunityByMaxDiameter()
        
        val verticesInCom = graph.partition(comToSplit)
        
        if(verticesInCom.size >= 2) {
	        val verticesWithHighestDegree = ((for(v <- verticesInCom.toArray) yield { (v, graph.degree(v)) }).sortBy(_._2).reverse.take(math.min(6, verticesInCom.size)).map(_._1))
	        val v1 = verticesWithHighestDegree(0)
	        val distV1 = geoDistancesWithinCommunities(v1, verticesInCom)
	        val v2 = (for(v <- verticesWithHighestDegree; if(v != v1)) yield { (v, distV1(v)) } ).maxBy(_._2)._1
	        graph.vertexEvents(v1).add(VertexEvents.leaderForCommunitySplit)
	        graph.vertexEvents(v2).add(VertexEvents.leaderForCommunitySplit)
	        
	        val com1 = HashSet[Int](v1)
	        val com2 = HashSet[Int](v2)
	        
	        // Compute the distance between each vertex in verticesInCom to v1 and v2
	        val distV2 = geoDistancesWithinCommunities(v2, verticesInCom)
	        
	        // Assign the vertices to their new communities
	        for(v <- verticesInCom; if(v != v1 && v != v2)) {
//	            val dV1 = Stat.distance(graph.attr(v1), graph.attr(v))
//	            val dV2 = Stat.distance(graph.attr(v2), graph.attr(v))
//	            if(distV1(v) < distV2(v) || (distV1(v) == distV2(v) && dV1 <= dV2)) { com1.add(v) } else { com2.add(v) }
	            
	            if(distV1(v) <= distV2(v)) { com1.add(v) } else { com2.add(v) }
	        }
	        
	        // Update between edges
	        for(v <- com1; n <- graph(v); if(com2.contains(n))) {
	            Stat.doWithProbability(0.3, { () =>
	                graph.removeEdge(v, n, comToSplit, comToSplit)
	            })
	        }
	        
	        // Update the structures
	        graph.removeCommunity(comToSplit)
	        val cIdx1 = graph.createCommunity(com1)
	        val cIdx2 = graph.createCommunity(com2)
	        
	        // Add within edges
	        for((com, cIdx) <- List((com1, cIdx1), (com2, cIdx2)); v <- com) {
	            var degBtw = 0
	            for(n <- graph(v); if(!graph.sameCommunity(v, n))) { degBtw += 1}
	            
	            while((graph(v).isEmpty || graph.degree(v) - degBtw <= degBtw) && graph.degree(v) - degBtw != com.size - 1) {
	                val candVerticesToConnect = com.filter(x => !graph.edgeExists(x, v))
	//                println(candVerticesToConnect.mkString("--"))
	                if(!candVerticesToConnect.isEmpty) {
	                    val v2 = Stat.weightedProbability(candVerticesToConnect, (a : Int) => graph.degree(a) + 1)
	            		if(v != v2) { graph.addEdge(v, v2, cIdx, cIdx) }
	                }
	                
	            }
	        }
	        
	        graph.updateRepresentatives(nbRep)
        }
        Timer.stopTimer("Macro - Split")
//        graph.testGraphStruct("split Out")
    }
    
    /**
     * Compute the size of the shortest path between a given vertex v and each other vertex in vertices 
     */
    def geoDistancesWithinCommunities(v : Int, vertices : HashSet[Int]) : HashMap[Int, Int] = {
        var cand = graph(v).clone
        val remainingVertices = HashSet[Int](vertices.diff(cand).toSeq : _*)
        remainingVertices.remove(v)
        val distances = HashMap[Int, Int]()
        var curDistance = 1
        while(!cand.isEmpty) {
            val newCand = HashSet[Int]()
            for(v2 <- cand) {
                distances(v2) = curDistance
                for(n <- graph(v2); if(remainingVertices.contains(n))) {
                    remainingVertices.remove(n)
                    newCand.add(n)
                }
            }
            cand = newCand
            curDistance += 1
        }
        // Update distance for not reachable vertices
        for(v2 <- remainingVertices) { distances(v2) = Int.MaxValue }
        distances
    }
    
    /**
     * Returns the community id having the highest number of vertices
     */
    def getCommunityByMaxNumberOfVertices() : Int = {
	    (for((cIdx, vertices) <- graph.partition) yield { (vertices.size, cIdx) }).maxBy(_._1)._2
	}
    
    
    /**
     * Returns the community id having the highest number of vertices
     * The first algorithm is easier to understand but slower
     * The second algorithm is slightly more complicated but much faster
     */
    /*
    def getCommunityByMaxDiameterOld() : Int = {
        Timer.startTimer("Max diameter community")
        var largestDiameter = 0
        var bestCommunity = -1
        var curDiameter = 0
	    for((cIdx, com) <- graph.partition) {
	        val vertices = com.clone
	        for(v <- vertices) {
	            var cand = graph(v)
		        val remainingVertices = vertices.diff(cand)
		        remainingVertices.remove(v)
		        curDiameter = 0
		        while(!cand.isEmpty) {
		            val newCand = HashSet[Int]()
		            for(v2 <- cand; n <- graph(v2); if(remainingVertices.remove(n))) { newCand.add(n) }
		            cand = newCand
		            curDiameter += 1
		        }
	            if(curDiameter > largestDiameter) {
		            bestCommunity = cIdx
		            largestDiameter = curDiameter
		        }
	        }
	    }
        Timer.stopTimer("Max diameter community")
//        println("Largest Diam 1: " + largestDiameter)
	    bestCommunity
	}
	*/
    
    /**
     * Return the id of the community having the largest diameter
     * This algorithm consider each node of the community and for each node
     * expand its neighborhood until no more vertices can be visited.
     */
    def getCommunityByMaxDiameter() : Int = {
        var largestDiameter = 0
        var bestCommunity = -1
        
	    for((cIdx, com) <- graph.partition) {
	        
	        // Build a new community graph having an efficient in memory struture
	        val nbVertices = com.size
	        val mappings = HashMap[Int, Int]()
	        val graph2 = new Array[Array[Int]](com.size)
	        for((v1, i) <- com.zipWithIndex) { mappings(v1) = i }
	        for((v1, i) <- com.zipWithIndex) {
	            val withinNeighbors = graph(v1).filter(n => com.contains(n))
	            graph2(i) = new Array[Int](withinNeighbors.size)
	            for((n, i2) <- withinNeighbors.zipWithIndex) {
	                graph2(i)(i2) = mappings(n) 
                }
	        }
	        
	        var cand = Array.fill(2 * nbVertices)(0)
	        var nbCand = 0
	        var v = 0
	        
	        // Main loop
	        for(vIdx <- com) {
	            val v = mappings(vIdx)
	            val remainingVertices = new MBitSet(nbVertices)
	            for(i <- 0 until nbVertices) { remainingVertices.add(i) }
	            
	            // Build the initial set of candidates and remaining vertices
	            for(x <- graph2(v)) {
	                cand(nbCand) = x
	                remainingVertices.remove(x)
	                nbCand += 1
	            }
		        remainingVertices.remove(v)
	        	
		        var curDiameter = 0
		        while(nbCand != 0 && nbCand != nbVertices) {
		            var (newNbCand, i) = if(nbCand > nbVertices) (0, nbVertices) else (nbVertices, 0)
		            while(i != nbCand) {
		                val v2 = cand(i)
		                var j = 0
		                val neighbors = graph2(v2)
		                val nbNeighbors = neighbors.length
		                while(j != nbNeighbors) {
		                    val n = neighbors(j)
		                    if(remainingVertices.remove(n)) {
		                        cand(newNbCand) = n
	                    		newNbCand += 1
		                    }
		                    j +=1
		                }
		                i += 1
		            }
		            nbCand = newNbCand
		            curDiameter += 1
		        }
		        if(curDiameter > largestDiameter) {
		            bestCommunity = cIdx
		            largestDiameter = curDiameter
		        }
	        }
	    }
        bestCommunity
    }
    
    /*
    def getCommunityDiameter(community : HashSet[Int], g : Graph) {
        var curDiameter = 0
        var largestDiameter = 0
        val vertices = community.clone
        for(v <- vertices) {
            var cand = g(v)
	        val remainingVertices = vertices.diff(cand)
	        remainingVertices.remove(v)
	        curDiameter = 0
	        while(!cand.isEmpty) {
	            val newCand = HashSet[Int]()
	            for(v2 <- cand; n <- g(v2); if(remainingVertices.remove(n))) { newCand.add(n) }
	            cand = newCand
	            curDiameter += 1
	        }
            if(curDiameter > largestDiameter) { largestDiameter = curDiameter }
        }
        largestDiameter
    }
    */
    
    /**
     * Returns the community id having the highest within inertia
     */
	def getCommunityByMaximalWithinInertia() : Int = {
	    (for((cIdx, vertices) <- graph.partition) yield {
	        val gCommunity = Stat.gravityCenterOfPoints(vertices.toArray.map(graph.attr))
            val withinInertia = (for (v <- vertices) yield { Stat.distanceSquared(gCommunity, graph.attr(v)) }).sum
            (cIdx, withinInertia)
	    }).maxBy(_._2)._1
	}
}