/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package algo

import java.io.PrintWriter
import edu.uci.ics.jung.graph.UndirectedSparseGraph
import edu.uci.ics.jung.visualization.VisualizationImageServer
import java.awt.Dimension
import edu.uci.ics.jung.algorithms.layout.CircleLayout
import edu.uci.ics.jung.algorithms.layout.KKLayout
import utils.PdfOutput
import java.awt.Paint
import org.apache.commons.collections15.Transformer
import java.awt.Color
import utils.Colors
import edu.uci.ics.jung.visualization.decorators.EdgeShape
import edu.uci.ics.jung.io.GraphMLWriter
import java.io.FileWriter
import edu.uci.ics.jung.graph.Hypergraph
import edu.uci.ics.jung.algorithms.layout.FRLayout2
import edu.uci.ics.jung.algorithms.layout.FRLayout
import gui.graph.GraphView
import gui.components.DegreeDistribution
import algo.structures.Graph

/**
 * This class is used to output a generated graph into different file format.
 */
class Output(gen : Generator1) {

    /**
     * Custom CSV based file format with the generator parameters
     */
    def outputCSV(file : String) = {
        val out = new PrintWriter(file)
        out.println("# Parameters")
        out.println("# Random Seed: " + gen.selectedSeed)
        out.println("# Nb communities: " + gen.k)
        out.println("# Nb Vertice: " + gen.nbVertices)
        out.println("# Standard Deviation: " + gen.dev.mkString(","))
        out.println("# Nb Representents: " + gen.nbRep)
        out.println("# Max. Within Edges: " + gen.maxInternalEdges)
        out.println("# Max. Between Edges: " + gen.maxExternalEdges)
        out.println("# Theta: " + gen.theta)
        out.println("# MTE: " + gen.minimumTotalEdges)
        out.println("#")
        out.println("# Vertices")
        for(v1 <- 0 until gen.nbVertices) {
            out.println(v1 + ";" + gen.graph.attr(v1)(0) + ";" + gen.graph.attr(v1)(1) + ";" + gen.graph.vertexCommunities(v1))
        }
        out.println("#")
        out.println("# Edges")
        for(v1 <- 0 until gen.nbVertices; v2 <- gen.graph(v1)) {
            out.println(v1 + ";" + v2)
        }
        out.flush()
        out.close()
    }
    
    /**
     * Output the network as an image in pdf using the Fruchterman Reingold layout 
     */
    def toJungGraph(graph : Graph, outPath : String, replusionMultiplier : Double = 0.1D, attractionMultiplier : Double = 0.5D, size : Int = 1600) = {
        val jungGraph = new UndirectedSparseGraph[Int, Int]()
        for(v <- graph.vertices) { jungGraph.addVertex(v) }
        var edgeId = 0
        for(v1 <- graph.vertices; v2 <- graph(v1)) { 
            jungGraph.addEdge(edgeId, v1, v2)
            edgeId += 1
        }
        
        // val layout = new KKLayout(jungGraph)
        val layout = new FRLayout(jungGraph)
        layout.setRepulsionMultiplier(replusionMultiplier)
        layout.setAttractionMultiplier(attractionMultiplier)
        val vs = new VisualizationImageServer(layout, new Dimension(size, size))
        val palette = Colors.getMatchingColors(gen.k)
        vs.getRenderContext().setVertexFillPaintTransformer(new Transformer[Int, Paint]() { def transform(v : Int) : Paint = { palette(graph.vertexCommunities(v)) } })
        vs.getRenderContext().setEdgeShapeTransformer(new EdgeShape.Line())
//        layout.setMaxIterations(400)
        println("Performing graph layout...")
        while(!layout.done()) { Thread.sleep(2000) }
        println("Done performing graph layout")
        PdfOutput.createPdf(vs, outPath, Option(new Dimension(size, size)))
    }
    
    /*
    def toJungGraphBck(outPath : String) = {
        val jungGraph = new UndirectedSparseGraph[Int, Int]()
        for(v <- 0 until gen.nbVertices) { jungGraph.addVertex(v) }
        var edgeId = 0
        for(v1 <- 0 until gen.nbVertices; v2 <- gen.graph(v1)) { 
            jungGraph.addEdge(edgeId, v1, v2)
            edgeId += 1
        }
        val vertexPaint = new Transformer[Int, Paint]() {
        	val palette = Colors.getMatchingColors(gen.k) 
		    def transform(v : Int) : Paint = {
		        palette(gen.graph.vertexCommunities(v))
		    }
        }
        
        // val layout = new KKLayout(jungGraph)
        val layout = new FRLayout(jungGraph)
        layout.setRepulsionMultiplier(0.1D)
//        layout.setAttractionMultiplier(0.5D)
        val vs = new VisualizationImageServer(layout, new Dimension(800, 800))
        vs.getRenderContext().setVertexFillPaintTransformer(vertexPaint)
        vs.getRenderContext().setEdgeShapeTransformer(new EdgeShape.Line())
        
//        layout.setMaxIterations(400)
        
        while(!layout.done()) {
            println("Waiting...")
            Thread.sleep(500)
        }
        PdfOutput.createPdf(vs, outPath)
    }
    */
    
    def outputPlotXY(file : String) = {
//        val panel = new GraphViewOld()
//		panel.setGraph(gen)
//		PdfOutput.createPdf(panel, file)
    }
    
    /*
    def outputDegreeDistribution(file : String, distribution : Array[(Int, Int)]) = {
		val dd = new DegreeDistribution(gen, distribution)
    }
    */
    
    /**
     * GraphML format (used by some softwares)
     */
    def toGraphML(file : String) = {
        val jungGraph = new UndirectedSparseGraph[Int, Int]()
        for(v <- 0 until gen.nbVertices) { jungGraph.addVertex(v) }
        var edgeId = 0
        for(v1 <- 0 until gen.nbVertices; v2 <- gen.graph(v1)) { 
            jungGraph.addEdge(edgeId, v1, v2)
            edgeId += 1
        }
        val writer = new GraphMLWriter[Int, Int]()
        writer.save(jungGraph, new FileWriter(file))
    }
    
    /**
     * Output in the CSV format used by the Gephi platform
     */
    def outputGephiCSV(file : String) = {
        val out = new PrintWriter(file)
        for(v1 <- 0 until gen.nbVertices; v2 <- gen.graph(v1)) {
            out.println(v1 + "," + v2)
        }
        out.flush()
        out.close()
    }
    
    /**
     * Output for the R tools (not including the attribute values)
     */
    def outputR(fileGraph : String, fileGT : String) = {
        val out = new PrintWriter(fileGraph)
        for(v1 <- 0 until gen.nbVertices; v2 <- gen.graph(v1)) {
            out.println((v1) + " " + (v2))
        }
        out.flush()
        out.close()
        
//        val out2 = new PrintWriter(fileGT)
//        out2.println(gen.vertexCommunities.map(_ + 2).mkString(" "))
//        out2.flush()
//        out2.close()
    }
}