/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package algo

import scala.util.Random
import scala.collection.mutable.HashMap

/**
 * Generate power law distributions for a given means and different maximum values
 */
class PowerLawDistribution(mean : Double) {
    
    val distributions = HashMap[Int, (Double, Array[Double], Double)]()
    
    // Exponent precision
    val delta = math.pow(10, -7)
    
    // Maximum number of iterations
    val maxTest = 100
        
    /**
     * Given a power-law distribution i^(-x), returns an approximate value of x to get an average value of avgDegree 
     * when considering the sum over nbVertices.
     */
    def getExponentGivenMean(mean : Double, maxValue : Int) : Double = {
        var minExp = 0D
        var maxExp = 50D
        var nbTest = 0
        var curMean = 0D
        while(nbTest != maxTest && math.abs(mean - curMean) > delta) {
            val avgExp = minExp + (maxExp - minExp) / 2
            var s1 = 0D
	        var s2 = 0D
	        var i = 1
	        while (i != maxValue + 1) {
	            s1 += math.pow(i, 1-avgExp)
	            s2 += math.pow(i, -avgExp)
	            i += 1
	        }
            if(s1 / s2 < mean) { maxExp = avgExp } else { minExp = avgExp }
            curMean = s1 / s2
            nbTest += 1
        }
        minExp + (maxExp - minExp) / 2 
    }
    
    
    /**
     * Returns an integer value selected using a power law
     */
    def pickValue(maxValue : Int) : Int = {
        val (_, exponentValues, exponentSum) = distributions.getOrElseUpdate(maxValue, {
            val exp = getExponentGivenMean(mean, maxValue)
            val exponentValues = (for(i <- 1 to maxValue) yield { math.pow(i, -exp) }).toArray
            val exponentSum = exponentValues.sum
            (exp, exponentValues, exponentSum)
        })
        
        val idx = Random.nextDouble * exponentSum
        var acc = 0D
        var i = -1
        while(acc < idx) {
            i += 1
            acc += exponentValues(i)
        }
        i + 1
    }
}