/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package algo.structures

import scala.collection.mutable.HashSet
import scala.collection.mutable.HashMap
import scala.collection.mutable.ArrayBuffer
import utils.Colors
import java.awt.Color
import java.io.File
import utils.FileUtils
import guidyn.components.graph.VertexEvents
import scala.Array.canBuildFrom

class DynamicGraph {
    val graphs = ArrayBuffer[Graph]()
    
	val allGraphs = HashMap[Int, HashSet[Int]]()
	
	def nbTimestamps = graphs.size
	
	val communityColors = HashMap[Int, Color]()
	
	val communityProperties = HashMap[(Int, Int, CommunityProperty), Double]()
	def computeCommunityProperties() = {
        for((g, t) <- graphs.zipWithIndex; cIdx <- g.partition.keySet; p <- CommunityProperties.allProperties) {
            communityProperties((cIdx, t, p)) = p.compute(this, cIdx, t)
        }
    }
    
    def getCommunityInformation(cIdx : Int, t : Int) : Array[(String, Double)] = {
        for(p <- CommunityProperties.allProperties) yield { (p.name, communityProperties(cIdx, t, p)) }
    }
	
	var nbAttributes = 0
	
	def addGraph(g : Graph) = {
        graphs.append(g)
        nbAttributes = g.nbAttributes
        for(v1 <- g.vertices; v2 <- g(v1)) { allGraphs.getOrElseUpdate(v1, HashSet[Int]()).add(v2) }
        
        // Update communities colors
        for((cIdx, _) <- g.partition) { communityColors(cIdx) = Color.white }
        val colors = Colors.getMatchingColors(communityColors.size)
        for((cIdx, i) <- communityColors.keySet.zipWithIndex) { communityColors(cIdx) = colors(i) }
    }
    
    def updateVertexEvents() = {
        for((g, t) <- graphs.zipWithIndex) {
            for(v <- g.vertices) {
                val events = g.vertexEvents(v)
	            if((t == 0) || !graphs(t - 1).vertices.contains(v)) events.add(VertexEvents.createVertex)
	            if((t == nbTimestamps - 1) || !graphs(t + 1).vertices.contains(v)) events.add(VertexEvents.removeVertex)
	            if(events.contains(VertexEvents.createVertex) && events.contains(VertexEvents.removeVertex)) { events.add(VertexEvents.singleTimestamp) }
                if(events.isEmpty) { events.add(VertexEvents.noChanges) }
            }
        }
    }
    
    def save(name : String) = {
        val file = new File("out/" + name)
        FileUtils.deleteRecursive(file)
        file.mkdirs()

        // Output the graph
        for((graph, t) <- graphs.zipWithIndex) { graph.writeGraph(s"out/${name}/t${t}.graph") }

        // Output vertex community evolution
        val allVertices = HashMap[Int, HashSet[Int]]()
        for((graph, t) <- graphs.zipWithIndex; v <- graph.vertices) { allVertices.getOrElseUpdate(v, HashSet[Int]()).add(t) } 
        
        // Output vertex dynamics
        FileUtils.using(s"out/${name}/vertexCommunities") { w =>
            for((v, timestamps) <- allVertices.toList.sortBy(_._1)) { w.println(v + " " + timestamps.toList.sorted.mkString(",")) }
        }
    }
}

case class CommunityProperty(name : String, compute : (DynamicGraph, Int, Int) => Double)
object CommunityProperties {
    def getCommunityDiameter(community : HashSet[Int], g : Graph) : Double = {
        var curDiameter = 0
        var largestDiameter = 0
        val vertices = community.clone
        for(v <- vertices) {
            var cand = g(v)
	        val remainingVertices = vertices.diff(cand)
	        remainingVertices.remove(v)
	        curDiameter = 0
	        while(!cand.isEmpty) {
	            val newCand = HashSet[Int]()
	            for(v2 <- cand; n <- g(v2); if(remainingVertices.remove(n))) { newCand.add(n) }
	            cand = newCand
	            curDiameter += 1
	        }
            if(curDiameter > largestDiameter) { largestDiameter = curDiameter }
        }
        largestDiameter
    }
    
    val nbVerticesInCommunity = CommunityProperty("Nb Vertices", { (g : DynamicGraph, cIdx : Int, t : Int) => g.graphs(t).partition(cIdx).size })
    val nbEdges = CommunityProperty("Nb Edges", { (g : DynamicGraph, cIdx : Int, t : Int) =>
        val curGraph = g.graphs(t)
        val vertices = curGraph.partition(cIdx)
        var nbEdges = 0
        for(v1 <- vertices; v2 <- curGraph(v1); if(!vertices.contains(v2) || v1 < v2 )) { nbEdges += 1 }
        nbEdges
        // (for(v <- curGraph.partition(cIdx).toArray) yield { curGraph.degree(v) }).sum / 2 
    })
    val nbWithinEdges = CommunityProperty("Nb Wth. Edges", { (g : DynamicGraph, cIdx : Int, t : Int) =>
        val curGraph = g.graphs(t)
        (for(v <- curGraph.partition(cIdx).toArray) yield { curGraph.withinDegree(v) }).sum / 2 
    })
    val nbBetweenEdges = CommunityProperty("Nb Btw. Edges", { (g : DynamicGraph, cIdx : Int, t : Int) =>
        val curGraph = g.graphs(t)
        (for(v <- curGraph.partition(cIdx).toArray) yield { curGraph.betweenDegree(v) }).sum 
    })
    val age = CommunityProperty("Age", { (g : DynamicGraph, cIdx : Int, t : Int) =>
        var curTimestamp = t
        while(curTimestamp != -1 && g.graphs(curTimestamp).partition.contains(cIdx)) { curTimestamp -= 1 }
        t - curTimestamp
    })
    val diameter = CommunityProperty("Diameter", { (g : DynamicGraph, cIdx : Int, t : Int) =>
        val curGraph = g.graphs(t)
        getCommunityDiameter(curGraph.partition(cIdx), curGraph)
    })
    val allProperties = Array(nbVerticesInCommunity, nbEdges, nbWithinEdges, nbBetweenEdges, age, diameter)
}
