/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package algo.structures

import scala.collection.mutable.HashMap
import scala.collection.mutable.HashSet
import scala.io.Source
import utils.FileUtils
import utils.Stat
import scala.util.Random
import guidyn.components.graph.VertexEvent
import guidyn.components.graph.VertexEvent
import guidyn.components.graph.VertexEvent
import scala.Array.canBuildFrom
import utils.Timer
import scala.collection.mutable.ArrayBuffer

/**
 * This class contains everything related to the graph and the communities
 * Any actions related to the graph should be performed using the methods of this class
 * It also keep internal states to fasten some computations
 * 
 * It is strongly unadvised to performed graph updates without using the available methods since the graph state might get invalid
 * 
 * Potential improvements
 * 
 * - Everything related to the communities should be kept in a separate class
 * - The class attributes should be protected to avoid their manipulations from outside
 */
class Graph {
    val attr = HashMap[Int, Array[Double]]()
    val partition = HashMap[Int, HashSet[Int]]()
    val debug = false
    
    var graphTab = Array.fill(500)(HashSet[Int]())
    var degree = Array.fill(500)(0)
    
    /**
     * Internal structures
     */
    val vertexEvents = new HashMap[Int, HashSet[VertexEvent]]
    val vertexCommunities = new HashMap[Int, Int]()
	for((c, vertices) <- partition; v <- vertices) { vertexCommunities(v) = c }
    val communitiesSumDegree = HashMap[Int, Int]()
    var nbEdges = 0
    var curVertexId = 0
    var curComId = 0
    
    // Within and between neighborhood
    val withinNeighborhood = HashMap[Int, HashSet[Int]]()
    val withinDegree = HashMap[Int, Int]()
    val betweenNeighborhood = HashMap[Int, HashSet[Int]]()
    val betweenDegree = HashMap[Int, Int]()
    
    /**
     * Structures to log the number of micro updates
     */
    var nbBtwEdgesAdded = 0
    var nbBtwEdgesRemoved = 0
    var microCommunitiesUpdates = HashMap[Int, MicroCommunityUpdate]()
    
    /**
     * Min and max values of the attributes in each community
     */
    val minMaxAttributeValues = HashMap[Int, Array[(Double, Double)]]()
    
    /**
     * Communities representatives
     */
    val allRepresentents = HashSet[Int]()
    val representatives = HashMap[Int, HashSet[Int]]()
    val representativesNotInCommunity = HashMap[Int, HashSet[Int]]()
    
    /**
     * Build an exact copy of a graph
     */
    override def clone() : Graph = {
        val g = new Graph()
        g.nbEdges = nbEdges
        g.graphTab = new Array[HashSet[Int]](graphTab.size)
        g.degree = degree.clone()
        for((v, i) <- graphTab.zipWithIndex; if(vertices.contains(i))) { g.graphTab(i) = v.clone }
        
        for(v <- vertices) { g.vertices.add(v) }
        
        for((v, cIdx) <- vertexCommunities) { g.vertexCommunities(v) = cIdx }
        for((cIdx, com) <- partition) { g.partition(cIdx) = com.clone }
        for((cIdx, deg) <- communitiesSumDegree) { g.communitiesSumDegree(cIdx) = deg }
        for(v <- allRepresentents) { g.allRepresentents.add(v) }
        for((cIdx, rep) <- representatives) { g.representatives(cIdx) = rep.clone }
        for((cIdx, rep) <- representativesNotInCommunity) { g.representativesNotInCommunity(cIdx) = rep.clone }
        
        for((v, a) <- attr) { g.attr(v) = a.clone }
        g.nbEdges = nbEdges
        g.curVertexId = curVertexId
        g.curComId = curComId
        g.nbBtwEdgesAdded = nbBtwEdgesAdded
        g.nbBtwEdgesRemoved = nbBtwEdgesRemoved
        for((v, deg) <- withinDegree) g.withinDegree(v) = deg
        for((v, deg) <- betweenDegree) g.betweenDegree(v) = deg
        for((v, neighborhood) <- withinNeighborhood) g.withinNeighborhood(v) = neighborhood.clone
        for((v, neighborhood) <- betweenNeighborhood) g.betweenNeighborhood(v) = neighborhood.clone
        for((v, events) <- vertexEvents) g.vertexEvents(v) = events.clone
        
        for((cIdx, updates) <- microCommunitiesUpdates) { g.microCommunitiesUpdates(cIdx) = updates }
        for((cIdx, minMax) <- minMaxAttributeValues) { g.minMaxAttributeValues(cIdx) = minMax.clone }
        
        g
    }
    
    /**
     * Returns the neighborhood of a vertex using the parenthesis syntax
     */
    def apply(v : Int) = graphTab(v)
    
    /**
     * Methods to access the vertices as either a sequence of vertices or a set of vertices
     */
    val vertices = HashSet[Int]()
    def nbVertices = vertices.size
    
    /*
    def degree(v : Int) = {
        val expected = graphTab(v).size
        val found = degreeTab(v)
        if(expected != found) {
            println("Argh !!!!")
            System.exit(0)
    	}
    	found
    }
    */
    def nbCommunities = partition.size
	
    /**
     * The number of attributes associated to the vertices
     */
    def nbAttributes = {
        val firstVertex = attr.find(x => true)
        if(firstVertex.isDefined) { firstVertex.get._2.length } else { 0 }
    }
    
    def edgeExists(v1 : Int, v2 : Int) : Boolean = graphTab(v1).contains(v2)
	def sameCommunity(v1 : Int, v2 : Int) = vertexCommunities(v1) == vertexCommunities(v2)
    
    /**
     * Add a vertex v to the community having identifier cIdx
     */
    def addCommunityRepresentative(v : Int, cIdx : Int) = { 
        allRepresentents.add(v)
        representatives.getOrElseUpdate(cIdx, HashSet[Int]()).add(v)
        for((curComIdx, _) <- representatives; if(cIdx != curComIdx)) { representativesNotInCommunity.getOrElseUpdate(cIdx, HashSet[Int]()).add(v) }
    }
    
    /**
     * Update the representatives of all the communities
     */
    def updateRepresentatives(nbRep : Int) = {
        allRepresentents.clear
        representativesNotInCommunity.clear
        for((cIdx, vertices) <- partition) { 
            val newRepresentatives = Stat.sample(vertices, math.min(nbRep, vertices.size))
            representatives(cIdx) = newRepresentatives
            allRepresentents ++= newRepresentatives
            representativesNotInCommunity(cIdx) = HashSet[Int]()
        }
        for((cIdx, r) <- representatives) { representativesNotInCommunity(cIdx) = allRepresentents.diff(r) }
    }
    
    def setVertexCommunity(v : Int, c : Int) = { 
        partition.getOrElseUpdate(c, HashSet[Int]()).add(v)
        vertexCommunities(v) = c
        withinDegree(v) = 0
        betweenDegree(v) = 0
        withinNeighborhood(v).clear
        betweenNeighborhood(v).clear
        for(n <- graphTab(v)) {
            if(sameCommunity(v, n)) {
                withinDegree(v) += 1
                withinNeighborhood(v).add(n)
            } else {
                betweenDegree(v) += 1
                betweenNeighborhood(v).add(n)
            }
        }
        
        for(((min, max), attrId) <- minMaxAttributeValues(c).zipWithIndex) {
            minMaxAttributeValues(c)(attrId) = (math.min(attr(v)(attrId), min), math.max(attr(v)(attrId), max))
        }
    }
    
    /**
     * Add a new community in the graph
     * @param vert a set containing the vertices in the new community
     * @return the id of the created community
     */
    def createCommunity(vert : HashSet[Int]) : Int = {
        partition(curComId) = vert.clone
        for(v <- vert) { 
            vertexCommunities(v) = curComId
            communitiesSumDegree(curComId) = communitiesSumDegree.getOrElse(curComId, 0) + degree(v)
        }
        // Update within and between degrees
        for(v <- vert) {
            withinDegree(v) = 0
            betweenDegree(v) = 0
            withinNeighborhood(v).clear
            betweenNeighborhood(v).clear
            for(n <- graphTab(v)) {
                if(sameCommunity(v, n)) {
                    withinDegree(v) += 1
                    withinNeighborhood(v).add(n)
                } else {
                    betweenDegree(v) += 1
                    betweenNeighborhood(v).add(n)
                }
            }
            vertexCommunities(v) = curComId
        }
        
        // Set min and max attribute values of the community
        minMaxAttributeValues(curComId) = Array.fill(nbAttributes)((Double.MaxValue, Double.MinValue))
//        testMinMaxAttributeValues()
        for(attrId <- 0 until nbAttributes) {
            minMaxAttributeValues(curComId)(attrId) = vert.foldLeft((Double.MaxValue, Double.MinValue))((b, cur) => 
	            (math.min(attr(cur)(attrId), b._1), math.max(attr(cur)(attrId), b._2))
	        )
        }
        representatives(curComId) = HashSet[Int]()
        microCommunitiesUpdates(curComId) = new MicroCommunityUpdate()
        curComId += 1
        curComId - 1
    }
    
    def updateWithinBetweenDegrees(vertices : HashSet[Int]) {
        for(v <- vertices) {
            withinDegree(v) = 0
            betweenDegree(v) = 0
            withinNeighborhood(v).clear
            betweenNeighborhood(v).clear
            for(n <- graphTab(v)) {
                if(sameCommunity(v, n)) {
                    withinDegree(v) += 1
                    withinNeighborhood(v).add(n)
                } else {
                    betweenDegree(v) += 1
                    betweenNeighborhood(v).add(n)
                }
            }
        }
    }
    
    /**
     * Update the community of a vertex
     * @param v the vertex having its community updated
     * @param prevComIdx the previous community id of vertex v
     * @param newComIdx the id of the new community of v 
     */
    def changeVertexCommunity(v : Int, prevComIdx : Int, newComIdx : Int) {
        val degV = degree(v)
        communitiesSumDegree(prevComIdx) = communitiesSumDegree.getOrElse(prevComIdx, 0) - degV
        communitiesSumDegree(newComIdx) = communitiesSumDegree.getOrElse(newComIdx, 0) + degV
        
        // Update min and max attribute values of the community if required
        for(((min, max), attrId) <- minMaxAttributeValues(prevComIdx).zipWithIndex; if(min == attr(v)(attrId) || max == attr(v)(attrId))) {
            minMaxAttributeValues(prevComIdx)(attrId) = partition(prevComIdx).filter(_ != v).foldLeft((Double.MaxValue, Double.MinValue))((b, cur) => 
                (math.min(attr(cur)(attrId), b._1), math.max(attr(cur)(attrId), b._2))
            )
        }
        for(((min, max), attrId) <- minMaxAttributeValues(newComIdx).zipWithIndex) {
            minMaxAttributeValues(newComIdx)(attrId) = (math.min(attr(v)(attrId), min), math.max(attr(v)(attrId), max))
        }
        
        vertexCommunities(v) = newComIdx
        partition(prevComIdx).remove(v)
        partition(newComIdx).add(v)
    }
    
    /**
     * Remove a community from the graph
     * @param cIdx The id of the removed community
     */
    def removeCommunity(cIdx : Int) = {
        communitiesSumDegree.remove(cIdx)
        partition.remove(cIdx)
        
        minMaxAttributeValues.remove(cIdx)
        
        // Update the representatives
        representativesNotInCommunity.remove(cIdx)
        for(v <- representatives(cIdx)) { 
            allRepresentents.remove(v)
            for((_, repr) <- representativesNotInCommunity) repr.remove(v)
        }
        representatives.remove(cIdx)
    }
    
    /**
	 * Write an attributed graph having a partition using the format 
	 * used for the reader
	 * @param outFile the path to the file
	 */
	def writeGraph(outFile : String) = {
	    FileUtils.using(outFile) { w =>
            w.println("# Vertices")
	        for(v1 <- vertices.toList.sorted) {
//	            allVertices.getOrElseUpdate(v1, HashSet[Int]()).add(t)
	            w.println(v1 + ";" + attr(v1).mkString("|") + ";" + vertexCommunities(v1)) 
            }
	        w.println("#\n# Edges")
	        for(v1 <- vertices.toList.sorted; v2 <- graphTab(v1)) { w.println(v1 + ";" + v2) }
        }
	}
	
    def writeGraphCsv(outFile : String) = {
	    FileUtils.using(outFile) { w =>
	        for(v1 <- vertices.toList.sorted; v2 <- graphTab(v1)) { w.println(v1 + ";" + v2) }
        }
	}
	
    /**
     * Generate a new vertex associated to a random set of attribute values
     */
    def generateVertex(dev : Array[Double]) : Int = {
//        testWithinAndBetweenDegrees()
        vertices.add(curVertexId)
        if(graphTab.length == curVertexId) {
            // Resize graph array
            val graphTmp = new Array[HashSet[Int]](graphTab.length * 2)
            System.arraycopy(graphTab, 0, graphTmp, 0, graphTab.length)
            graphTab = graphTmp
            
            // Resize degree array
            val degreeTmp = new Array[Int](degree.length * 2)
            System.arraycopy(degree, 0, degreeTmp, 0, degree.length)
            degree = degreeTmp
        }
        
        graphTab(curVertexId) = HashSet[Int]()
        degree(curVertexId) = 0
        attr(curVertexId) = (for (i <- 0 until dev.size) yield { (Random.nextGaussian * dev(i))}).toArray
        withinDegree(curVertexId) = 0
        withinNeighborhood(curVertexId) = HashSet[Int]()
        betweenDegree(curVertexId) = 0
        betweenNeighborhood(curVertexId) = HashSet[Int]()
        vertexEvents(curVertexId) = HashSet[VertexEvent]()
        curVertexId += 1
        curVertexId - 1
    }
    
    def generateVertexWithAttributes(attributes : Array[Double]) : Int = {
        // Reallocate array if necessary
        if(graphTab.length == curVertexId) {
            // Resize graph array
            val graphTmp = new Array[HashSet[Int]](graphTab.length * 2)
            System.arraycopy(graphTab, 0, graphTmp, 0, graphTab.length)
            graphTab = graphTmp
            
            // Resize degree array
            val degreeTmp = new Array[Int](degree.length * 2)
            System.arraycopy(degree, 0, degreeTmp, 0, degree.length)
            degree = degreeTmp
        }
        vertices.add(curVertexId)
        degree(curVertexId) = 0
        
//        graph.getOrElseUpdate(curVertexId, HashSet[Int]())
        graphTab(curVertexId) = new HashSet[Int]()
        attr(curVertexId) = attributes.clone
        withinDegree(curVertexId) = 0
        withinNeighborhood(curVertexId) = HashSet[Int]()
        betweenDegree(curVertexId) = 0
        betweenNeighborhood(curVertexId) = HashSet[Int]()
        vertexEvents(curVertexId) = HashSet[VertexEvent]()
        curVertexId += 1
        curVertexId - 1
    }
    
    /**
     * Insert an edge in the graph
     */
    def addEdge(v1 : Int, v2 : Int, cIdx1 : Int, cIdx2 : Int) : Boolean = {
        if(!graphTab(v1).contains(v2)) {
            
            graphTab(v1).add(v2)
            graphTab(v2).add(v1)
            nbEdges += 1
            communitiesSumDegree(cIdx1) = communitiesSumDegree.getOrElse(cIdx1, 0) + 1
            communitiesSumDegree(cIdx2) = communitiesSumDegree.getOrElse(cIdx2, 0) + 1
            degree(v1) += 1
            degree(v2) += 1
            
            if(cIdx1 == cIdx2) {
                withinDegree(v1) += 1
                withinNeighborhood(v1).add(v2)
                withinDegree(v2) += 1
                withinNeighborhood(v2).add(v1)
            } else {
                betweenDegree(v1) += 1
                betweenNeighborhood(v1).add(v2)
                betweenDegree(v2) += 1
                betweenNeighborhood(v2).add(v1)
            }
            if(cIdx1 == cIdx2) { microCommunitiesUpdates(cIdx1).nbWithinEdgesAdded += 1 }
            true
        } else {
            false 
        }
    }
    
    /*
    def testGraphStruct(message : String) = {
        for(v <- vertices) {
            if(graphTab(v).size != withinDegree(v) + betweenDegree(v)) {
                println(message + " " + v)
                println(graphTab(v).size)
                println(withinDegree(v))
                println(betweenDegree(v))
                println(withinNeighborhood(v).size)
                println(betweenNeighborhood(v).size)
                
                
                for(x <- withinNeighborhood(v)) { if(!graphTab(v).contains(x)) { println("Missing vertex " + x + " in within") } }
                for(x <- betweenNeighborhood(v)) { if(!graphTab(v).contains(x)) { println("Missing vertex " + x + " in between") } }
                
                System.exit(0)
            }
        }
    }
    */
    
    /*
    def testWithinAndBetweenDegrees() = {
        if(debug) {
            for(v <- vertices) {
	            var expectedWithin = 0
	            var expectedBetween = 0
	            for(n <- graph(v)) { 
	                if(sameCommunity(v, n)) {
		                expectedWithin += 1
		                if(!withinNeighborhood(v).contains(n)) {
		                    println("Vertex " + n + " not in the within neighborhood of " + v)
		                    (new Array(0))(-1)
		                }
		            } else {
		                expectedBetween += 1
		                if(!betweenNeighborhood(v).contains(n)) {
		                    println("Vertex " + n + " not in the between neighborhood of " + v)
		                    (new Array(0))(-1)
		                }
	                }
	            }
	            
	            if(expectedWithin != withinDegree(v)) {
	                println("Error in the within degree. Found " + withinDegree(v) + " expected " + expectedWithin)
	                (new Array(0))(-1)
	            }
	            if(expectedBetween != betweenDegree(v)) {
	                println("Error in the between degree. Found " + betweenDegree(v) + " expected " + expectedBetween)
	                (new Array(0))(-1)
	            }
	            if(betweenNeighborhood(v).size != betweenDegree(v)) {
	                println("Incoherent between neighborhood size")
	                (new Array(0))(-1)
	            }
	            if(withinNeighborhood(v).size != withinDegree(v)) {
	                println("Incoherent within neighborhood size")
	                (new Array(0))(-1)
	            }
	        }
        }
    }
    */
    
    /*
    def testMinMaxAttributeValues() {
        val nbAttr = nbAttributes
        for((cIdx, vert) <- partition) {
            for(i <- (0 until nbAttr).toArray) {
                var minAttr = Double.MaxValue
                var maxAttr = Double.MinValue
                for(x <- vert) { minAttr = math.min(minAttr, attr(x)(i)); maxAttr = math.max(maxAttr, attr(x)(i)) }
                if(minAttr != minMaxAttributeValues(cIdx)(i)._1 || maxAttr != minMaxAttributeValues(cIdx)(i)._2) {
                    println("Min expected / found: " + minAttr + " / " + minMaxAttributeValues(cIdx)(i)._1)
                    println("Max expected / found: " + maxAttr + " / " + minMaxAttributeValues(cIdx)(i)._2)
                    (new Array(0))(-1)
                }
            }
    	}
    }
    */
    
    /**
     * Remove an edge
     */
    def removeEdge(v1 : Int, v2 : Int, cIdx1 : Int, cIdx2 : Int) : Boolean = {
        if(graphTab(v2).contains(v1)) {
            graphTab(v1).remove(v2)
            graphTab(v2).remove(v1)
            nbEdges -= 1
            degree(v1) -= 1
            degree(v2) -= 1
            communitiesSumDegree(cIdx1) -= 1
            communitiesSumDegree(cIdx2) -= 1
            if(cIdx1 == cIdx2) {
                withinDegree(v1) -= 1
                withinNeighborhood(v1).remove(v2)
                withinDegree(v2) -= 1
                withinNeighborhood(v2).remove(v1)
            } else {
                betweenDegree(v1) -= 1
                betweenNeighborhood(v1).remove(v2)
                betweenDegree(v2) -= 1
                betweenNeighborhood(v2).remove(v1)
            }
            if(cIdx1 == cIdx2) { microCommunitiesUpdates(cIdx1).nbWithinEdgesRemoved += 1 }
            true
        } else { false }
    }
    
    /**
     * Remove a vertex from the graph
     */
    def removeVertex(v : Int) = {
        val cIdx = vertexCommunities(v)
        
        for(v2 <- graphTab(v).clone) {
            removeEdge(v, v2, cIdx, vertexCommunities(v2))
        }
        // Update min and max attribute values of the community if required
//        testMinMaxAttributeValues()
        for(((min, max), attrId) <- minMaxAttributeValues(cIdx).zipWithIndex; if(min == attr(v)(attrId) || max == attr(v)(attrId))) {
            minMaxAttributeValues(cIdx)(attrId) = partition(cIdx).filter(_ != v).foldLeft((Double.MaxValue, Double.MinValue))((b, cur) => 
                (math.min(attr(cur)(attrId), b._1), math.max(attr(cur)(attrId), b._2))
            )
        }
        
        graphTab(v).clear
        vertices.remove(v)
        attr.remove(v)
        vertexEvents.remove(v)
        
    	// This part should not be performed !!!!!
        // Update within and between degree
        for(n <- withinNeighborhood(v)) {
            withinDegree(n) -= 1
            withinNeighborhood(n).remove(v)
        }
        for(n <- betweenNeighborhood(v)) {
//            println("betweenNeighborhood: " + n)
            betweenDegree(n) -= 1
            betweenNeighborhood(n).remove(v)
        }
        betweenDegree.remove(v)
        betweenNeighborhood.remove(v)
        
        // Remove from its community
        partition(cIdx).remove(v)
        vertexCommunities.remove(v)
        
        // Update the representatives
        if(allRepresentents.contains(v)) {
            allRepresentents.remove(v)
            for((_, r) <- representativesNotInCommunity) { r.remove(v) }
            for((_, r) <- representatives) { r.remove(v) }
        }
        
        // Update the shortest paths
//        shortestPaths.remove(v)
//        for((_, n) <- shortestPaths) { n.remove(v) }
        
        microCommunitiesUpdates(cIdx).nbVerticesRemoved += 1
    }
    
    /**
     * Compute the Euclidean distance between the attribute values associated to two vertices
     * @param v1 the first vertex
     * @param v2 the second vertex
     * @return the euclidean distance
     */
    def distance(v1: Int, v2 : Int) : Double = {
        val attr1 = attr(v1)
        val attr2 = attr(v2)
        var d = 0D
        var i = 0
        while (i != attr1.length) {
            val delta = attr1(i) - attr2(i)
            d += delta * delta
            i += 1
        }
        math.sqrt(d).toFloat
    }
    
    /**
     * Distance between a vertex and a community
     */
    def distanceWithCommunity(v1: Int, cIdx : Int) : Double = {
//        testWithinAndBetweenDegrees()
        var minDistance = 0D
        var sumDistances = 0D
        for (v2 <- representatives(cIdx)) {
            val curDistance = distance(v1, v2)
            minDistance = math.min(curDistance, minDistance) 
            sumDistances += curDistance
        }
        (sumDistances / representatives(cIdx).size) + minDistance
    }
    
	/**
	 * Compute the number of vertices within the set vertices connected to v
	 */
	def degreeSet(v : Int, vertices : HashSet[Int]) = {
        var degW = 0
        val (smallSet, largeSet) = if(graphTab(v).size > vertices.size) { (vertices, graphTab(v)) } else { (graphTab(v), vertices) }
        for(n <- smallSet; if(largeSet.contains(n))) { degW += 1 }
        degW
    }
}

class MicroCommunityUpdate(
        var nbVerticesAdded : Int, 
        var nbVerticesRemoved : Int,
        var nbAttrUpdated : Int,
        var nbWithinEdgesAdded : Int,
        var nbWithinEdgesRemoved : Int) {
    def this() = this(0,0,0,0,0)
}

object GraphBuilder {
    /**
     * Read a graph from a file using the following format:
     * vertexId classId attr1 ... attrN;neighbor1 ... neighborN
     * where all vertexId, classId and neighbors are integers and attrs are double values
     */
	def readGraph(fileName : String) = {
	    val graph = HashMap[Int, HashSet[Int]]()
	    val attributes = HashMap[Int, Array[Double]]()
	    val partition = HashMap[Int, HashSet[Int]]()
	    
	    val lines = Source.fromFile(fileName).getLines
	    
	    // Parse the file
	    for(line <- lines) {
	        val vertex = line.split(";")(0)
	        val elems = vertex.split(" ")
	        val vertexId = elems(0).toInt
	        
	        // Read the attributes
	        attributes(vertexId) = (for(i <- 2 until elems.size) yield { elems(i).toDouble }).toArray
	        
	        // Read the graph
	        for(n <- line.split(";")(1).split(" ").map(a => a.toInt)) {
	            graph.getOrElseUpdate(vertexId, HashSet[Int]()).add(n)
	            graph.getOrElseUpdate(n, HashSet[Int]()).add(vertexId)
	        }
	        
	        // Read the partition
	        partition.getOrElseUpdate(elems(1).toInt, HashSet[Int]()).add(vertexId)
	    }
	    
//	    new Graph(graph, attributes, partition)
	}
	
	def readGraph2(fileName : String, nbRep : Int) : Graph = {
	    val g = new Graph()
	    println(fileName)
	    
	    var maxVertexId = 0
	    var maxComId = 0
	    var isReadingVertices = true
	    for(line <- Source.fromFile(fileName).getLines) {
	        if(line.startsWith("#")) {
	            if(line == "# Edges") {
	                isReadingVertices = false 
	            }
	        } else {
	            if(isReadingVertices) {
	                val elems = line.split(";")
	                val vertexId = elems(0).toInt
	                val vertexCommunity = elems(2).toInt
	                g.communitiesSumDegree(vertexCommunity) = 0
	                g.graphTab(vertexId) = HashSet[Int]()
	                g.attr(vertexId) = elems(1).split("\\|").map(_.toDouble)
	                g.partition.getOrElseUpdate(vertexCommunity, HashSet[Int]()).add(vertexId)
	                g.vertexCommunities(vertexId) = vertexCommunity
	                g.withinNeighborhood(vertexId) = HashSet[Int]()
	                g.betweenNeighborhood(vertexId) = HashSet[Int]()
	                g.vertexEvents(vertexId) = HashSet[VertexEvent]()
	                maxVertexId = math.max(maxVertexId, vertexId)
	                maxComId = math.max(maxComId, vertexCommunity)
	            } else {
	                val (v1, v2) = (line.split(";")(0).toInt, line.split(";")(1).toInt)
	                g.graphTab(v1).add(v2)
	                g.graphTab(v2).add(v1)
	                g.nbEdges += 1
	                
	                // Update the degree structures
	                if(g.sameCommunity(v1, v2)) {
	                    g.withinNeighborhood(v1).add(v2)
	                    g.withinNeighborhood(v2).add(v1)
	                    g.withinDegree(v1) = g.withinDegree.getOrElse(v1, 0) + 1
	                    g.withinDegree(v2) = g.withinDegree.getOrElse(v2, 0) + 1
	                } else {
	                    g.betweenNeighborhood(v1).add(v2)
	                    g.betweenNeighborhood(v2).add(v1)
	                    g.betweenDegree(v1) = g.betweenDegree.getOrElse(v1, 0) + 1
	                    g.betweenDegree(v2) = g.betweenDegree.getOrElse(v2, 0) + 1
	                }
	                g.communitiesSumDegree(g.vertexCommunities(v1)) += 1
	                g.communitiesSumDegree(g.vertexCommunities(v2)) += 1
	            }
	        }
	        g.nbEdges /= 2
	        g.curVertexId = maxVertexId + 1
	        g.curComId = maxComId + 1
	        g.updateRepresentatives(nbRep)
	    }
	    g
	}
}