/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package cli

import scalargs.argtypes.IntArg
import scalargs.argtypes.FlagArg
import java.io.PrintWriter
import measures.GraphMeasures
import algo.Generator1
import scalargs.argtypes.DoubleArg
import algo.Output
import scalargs.argtypes.StringArg
import scalargs.Scalargs
import scalargs.argtypes.ArrayDoubleArg
import scala.collection.mutable.HashMap
import xps.Measures
import java.io.File
import scala.io.Source

object GeneratorCLI extends Scalargs {
    val defaultParametersFileName = "parameters.default"
    val sep = File.separator
    
    val defaultParams = HashMap[String, String]()
    if(new File(defaultParametersFileName).exists()) {
        for(line <- Source.fromFile(defaultParametersFileName).getLines; if(!line.startsWith("#") && !line.trim.isEmpty())) {
	        val elems = line.split("=").map(_.trim())
	        defaultParams(elems(0)) = elems(1)
	    }
    }
    
    val defaultK = defaultParams.getOrElse("k", "3").toInt
    val defaultNbVertices = defaultParams.getOrElse("nbVertices", "100").toInt
    val defaultTheta = defaultParams.getOrElse("theta", "0.05").toDouble
    val defaultNbRep = defaultParams.getOrElse("nbRep", "5").toInt
    val defaultMaxInternalEdges = defaultParams.getOrElse("maxInternalEdges", "3").toInt
    val defaultMaxExternalEdges = defaultParams.getOrElse("maxExternalEdges", "2").toInt
    val defaultMinEdges = defaultParams.getOrElse("minEdges", "500").toInt
    val defaultAttributesDevs = defaultParams.getOrElse("attributesDevs", "7.0,7.0").split(",").map(_.toDouble)
    val defaultSeed = defaultParams.getOrElse("seed", "0").toInt
    val defaultNbRun = defaultParams.getOrElse("nbRun", "100").toInt
    
    override val progName = "Generator"
    override val description = s"An attributed graph generator. Default parameters settings are located in the file '${defaultParametersFileName}'."
    
    val outPath = new StringArg("out", help = "Output file name")
    val k = new IntArg("-K", List("-k"), "Number of communities", default = defaultK)
    val nbVertices = new IntArg("-nbVertices", List("-n"), "Number of vertices", defaultNbVertices)
    val theta = new DoubleArg("-theta", List("-t"), "Theta, between 0 and 1", defaultTheta)
    val nbRepresentatives = new IntArg("-nbRep", List("-r"), "Number of representatives", defaultNbRep)
    val maxInternalEdges = new IntArg("-maxWthEdges", List("-ei"), "Maximum number of within edges", defaultMaxInternalEdges)
    val maxExternalEdges = new IntArg("-maxBtwEdges", List("-ee"), "Maximum number of between edges", defaultMaxExternalEdges)
    val minimumNbEdges = new IntArg("-minEdges", List("-e"), "Minimum number of edges", defaultMinEdges)
    val seed = new IntArg("-seed", List("-s"), "Random generator seed", defaultSeed)
    val nbRun = new IntArg("-nbRun", List("-nr"), "Number of graphs generated", defaultNbRun)
    val computeMeasures = new FlagArg("-measures", List("-m"), "Compute measures on generated graph", false)
    val attributesDevs = new ArrayDoubleArg("-attributeDev", List("-d"), "Standard deviation value of the attributes, separate the attribute using a ','. E.g., 7.1,3.2", defaultAttributesDevs)
    
    def main(args: Array[String]): Unit = {
//        val curArgs = if(args.length == 0) Array("out") else args
        val curArgs = args
        
        if(!parseArgs(curArgs)) {
		    println(getFullHelpMessage())
		    System.exit(-1)
		} else {
	        val nbVerticesV = nbVertices.getVal
		    val kV = k.getVal
		    val thetaV = theta.getVal
		    val nbRepresentativesV = nbRepresentatives.getVal
		    val maxInternalEdgesV = maxInternalEdges.getVal
		    val maxExternalEdgesV = maxExternalEdges.getVal
		    val minimumNbEdgesV = minimumNbEdges.getVal
		    val attributesDevsV = attributesDevs.getVal
		    val seedVal = seed.getVal
		    val nbRunV = nbRun.getVal
		    
		    new File(outPath.getVal).mkdirs()
	        val parametersFile = new PrintWriter(outPath.getVal + ".params")
	        
		    parametersFile.println("# K: " + kV)
		    parametersFile.println("# NbVertices: " + nbVerticesV)
		    parametersFile.println("# Theta: " + thetaV)
		    parametersFile.println("# Nb Representatives: " + nbRepresentativesV)
		    parametersFile.println("# Max Within Edges: " + maxInternalEdgesV)
		    parametersFile.println("# Max Between Edges: " + maxExternalEdgesV)
		    parametersFile.println("# Min Total Edges: " + minimumNbEdgesV)
		    parametersFile.println("# Attributes Devs: " + attributesDevsV.mkString(";"))
		    parametersFile.println("# Seed: " + seedVal)
		    parametersFile.println("# NbRun: " + nbRunV)
		    parametersFile.flush()
	        parametersFile.close()
	        
		    val allMeasures = for(run <- 0 until nbRunV) yield {
		        val curSeed = run + seedVal
		        val gen = new Generator1(k = kV, nbVertices = nbVerticesV, dev = attributesDevsV, theta = thetaV, nbRep = nbRepresentativesV, 
		            maxInternalEdges = maxInternalEdgesV, maxExternalEdges = maxExternalEdgesV, minimumTotalEdges = minimumNbEdgesV, seed = Option(curSeed), probaCloseTriangle = 0)
			    val graph = gen.graph
			    
			    val graphFile = new PrintWriter(outPath.getVal + "_" + curSeed + ".edgeList")
		        val attrFile = new PrintWriter(outPath.getVal + "_" + curSeed + ".attributes")
			    val classFile = new PrintWriter(outPath.getVal + "_" + curSeed + ".community")
		        
		        
			    
			    for(v1 <- graph.vertices.toArray.sorted) {
			        attrFile.println(v1 + " " + graph.attr(v1).mkString(","))
			        classFile.println(v1 + " " + graph.vertexCommunities(v1))
			        for(v2 <- graph(v1)) { graphFile.println(v1 + " " + v2) }
			    }
		        
		        val measures = new GraphMeasures(graph, attributesDevsV)
		        if(computeMeasures.getVal) {
		            val measuresFile = new PrintWriter(outPath.getVal + "_" + curSeed + ".measures")
		            measuresFile.println("Modularity:" + measures.modularity)
		            measuresFile.println("Within Inertia:" + measures.withinInertiaRatio)
		            measuresFile.println("Between Inertia:" + measures.betweenInertiaRatio)
		            measuresFile.println("Clustering Coefficient:" + measures.clusteringCoefficient)
		            measuresFile.println("Average Degree:" + measures.avgDegree)
		            measuresFile.println("Nb Vertices:" + measures.nbVertices)
		            measuresFile.println("Nb Edges:" + measures.nbEdges)
		            measuresFile.println("Nb Within Edges:" + measures.nbWithinEdges)
		            measuresFile.println("Nb Between Edges:" + measures.nbBetweenEdges)
		            measuresFile.flush()
		            measuresFile.close()
		        }

		        attrFile.flush()
		        attrFile.close()
		        graphFile.flush()
		        graphFile.close()
		        classFile.flush()
		        classFile.close()
		        
		        measures
		    }
	        
	        if(computeMeasures.getVal) {
	            var m = utils.Stat.meanStdDev(allMeasures.map(_.modularity))
	            println("Modularity;" + m._1 + ";" + m._2)
	            m = utils.Stat.meanStdDev(allMeasures.map(_.withinInertiaRatio))
	            println("Within Inertia;" + m._1 + ";" + m._2)
	            m = utils.Stat.meanStdDev(allMeasures.map(_.betweenInertiaRatio))
	            println("Between Inertia;" + m._1 + ";" + m._2)
	            m = utils.Stat.meanStdDev(allMeasures.map(_.clusteringCoefficient))
	            println("Clustering Coefficient;" + m._1 + ";" + m._2)
	            m = utils.Stat.meanStdDev(allMeasures.map(_.avgDegree))
	            println("Average Degree;" + m._1 + ";" + m._2)
	            m = utils.Stat.meanStdDev(allMeasures.map(_.nbVertices.toDouble))
	            println("Nb Vertices;" + m._1 + ";" + m._2)
	            m = utils.Stat.meanStdDev(allMeasures.map(_.nbEdges.toDouble))
	            println("Nb Edges;" + m._1 + ";" + m._2)
	            m = utils.Stat.meanStdDev(allMeasures.map(_.nbWithinEdges.toDouble))
	            println("Nb Within Edges;" + m._1 + ";" + m._2)
	            m = utils.Stat.meanStdDev(allMeasures.map(_.nbBetweenEdges.toDouble))
	            println("Nb Between Edges;" + m._1 + ";" + m._2)
	        }
	        
		}
    }
}