/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package cli
import scalargs.argtypes.IntArg
import scalargs.argtypes.StringArg
import scalargs.Scalargs
import scala.io.Source
import java.io.File
import scalargs.argtypes.DoubleArg
import scalargs.argtypes.ArrayDoubleArg
import scala.collection.mutable.HashMap
import measures.GraphMeasures
import java.io.PrintWriter
import utils.PdfOutput
import java.awt.Dimension
import edu.uci.ics.jung.graph.UndirectedSparseGraph
import edu.uci.ics.jung.visualization.VisualizationImageServer
import edu.uci.ics.jung.algorithms.layout.FRLayout
import utils.Colors
import edu.uci.ics.jung.visualization.decorators.EdgeShape
import org.apache.commons.collections15.Transformer
import java.awt.Paint
import java.awt.Color
import algo.Generator1
import algo.structures.Graph
import algo.MacroDynamic
import scala.Array.canBuildFrom
import algo.MicroProba

object GeneratorDynCLI extends Scalargs {
    val defaultParametersFileName = "parameters.default"
    val measuresListFileName = "measures"
    val sep = File.separator
    
    val defaultParams = HashMap[String, String]()
    if(new File(defaultParametersFileName).exists()) {
        for(line <- Source.fromFile(defaultParametersFileName).getLines; if(!line.startsWith("#") && !line.trim.isEmpty())) {
	        val elems = line.split("=").map(_.trim())
	        if(elems.size == 2) { defaultParams(elems(0)) = elems(1) }
	    }
    }
    
    val defaultK = defaultParams.getOrElse("k", "5").toInt
    val defaultNbVertices = defaultParams.getOrElse("nbVertices", "1000").toInt
    val defaultTheta = defaultParams.getOrElse("theta", "0.00").toDouble
    val defaultNbRep = defaultParams.getOrElse("nbRep", "10").toInt
    val defaultAvgWthDegree = defaultParams.getOrElse("avgWthDegree", "6").toInt
    val defaultAvgBtwDegree = defaultParams.getOrElse("avgBtwDegree", "3").toInt
    val defaultMinEdges = defaultParams.getOrElse("minEdges", "0").toInt
    val defaultAttributesDevs = defaultParams.getOrElse("attributesDevs", "1.0,1.0").split(",").map(_.toDouble)
    
    val defaultProbaMicro = defaultParams.getOrElse("probaMicro", "0.8").toDouble
    val defaultProbaAddVertex = defaultParams.getOrElse("probaAddVertex", "0.8").toDouble
    val defaultProbaRemoveVertex = defaultParams.getOrElse("probaRemoveVertex", "0.1").toDouble
    val defaultProbaUpdateAttributes = defaultParams.getOrElse("probaUpdateAttributes", "0.8").toDouble
    val defaultProbaAddBtwEdges = defaultParams.getOrElse("probaAddBtwEdges", "0.8").toDouble
    val defaultProbaRemoveBtwEdges = defaultParams.getOrElse("probaRemoveBtwEdges", "0.8").toDouble
    val defaultProbaAddWthEdges = defaultParams.getOrElse("probaAddWthEdges", "0.8").toDouble
    val defaultProbaRemoveWthEdges = defaultParams.getOrElse("probaRemoveWthEdges", "0.8").toDouble
    
    val defaultProbaMacro = defaultParams.getOrElse("probaMacro", "0.8").toDouble
    val defaultNbTimestamps = defaultParams.getOrElse("nbTimestamps", "10").toInt
    val defaultCloseTriangle = defaultParams.getOrElse("closeTriangle", "0").toDouble
    
    val defaultProbaMerge = defaultParams.getOrElse("mergeCommunities", "0.1").toDouble
    val defaultProbaSplit = defaultParams.getOrElse("splitCommunities", "0.1").toDouble
    val defaultProbaChange = defaultParams.getOrElse("changeCommunities", "0.1").toDouble
    
    override val progName = "DynGenerator"
    override val description = s"An attributed graph generator. Default parameters settings are located in the file '${defaultParametersFileName}'. The list of measures to compute is located in the file '${measuresListFileName}'"
    
    val outPath = new StringArg("outPath", help = "Name of the directory to output the results. The content of the directory will be DELETED.")
    
    val k = new IntArg("-K", List("-k"), "Number of communities", default = defaultK)
    val nbVertices = new IntArg("-nbVertices", List("-n"), "Number of vertices", defaultNbVertices)
    val theta = new DoubleArg("-theta", List("-t"), "Theta, between 0 and 1", defaultTheta)
    val nbRepresentatives = new IntArg("-nbRep", List("-r"), "Number of representatives", defaultNbRep)
    val avgWthDegree = new IntArg("-avgWthDegree", List("-eW"), "Avg. within degree", defaultAvgWthDegree)
    val avgBtwDegree = new IntArg("-avgBtwDegree", List("-eB"), "Avg. between degree", defaultAvgBtwDegree)
    val minimumNbEdges = new IntArg("-minEdges", List("-e"), "Minimum number of edges", defaultMinEdges)
    val attributesDevs = new ArrayDoubleArg("-attributeDev", List("-d"), "Standard deviation value of the attributes, separate the attribute using a ','. E.g., 7.1,3.2", defaultAttributesDevs)
    val closeTriangles = new DoubleArg("-closeTriangle", List("-c"), "Probability to close a triangle during edge insertion", defaultCloseTriangle)
    
    val probaMicro = new DoubleArg("-probaMicro", List("-pm"), "Probability to perform micro update", defaultProbaMicro, group = "Micro updates")
    val probaAddVertex = new DoubleArg("-pAddVertex", List("-pVa"), "Probability to insert a new vertex", defaultProbaAddVertex, group = "Micro updates")
    val probaRemoveVertex = new DoubleArg("-pRemoveVertex", List("-pVr"), "Probability to remove a vertex", defaultProbaRemoveVertex, group = "Micro updates")
    val probaUpdateAttributes = new DoubleArg("-pUpdateAttributes", List("-pAu"), "Probability to update vertex attributes", defaultProbaUpdateAttributes, group = "Micro updates")
    val probaAddBtwEdges = new DoubleArg("-pAddBtwEdges", List("-pBa"), "Probability to add a between edg", defaultProbaAddBtwEdges, group = "Micro updates")
    val probaRemoveBtwEdges = new DoubleArg("-pRemoveBtwEdges", List("-pBr"), "Probability to remove a between edge", defaultProbaRemoveBtwEdges, group = "Micro updates")
    val probaAddWthEdges = new DoubleArg("-pAddWthEdges", List("-pWa"), "Probability to add a within edge", defaultProbaAddWthEdges, group = "Micro updates")
    val probaRemoveWthEdges = new DoubleArg("-pRemoveWthEdges", List("-pWr"), "Probability to remove a within edge", defaultProbaRemoveWthEdges, group = "Micro updates")
    
//    val probaMacro = new DoubleArg("-probaMacro", List("-pM"), "Probability to perform macro update (between 0 and 1)", defaultProbaMacro, group = "Macro updates")
    val nbTimestamps = new IntArg("-nbTimestamps", List("-T"), "Number of timestamps", defaultNbTimestamps, group = "Macro updates")
    val probaMerge = new DoubleArg("-pMergeCommunities", List("-pMc"), "Probability to merge two communities", defaultProbaMerge, group = "Macro updates")
    val probaSplit = new DoubleArg("-pSplitCommunities", List("-pSc"), "Probability to split a community", defaultProbaSplit, group = "Macro updates")
    val probaChange = new DoubleArg("-pChangeCommunities", List("-pCc"), "Probability to change the community of a set of vertices", defaultProbaChange, group = "Macro updates")
    
    def main(args: Array[String]): Unit = {
        val curArgs = if(args.length == 0) Array[String]("out") else args
        // val curArgs = if(args.length == 0) Array[String]() else args
        deleteRecursive(new File(outPath.getVal))
        Thread.sleep(500)
        
        if(!parseArgs(curArgs)) {
		    println(getFullHelpMessage())
		    System.exit(-1)
		} else {
	        new File(outPath.getVal).mkdirs()
	        var parametersFile = new PrintWriter(outPath.getVal + sep + "parameters")
	        var resultsFile = new PrintWriter(outPath.getVal + sep + "resultsSemiColonSeparated.csv")
	        var resultsFileEn = new PrintWriter(outPath.getVal + sep + "resultsCommaSeparated.csv")
		    
		    val nbVerticesV = nbVertices.getVal
		    val kV = k.getVal
		    val thetaV = theta.getVal
		    val nbRepresentativesV = nbRepresentatives.getVal
		    val avgWthDegreeV = avgWthDegree.getVal
		    val avgBtwDegreeV = avgBtwDegree.getVal
		    val minimumNbEdgesV = minimumNbEdges.getVal
		    val attributesDevsV = attributesDevs.getVal
		    val closeTriangleV = closeTriangles.getVal
		    
		    // Initialize the micro probabilities
		    val microProba = MicroProba(probaMicro.getVal, /* probaMacro.getVal, */ probaAddVertex.getVal, probaRemoveVertex.getVal, probaUpdateAttributes.getVal, 
		            probaRemoveWthEdges.getVal, probaAddWthEdges.getVal,
		            probaRemoveBtwEdges.getVal, probaAddBtwEdges.getVal, probaMerge.getVal, probaSplit.getVal, probaChange.getVal)
		    
		    parametersFile.println("# K: " + kV)
		    parametersFile.println("# NbVertices: " + nbVerticesV)
		    parametersFile.println("# Theta: " + thetaV)
		    parametersFile.println("# Nb Representatives: " + nbRepresentativesV)
		    parametersFile.println("# Avg. Btw. Degree: " + avgWthDegreeV)
		    parametersFile.println("# Avg. Wth. Degree: " + avgBtwDegreeV)
		    parametersFile.println("# Min Edges: " + minimumNbEdgesV)
		    parametersFile.println("# Attributes Devs: " + attributesDevsV.mkString(";"))
		    
	        val gen = new Generator1(k = kV, nbVertices = nbVerticesV, dev = attributesDevsV, theta = thetaV, nbRep = nbRepresentativesV, 
            		maxInternalEdges = avgWthDegreeV, maxExternalEdges = avgBtwDegreeV, minimumTotalEdges = minimumNbEdgesV, seed = Option(0), 
            		microProba = Option(microProba), probaCloseTriangle = closeTriangleV)
	        
	        val graph = gen.graph
	        val macroDynamic = new MacroDynamic(graph, nbRepresentativesV, microProba)

//	        println(List("t", "Clustering Coeff.", "Edges", "Vert", "Diameter").mkString("\t"))
	        for(t <- 0 until nbTimestamps.getVal) {
//	            outputGraph(graph, s"graph_${t}.pdf")
	            if(t > 0) {
	                gen.doMicroDynamics
	                macroDynamic.doMacroDynamics
	            }
	            // Output the measures
//	            val m = new GraphMeasures(graph, gen.dev)
//	            val measures = List(t, m.clusteringCoefficient, m.nbEdges, m.nbVertices, m.diameter)
//        		println(measures.mkString("\t"))
	        }
	        
	        parametersFile.flush()
	        parametersFile.close()
	        resultsFileEn.flush()
	        resultsFileEn.close()
	        resultsFile.flush()
	        resultsFile.close()
		}
    }
    
    def outputGraph(graph : Graph, outPath : String, maxIteration : Int = 400) = {
        val jungGraph = new UndirectedSparseGraph[Int, Int]()
        for(v <- graph.vertices) { jungGraph.addVertex(v) }
        var edgeId = 0
        for(v1 <- graph.vertices; v2 <- graph(v1)) { jungGraph.addEdge(edgeId, v1, v2) ; edgeId += 1 }
        // val layout = new KKLayout(jungGraph)
        val layout = new FRLayout(jungGraph)
        layout.setRepulsionMultiplier(0.1D)
        layout.setMaxIterations(400)
//        layout.setAttractionMultiplier(0.5D)
        val vs = new VisualizationImageServer(layout, new Dimension(800, 800))
        val palette = HashMap[Int, Color]()
        val colors = Colors.getMatchingColors(graph.nbCommunities)
        for((cIdx, i) <- graph.partition.keys.zipWithIndex) { palette(cIdx) = colors(i) }
        vs.getRenderContext().setVertexFillPaintTransformer(new Transformer[Int, Paint]() { def transform(v : Int) : Paint = { palette(graph.vertexCommunities(v)) } })
        vs.getRenderContext().setEdgeShapeTransformer(new EdgeShape.Line())
        while(!layout.done()) { println("Waiting..."); Thread.sleep(500) }
        PdfOutput.createPdf(vs, outPath)
    }
    
    def deleteRecursive(path : File) : Unit = {
        if(path.isDirectory()) { for(f <- path.listFiles()) { deleteRecursive(f) } }
        path.delete()
    }
    
    def formatDouble(x : Double) = if(x == Math.floor(x)) { "" + x.toInt } else { f"${x}%.2f" }
    def formatDoubleEn(x : Double) = if(x == Math.floor(x)) { "" + x.toInt } else { f"${x}%.2f".replaceAllLiterally(",", ".") }
    
    def printlne(x : String) = System.err.println(x)
}