/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package gui

import java.awt.BorderLayout
import java.awt.Color
import java.awt.FlowLayout
import java.awt.event.ActionEvent
import java.awt.event.ActionListener
import java.io.File
import scala.collection.mutable.HashMap
import com.alee.extended.button.WebSwitch
import com.alee.laf.button.WebButton
import com.alee.laf.label.WebLabel
import algo.Generator1
import algo.Output
import gui.components.SpinSlider
import javax.swing.BoxLayout
import javax.swing.ImageIcon
import javax.swing.JButton
import javax.swing.JCheckBox
import javax.swing.JFileChooser
import javax.swing.JPanel
import javax.swing.JTextField
import utils.Icon
import utils.SwingEvents._
import gui.graph.GraphView
import gui.components.Savable
import gui.components.SliderDescriptor
import gui.components.SavableManager
import gui.components.RoundedSwitch
import gui.components.TitledSeparator
import utils.SwingEvents._
import guidyn.components.ProgressBar
import gui.graph.GraphOptions
import gui.graph.GraphLayouts

object RunPanel extends Savable with Runnable {
    
    val parameterPanel = new JPanel()
    val l = new BoxLayout(parameterPanel, BoxLayout.Y_AXIS)
    parameterPanel.setLayout(l)
    val kField = new SpinSlider("K:", 1, 1000, 3, help = "Number of communities")
    val nbVerticesSliderDescriptor = SliderDescriptor(1, 7, 1, 1, {e : Int => Math.pow(10, e)})
    val nbVerticesField = new SpinSlider("n:", 1, 1000000, 100, help = "Number of vertices")
    
    val maxNbAttr = 100
    val attributesParams = new Array[SpinSlider](maxNbAttr)
    for(i <- 0 until maxNbAttr) {
        attributesParams(i) = new SpinSlider(s"Dev. ${i + 1}:", 0, 1000, 7, 0.01D, help = s"Standard deviation value of the attribute ${i + 1}")
        attributesParams(i).setVisible(false)
    }
    val nbAttributesField = new SpinSlider("Nb. Attr.:", 1, maxNbAttr, 2, help = "Number of attributes")
    nbAttributesField.onChange { e => for((a, i) <- attributesParams.zipWithIndex) { a.setVisible(i < nbAttributesField.value) } }
    val nbMaxInternalEdgesField = new SpinSlider("Edges Within:", 0, 1000, 3, help = "Maximum number of edges added connecting a newly inserted vertex to vertices in the same community")
    val nbMaxExternalEdgesField = new SpinSlider("Edges Between:", 0, 1000, 2, help = "Maximum number of edges added connecting a newly inserted vertex to vertices in different communities")
    val nbRepresententsField = new SpinSlider("Nb. Rep.:", 1, 100000, 10, help = "Number of representatives per community")
    val thetaField = new SpinSlider("Theta:", 0, 100, 10, help = "Probability for a vertex to be assigned to a random community")
    val minimalTotalEdgesField = new SpinSlider("MTE:", 0, 100000, 500, help = "Minimum number of edges in the graph")
    val seedField = new JTextField(12)
    var runButton = new WebButton("Run", Icon.get("resultset_next"))

    nbRepresententsField.onChange { e =>
        nbRepresententsField.setValue(math.min(nbRepresententsField.value, math.ceil(nbVerticesField.value / kField.value)))
    }
    kField.onChange { e => 
        nbRepresententsField.setValue(math.min(nbRepresententsField.value, math.ceil(nbVerticesField.value / kField.value)))
        kField.setValue(math.min(kField.value, nbVerticesField.value))
    }
    nbVerticesField.onChange { e => 
        nbRepresententsField.setValue(math.min(nbRepresententsField.value, math.ceil(nbVerticesField.value / kField.value)))
        kField.setValue(math.min(kField.value, nbVerticesField.value))
    }
    
    
    val useSeed = new RoundedSwitch()
    var gen: Generator1 = _
//    run.onAction { e => runGenerator }

    
    var curThread : Option[Thread] = None
    runButton.onAction { e =>
        val shouldStop = if(curThread.isDefined) {
            curThread.get.isAlive()
        } else { false }
        if(shouldStop) {
            runButton.setIcon(Icon.get("resultset_next"))
            runButton.setText("Run")
            ProgressBar.progressAction.setText("")
		    ProgressBar.timestamp.setText("Generation Stopped")
		    ProgressBar.progressBar.setIndeterminate(false)
		    ProgressBar.progressBar.setMaximum(1)
		    ProgressBar.progressBar.setValue(1)
            for(th <- curThread) { th.stop() }
        } else {
            runButton.setIcon(Icon.get("cross"))
            runButton.setText("Stop")
            curThread = Option(new Thread(this))
            curThread.get.start()
        }
    }
    
    override def run() = {
        val seed = if(useSeed.isSelected() && !seedField.getText().isEmpty()) Option(seedField.getText().toLong) else None
        val devs = attributesParams.take(nbAttributesField.value.toInt).map(_.value)
        gen = new Generator1(k = kField.value.toInt, nbVertices = nbVerticesField.value.toInt,
            dev = devs, theta = thetaField.value.toDouble / 100,
            nbRep = nbRepresententsField.value.toInt,
            maxInternalEdges = nbMaxInternalEdgesField.value.toInt, maxExternalEdges = nbMaxExternalEdgesField.value.toInt,
            minimumTotalEdges = minimalTotalEdgesField.value.toInt, seed = seed, probaCloseTriangle = 0)
        
        
        ProgressBar.startTask("Updating GUI")
        GraphView.setGraph(gen)
        GraphView.repaint()
        seedField.setText(gen.selectedSeed.toString)

        GraphOptions.layoutComboBoxModel.removeAllElements()
        for(l <- GraphLayouts.allLayouts; if(l != GraphLayouts.plotXY && devs.size < 2) || devs.size >= 2) { GraphOptions.layoutComboBoxModel.addElement(l) }
        
        
        val degDistMap = HashMap[Int, Int]()
        val degrees = for (v1 <- 0 until gen.nbVertices) { degDistMap(gen.graph.degree(v1)) = degDistMap.getOrElse(gen.graph.degree(v1), 0) + 1 }
//        Main.degreeDistribution.setDistribution(degDistMap.toArray.sortBy(a => a._1), Option(gen))
        gui.graph.GraphView.setGraph(gen)
//        Main.degreeDistribution.repaint()
//        Main.communitiesInformationPanel.setGenerator(gen)
        Main.measuresView.setGenerator(gen)
        if (gen.nbVertices <= 500) { Main.measuresView.updateMeasures } else { 
            Main.measuresView.resetMeasures 
            Main.measuresView.updateFastMeasures
        }
        ProgressBar.allDone()
        runButton.setIcon(Icon.get("resultset_next"))
        runButton.setText("Run")
    }

    
    
    val saveButton = new JButton(Icon.get("disk"))
    saveButton.onAction { e =>
        val fileChooser = new JFileChooser()
        fileChooser.setCurrentDirectory(new File(""))
        val ret = fileChooser.showSaveDialog(Main)
        if (ret == JFileChooser.APPROVE_OPTION) {
            val out = new Output(gen)
            val path = fileChooser.getSelectedFile()
            out.outputCSV(path + ".graph")
        }
    }
    
    
    val runPanel = new JPanel(new FlowLayout(FlowLayout.LEFT))
    runPanel.add(runButton)
    runPanel.add(seedField)
    runPanel.add(useSeed)
    runPanel.add(saveButton)
    useSeed.onAction { e => 
        seedField.setEditable(useSeed.isSelected())
        seedField.setEnabled(useSeed.isSelected())
    }
    
    parameterPanel.add(runPanel)
    parameterPanel.add(new TitledSeparator("Communities"))
    parameterPanel.add(kField)
    parameterPanel.add(nbVerticesField)
    parameterPanel.add(nbRepresententsField)
    parameterPanel.add(thetaField)
    parameterPanel.add(new TitledSeparator("Attributes"))
//    parameterPanel.add(meanXField)
    
    parameterPanel.add(nbAttributesField)
    for(i <- 0 until maxNbAttr) { parameterPanel.add(attributesParams(i)) }
    attributesParams(0).setVisible(true)
    attributesParams(1).setVisible(true)
    parameterPanel.add(new TitledSeparator("Edges"))
    parameterPanel.add(nbMaxInternalEdgesField)
    parameterPanel.add(nbMaxExternalEdgesField)
    parameterPanel.add(minimalTotalEdgesField)

    val topPanel = new JPanel(new BorderLayout())
    topPanel.add(parameterPanel, BorderLayout.NORTH)

    def load() = { 
        seedField.setText(SavableManager.load("seed", "1")) 
        useSeed.setSelected(SavableManager.load("useSeed", true))
    }
    def save() = {
        SavableManager.save("seed", seedField.getText())
        SavableManager.save("useSeed", useSeed.isSelected())
    }
}

