/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package gui.graph

import java.awt.BorderLayout
import java.awt.Dimension
import java.awt.FlowLayout
import java.awt.geom.Point2D
import scala.collection.JavaConversions.collectionAsScalaIterable
import com.alee.laf.button.WebButton
import com.alee.laf.combobox.WebComboBox
import com.alee.laf.label.WebLabel
import com.alee.managers.tooltip.TooltipManager
import com.alee.managers.tooltip.TooltipWay
import edu.uci.ics.jung.algorithms.layout.AggregateLayout
import edu.uci.ics.jung.algorithms.layout.FRLayout
import edu.uci.ics.jung.algorithms.layout.ISOMLayout
import edu.uci.ics.jung.algorithms.layout.KKLayout
import edu.uci.ics.jung.algorithms.layout.Layout
import edu.uci.ics.jung.algorithms.util.IterativeContext
import edu.uci.ics.jung.graph.Graph
import javax.swing.JCheckBox
import javax.swing.JPanel
import utils.Icon
import utils.SwingEvents._
import edu.uci.ics.jung.algorithms.layout.StaticLayout
import org.apache.commons.collections15.Transformer
import com.alee.laf.checkbox.WebCheckBox
import com.alee.extended.progress.WebStepProgress
import java.awt.event.MouseAdapter
import java.awt.event.MouseEvent
import javax.swing.DefaultComboBoxModel

object GraphLayouts {
    val plotXY = GraphLayout("PlotXY", classOf[PlotLayout])
    val fruchtermanReingold = GraphLayout("Fruchterman-Reingold", classOf[FRLayout[Int, Int]])
    val kamadaKawai = GraphLayout("Kamada-Kawai", classOf[KKLayout[Int, Int]])
    val selfOrganizingMap = GraphLayout("Self Organizing Map", classOf[ISOMLayout[Int, Int]])

    val allLayouts = Array[Object](plotXY, fruchtermanReingold, kamadaKawai, selfOrganizingMap)
}

case class GraphLayout(label : String, layoutClass : Class[_]) {
    override def toString() = label
}

object GraphOptions extends JPanel(new BorderLayout()) {
    var doComputeLayout = true
    
    def getLayoutFor(layoutClass : Class[_], graph : Graph[Int, Int]) : Layout[Int, Int] = {
        val constructor = layoutClass.getDeclaredConstructor(Array(classOf[Graph[Int, Int]]) : _*)
        val l = constructor.newInstance(Array(graph) : _*).asInstanceOf[Layout[Int, Int]]
        if(l.isInstanceOf[FRLayout[Int, Int]]) {
            l.asInstanceOf[FRLayout[Int, Int]].setAttractionMultiplier(0.08)
            l.asInstanceOf[FRLayout[Int, Int]].setRepulsionMultiplier(0.3)
        }
        l
        
    }
    
    def getSelectedLayout() = {
        /*
        if(useCommunityLayout.isSelected()) {
        	val l = new AggregateLayout[Int, Int](new PlotLayout(GraphView.jungGraph))
            l.setSize(GraphView.getSize())
            for(communityGraph <- GraphView.communitiesGraph) {
                val localLayout = layouts.getSelectedItem() match {
		            case x : GraphLayout => getLayoutFor(x.layoutClass, communityGraph)
		            case _ => getLayoutFor(GraphLayouts.plotXY.layoutClass, communityGraph)
		        }
                localLayout.setSize(new Dimension(600, 600))
                val center = new Point2D.Double()
                for(v <- communityGraph.getVertices()) {
                    center.setLocation(l.transform(v).getX() + center.getX(), l.transform(v).getY() + center.getY())
                }
                center.setLocation(center.getX() / communityGraph.getVertexCount(), center.getY() / communityGraph.getVertexCount())
                localLayout.setInitializer(l)
                localLayout.setSize(new Dimension(300, 300))
                l.put(localLayout, center)
            }
            l
        } else {
        */
            layouts.getSelectedItem() match {
	            case x : GraphLayout => getLayoutFor(x.layoutClass, GraphView.jungGraph)
	            case _ => getLayoutFor(GraphLayouts.plotXY.layoutClass, GraphView.jungGraph)
	        }
//        }
    }
    
    def updateLayout() = {
        val curLayout = getSelectedLayout
        new Thread(new Runnable {
            def run() = {
                stopLayout.setVisible(true)
                GraphView.viewer.setGraphLayout(curLayout)
                if (curLayout.isInstanceOf[IterativeContext]) {
		            val l = curLayout.asInstanceOf[IterativeContext]
		            while (!l.done() && doComputeLayout) { Thread.sleep(500) }
		            if(!doComputeLayout) {
		                // Build a static layout corresponding to the current state
		                val transformer = new Transformer[Int,Point2D] { def transform(v : Int) = curLayout.transform(v) }
		                val staticLayout = new StaticLayout(GraphView.jungGraph, transformer, curLayout.getSize())
		                GraphView.viewer.setGraphLayout(staticLayout)
		            }
		        }
                doComputeLayout = true
                stopLayout.setVisible(false)
            }
        }).start
        GraphView.repaint()
    }
    
    // Components
    val layoutLabel = new WebLabel("Layout:")
    val layoutComboBoxModel = new DefaultComboBoxModel(GraphLayouts.allLayouts) 
    val layouts = new WebComboBox(layoutComboBoxModel)
    layouts.onAction { e => updateLayout() }
//    val useCommunityLayout = new JCheckBox("By community")
//    useCommunityLayout.onAction { e => updateLayout() }
    val stopLayout = new WebButton(Icon.get("cross"))
    stopLayout.setVisible(false)
    TooltipManager.addTooltip(stopLayout, "Stop computing the layout", TooltipWay.down, 0)
    stopLayout.onAction { e => doComputeLayout = false }
    val toggleEdgeDisplay = new WebCheckBox("Display Edges")
    toggleEdgeDisplay.onAction { e => GraphView.repaint() }
    val step = new WebStepProgress(4)
    step.onClick { e => GraphView.repaint() }
    step.onRelease { e => GraphView.repaint() }
    step.onDrag { e => GraphView.repaint() }
    val onlyCurrentStep = new WebCheckBox("Current")
    onlyCurrentStep.onAction { e => GraphView.repaint() }
    val showRepresentatives = new WebCheckBox("Representatives")
    showRepresentatives.onAction { e => GraphView.repaint() }
    
    val highlightBetweenEdges = new WebCheckBox("Highlight between vertices")
    highlightBetweenEdges.onAction { e => GraphView.repaint() }
    
    // Layout the components
    val leftPanel = new JPanel(new FlowLayout(FlowLayout.LEFT))
    leftPanel.add(layoutLabel)
    leftPanel.add(layouts)
    leftPanel.add(stopLayout)
//    leftPanel.add(useCommunityLayout)
    leftPanel.add(toggleEdgeDisplay)
//    leftPanel.add(highlightBetweenEdges)
//    leftPanel.add(step)
//    leftPanel.add(onlyCurrentStep)
//    leftPanel.add(showRepresentatives)
    add(leftPanel)
}