/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package guidyn.components

import java.awt.BasicStroke
import java.awt.Color
import java.awt.Dimension
import java.awt.Font
import java.awt.Graphics
import java.awt.Graphics2D
import java.awt.Point
import java.awt.RenderingHints
import java.awt.event.MouseAdapter
import java.awt.event.MouseEvent
import scala.collection.mutable.HashMap
import scala.collection.mutable.HashSet
import algo.structures.DynamicGraph
import algo.structures.DynamicGraph
import algo.structures.Graph
import guidyn.Main
import javax.swing.JPanel
import javax.swing.JViewport
import utils.Colors
import utils.SwingEvents._
import guidyn.components.listeners.GraphUpdateListener

class CommunityDynamics extends JPanel with GraphUpdateListener {
    val padding = 60
    val columnSpacing = 60
    val minNodeSize = 20
    val maxNodeSize = 50
//    var paddingBottom = 150
    
    val minVerticalspacingBetweenNodes = 30
    val paddingTop = 40
    val paddingBottom = 0
    
    var selectedCommunity : Option[(Int, Int, Graph, Int, Int, Int)] = None
    
    var totalWidth = 0
    var totalHeight = 0
    
    def getDimension() = new Dimension(totalWidth, totalHeight)
    override def getPreferredSize() = getDimension()
//    override def getSize() = getDimension()
//    override def getWidth() = totalWidth
//    override def getHeight() = totalHeight
    
    this.onMove { e =>
        val (px, py) = (e.getX(), e.getY())
        selectedCommunity = communitiesLocation.find { e =>
            val (cIdx, t, graph, x, y, w) = e 
            val (dx, dy) = ((px - x), (py - y))
            math.sqrt(dx * dx + dy * dy) < w
        }
//        for(c <- selectedCommunity) { val (cIdx, t, graph, x, y, w) = c }
        repaint()
    }
    
    val communitiesLocation = HashSet[(Int, Int, Graph, Int, Int, Int)]()
    
    var g = new DynamicGraph()
    var maxNbCommunities = 0
    var largestCommunitySize = 0
    var smallestCommunitySize = 0
    
    def graphUpdated(g : DynamicGraph) = {
        this.g = g
        maxNbCommunities = math.max(2, (for(graph <- g.graphs) yield { graph.partition.size }).max)
        largestCommunitySize = (for(graph <- g.graphs; com <- graph.partition.values) yield { com.size }).max
        smallestCommunitySize = (for(graph <- g.graphs; com <- graph.partition.values) yield { com.size }).min
        
        totalHeight = maxNbCommunities * maxNodeSize + (maxNbCommunities - 1) * minVerticalspacingBetweenNodes + paddingTop + paddingBottom
        totalWidth = g.nbTimestamps * maxNodeSize + columnSpacing * (g.nbTimestamps - 1) + padding
        revalidate()
        repaint()
    }
    
    def getCommunityWidth(size : Int) = {
        minNodeSize + (maxNodeSize - minNodeSize) * (size - smallestCommunitySize).toDouble / (largestCommunitySize - smallestCommunitySize)
    }
    
    def getX(t : Int) = {
        if(getVisibleRect().width >= totalWidth) {
            padding + (getWidth() - 2 * padding) * (t.toDouble / (g.nbTimestamps - 1))
        } else {
            padding + t * (columnSpacing + maxNodeSize)
        }
        
    }
    
    def getY(pos : Int, nbCom : Int) = {
        val h = getHeight() - 2 * padding - paddingBottom
        val ySpacing = h / (maxNbCommunities - 1)
        pos * ySpacing + padding + (h - (nbCom - 1) * ySpacing) / 2
    }
    
    def drawInformations(g : Graphics2D, measures : Array[(String, Double)], px : Int, y : Int, nodeDiameter : Int) = {
        val columnPadding = 20
        val rowPadding = 5
        val topBottomPadding = 10
        
        // Compute height of the box
        val measuresArray = measures.toArray
        val font = g.getFont()
        val fontMetric = g.getFontMetrics()
        val maxNbRows = 5
        val nbRows = math.min(maxNbRows, measuresArray.length)
        val height = nbRows * (font.getSize() + rowPadding) + topBottomPadding * 2  
        
        // Compute width of the box
        var width = 0
        val columnWidths = (for(column <- measuresArray.grouped(maxNbRows)) yield {
            var columnMaxWidth = 0
            for((key, value) <- column) {
                val valStr = if(value == math.floor(value)) value.toInt.toString else f"$value%.3f" 
                columnMaxWidth = math.max(columnMaxWidth, fontMetric.stringWidth(key + ": " + valStr))
            }
            width += (columnMaxWidth + columnPadding)
            columnMaxWidth
        }).toArray
        
        // Locate box position
        val py = y + nodeDiameter + 5
        val topX = math.min(math.max(this.getVisibleRect().x + 4, px - width / 2), getVisibleRect().x + getVisibleRect().width - 4 - width)
        val topY = py
        val triangleWidth = (width * 0.1).toInt
        val triangleHeight = triangleWidth / 2
        val prevColor = g.getColor()
        g.setColor(Color.black)
        
        // Draw box
        g.fillRoundRect(topX, py, width, height, 10, 10)
        val pyt = py + (height - triangleWidth) / 2
        val shapeTopBottomX = Array(px - triangleWidth / 2, px , px + triangleWidth / 2)
        val shapeLeftX = Array(px, px - triangleHeight, px)
        val shapeRightX = Array(px, px + triangleHeight, px)
        val shapeBottomY = Array(py + height, py + height + triangleHeight, py + height)
        val shapeTopY = Array(py, py - triangleHeight, py)
        val shapeLeftRightY = Array(pyt, pyt - triangleHeight / 2, pyt)
        g.fillPolygon(shapeTopBottomX, shapeTopY, 3)
        
        // Draw text
        g.setColor(Color.white)
        var curX = topX + 10
        for((column, i) <- measuresArray.grouped(maxNbRows).zipWithIndex) {
            if(i != 0) curX += columnWidths(i - 1) + columnPadding
            var curY = topY + topBottomPadding + font.getSize()
            
            for((key, value) <- column) {
                val valStr = if(value == math.floor(value)) value.toInt.toString else f"$value%.3f" 
                g.drawString(key + ": " + valStr, curX, curY)
                curY += font.getSize() + rowPadding
            }
            
        }
        g.setColor(prevColor)
    }
    
    def centerText(g : Graphics2D, t : String, x : Int, y : Int, boxed : Boolean = false) = {
        val w = g.getFontMetrics().stringWidth(t)
        val h = g.getFontMetrics().getHeight()
        val padding = 3
        val px = x - w / 2
        if(boxed) {
            g.setColor(Color.white)
            g.fillRoundRect(px - padding, y + padding, w + 2 * padding - 1, h + padding - 3, 4, 4)
            g.setColor(Color.darkGray)
            g.drawRoundRect(px - padding, y + padding, w + 2 * padding - 1, h + padding - 3, 4, 4)
        }
        g.drawString(t, px, y + h)
        
    }
    
    def getXYAlongLine(x1 : Int, y1 : Int, x2 : Int, y2 : Int, ratio : Double) = {
         ((x1 + (x2 - x1) * ratio).toInt, (y1 + (y2 - y1) * ratio).toInt)
    }
    
    override def paint(graphics : Graphics) = {
        val g2d = graphics.asInstanceOf[Graphics2D]
        g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        g2d.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
        val boldFont = g2d.getFont().deriveFont(Font.BOLD)
		val stdFont = g2d.getFont()
		
		g2d.setColor(Color.white)
		g2d.fillRect(0, 0, getWidth(), getHeight())
        communitiesLocation.clear
        g2d.setColor(Color.black)
		
        for((graph, t) <- g.graphs.zipWithIndex) {
            val xt = getX(t).toInt
//            g2d.setFont(boldFont)
            // centerText(g2d, s"t${t + 1}", xt, getHeight() - 30, false)
            centerText(g2d, s"t${t + 1}", xt, 10, false)
//            g2d.setFont(stdFont)
            for(((cIdx, com), pos) <- graph.partition.zipWithIndex) {
                val width = getCommunityWidth(com.size).toInt
                val yt = getY(pos, graph.partition.size)
                val (xc, yc) = (xt - width / 2, yt - width / 2)
                
                
                g2d.setStroke(new BasicStroke(2))
                if(t != g.nbTimestamps - 1) {
                    val nextPartitions = g.graphs(t + 1).partition 
                	for(((cIdxNext, comNext), posNext) <- nextPartitions.zipWithIndex) {
                	    val nbCommonVertices = com.intersect(comNext).size
                	    if(nbCommonVertices != 0) {
                	        val xt2 = getX(t + 1).toInt
                	        val yt2 = getY(posNext, nextPartitions.size)
                	        g2d.drawLine(xt, yt, xt2, yt2)
                	        val (xl, yl) = getXYAlongLine(xt, yt, xt2, yt2, 3D/7)
                	        g2d.setStroke(new BasicStroke(1))
                	        centerText(g2d, nbCommonVertices.toString, xl, yl - 10, true)
                	        g2d.setStroke(new BasicStroke(2))
                	    }
                	}
                }
                
                val curColor = g.communityColors(cIdx)
                val labelColor = 
                
                g2d.setColor(curColor)
                g2d.fillOval(xc, yc, width, width)
                g2d.setColor(Color.black)
                g2d.drawOval(xc, yc, width, width)
                communitiesLocation.add((cIdx, t, graph, xc + width / 2, yc + width / 2, width))
                g2d.setStroke(new BasicStroke(1))
                centerText(g2d, com.size.toString, xt, yc + width, true)
                
                g2d.setColor(Colors.getTextColor(curColor))
                g2d.setFont(boldFont)
                centerText(g2d, cIdx.toString, xt, yt - 10, false)
                g2d.setFont(stdFont)
                g2d.setColor(Color.black)
            }
        }
        
        if(selectedCommunity.isDefined) {
		    val (cIdx, t, graph, xc, yc, wc) = selectedCommunity.get
		    drawInformations(g2d, g.getCommunityInformation(cIdx, t), xc, yc, wc)
		}
    }
}

class HandScrollListener extends MouseAdapter {
    val pp = new Point()

    override def mouseDragged(e : MouseEvent) = {
        val source = e.getSource().asInstanceOf[JPanel]
        val vport = Main.xx.getViewport()
        val cp = e.getPoint()
        val vp = vport.getViewPosition()
        
        vp.translate(pp.x-cp.x, pp.y-cp.y)
        vp.setLocation(math.min(0, vp.x), math.min(0, vp.y))
        vport.setViewPosition(vp)
    }

    override def mousePressed(e : MouseEvent) = {
        pp.setLocation(e.getPoint());
    }
}
