/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package guidyn.components

import java.awt.BorderLayout
import java.awt.Color
import java.awt.FlowLayout
import scala.Array.canBuildFrom
import algo.structures.DynamicGraph
import javax.swing.JPanel
import javax.swing.JScrollPane
import javax.swing.ScrollPaneConstants
import measures.GraphMeasures
import guidyn.components.listeners.GraphUpdateListener

object Measures extends JPanel(new BorderLayout) with GraphUpdateListener {
    val plotWidth = 300
    val nbVerticesEdgesPlot = new Plot("Nb Vertices", false, "Timestamp", "Nb Vertices", plotWidth, 3D/4, useIntRangeAxis = true)
    
    val inertiaWithinPlot = new Plot("Inertia", false, "Timestamp", "Inertia", plotWidth, 3D/4, 
            legend = List(("Within", Color.black), ("Between", Color.red)))
    
    val homophiliesPlot = new Plot("Homophily", false, "Timestamp", "Homophily", plotWidth, 3D/4, 
            legend = List(("Observed Homo.", Color.black), ("Expected. Homo.", Color.red)))
    
    val clusteringCoefficientPlot = new Plot("Clustering Coefficient", false, "Timestamp", "Clustering Coefficient", plotWidth, 3D/4, legend = List(("Clust. Coeff.", Color.black), ("Rand. Clust. Coeff.", Color.red)))
    
    val diameterPlot = new Plot("Diameter", false, "Timestamp", "Diameter", plotWidth, 3D/4, useIntRangeAxis = true)
    val avgDegreePlot = new Plot("Average Degree", false, "Timestamp", "Average Degree", plotWidth, 3D/4)
    val nbEdgesVerticesPlot = new Plot("Densification", true, "Nb Vertices", "Nb Edges", plotWidth, 3D/4, drawLine = false, useIntRangeAxis = true, drawDiagonalLine = true)
    val modularityPlot = new Plot("Modularity", false, "Timestamp", "Modularity", plotWidth, 3D/4)
    val nbCommunityPlot = new Plot("Nb Communities", false, "Timestamp", "Nb Communities", plotWidth, 3D/4, useIntRangeAxis = true)
    val avgShortestPathPlot = new Plot("Avg. Shortest Path", false, "Timestamp", "Avg. Shortest Path", plotWidth, 3D/4, useIntRangeAxis = true)
    val communitiesSizesPlot = new Plot("Community Size", false, "Timestamp", "Community Size", plotWidth, 3D/4, useIntRangeAxis = true)
    val componentsSizesPlot = new Plot("Component Size", false, "Timestamp", "Component Size", plotWidth, 3D/4, useIntRangeAxis = true)
    val nbComponentsPlot = new Plot("Nb. Components", false, "Timestamp", "Nb. Components", plotWidth, 3D/4, useIntRangeAxis = true)
    val edgesPlot = new Plot("Nb Edges", false, "Timestamp", "Nb Edges", plotWidth, 3D/4, 
            legend = List(("Edges Within", Color.black), ("Edges Between", Color.red)), useIntRangeAxis = true)
    
    val plotsContainer = new JPanel(new FlowLayout(FlowLayout.LEADING))
//    val plotsContainer = new JPanel()
//    plotsContainer.setLayout(new BoxLayout(plotsContainer, BoxLayout.Y_AXIS))
//    val plotsContainer = new JPanel(new GridLayout(2, 0))
    plotsContainer.add(nbVerticesEdgesPlot)
    plotsContainer.add(diameterPlot)
    plotsContainer.add(communitiesSizesPlot)
    plotsContainer.add(componentsSizesPlot)
    plotsContainer.add(clusteringCoefficientPlot)
    plotsContainer.add(avgDegreePlot)
    plotsContainer.add(modularityPlot)
    plotsContainer.add(nbCommunityPlot)
    plotsContainer.add(avgShortestPathPlot)
    plotsContainer.add(inertiaWithinPlot)
    plotsContainer.add(homophiliesPlot)
    plotsContainer.add(nbEdgesVerticesPlot)
    plotsContainer.add(edgesPlot)
    plotsContainer.add(nbComponentsPlot)
    
    plotsContainer.setBackground(Color.white)
     add(new JScrollPane(plotsContainer, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED), BorderLayout.CENTER)
//    add(plotsContainer, BorderLayout.CENTER)
    
    def graphUpdated(g: DynamicGraph) = {
        val timestampsValues = (for(t <- 0 until g.nbTimestamps) yield (t.toDouble + 1)).toArray
        val devs = Parameters.attributesParams.take(Parameters.nbAttributesField.value.toInt).map(_.value)
        val measures = for(graph <- g.graphs.toArray) yield { new GraphMeasures(graph, devs) }
        
        ProgressBar.startTask("Measures - Diameter")
        if(diameterPlot.plotAutoUpdate.isSelected()) {
            val diameters = measures.map(m => m.effectiveDiameter.toDouble)
    		diameterPlot.setDistribution(timestampsValues, diameters)
        } else {
            diameterPlot.setDistribution(Array(1D), Array(1D))
        }
        
        val nbVertices = measures.map(m => m.nbVertices.toDouble)
        val nbEdges = measures.map(m => m.nbEdges.toDouble)
        nbVerticesEdgesPlot.setDistribution(timestampsValues, nbVertices)
        nbVerticesEdgesPlot.setSecondRangeAxis("Nb Edges", timestampsValues, nbEdges, 2)
        
        nbEdgesVerticesPlot.setDistribution(nbVertices, nbEdges)
        
        ProgressBar.startTask("Measures - Clustering Coefficient")
        val clusteringCoefficients = measures.map(m => m.clusteringCoefficient)
        val randClusteringCoefficients = measures.map(m => m.randomClusteringCoefficient)
        clusteringCoefficientPlot.setDistribution(timestampsValues, clusteringCoefficients)
        clusteringCoefficientPlot.setDistribution(timestampsValues, randClusteringCoefficients, 2)
        
        ProgressBar.startTask("Measures - Average Degree")
        val avgDegrees = measures.map(m => m.avgDegree)
        avgDegreePlot.setDistribution(timestampsValues, avgDegrees)

        ProgressBar.startTask("Measures - Modularity")
        val modularities = measures.map(m => m.modularity)
        modularityPlot.setDistribution(timestampsValues, modularities)

        ProgressBar.startTask("Measures - Nb Communities")
        val nbCommunities = measures.map(m => m.nbCommunities.toDouble)
        nbCommunityPlot.setDistribution(timestampsValues, nbCommunities)
        
        ProgressBar.startTask("Measures - Inertia")
        val inertiaWithins = measures.map(m => m.withinInertiaRatio)
        val inertiaBetweens = measures.map(m => m.betweenInertiaRatio)
        inertiaWithinPlot.setDistribution(timestampsValues, inertiaWithins)
        inertiaWithinPlot.setDistribution(timestampsValues, inertiaBetweens, 2)
    
        ProgressBar.startTask("Measures - Homophily")
        val expectedHomophilies = measures.map(m => m.homophilyExpected)
        val observedHomophilies = measures.map(m => m.homophilyObserved)
        homophiliesPlot.setDistribution(timestampsValues, observedHomophilies)
        homophiliesPlot.setDistribution(timestampsValues, expectedHomophilies, 2)
    
        ProgressBar.startTask("Measures - Nb Edges")
        val edgesWithin = measures.map(m => m.nbWithinEdges.toDouble)
        val edgesBetween = measures.map(m => m.nbBetweenEdges.toDouble)
        edgesPlot.setDistribution(timestampsValues, edgesWithin)
        edgesPlot.setDistribution(timestampsValues, edgesBetween, 2)
//        val inertiaBetweens = measures.map(m => m.betweenInertiaRatio)
//        inertiaBetweenPlot.setDistribution(timestampsValues, inertiaBetweens)
//    
//        val inertiaBetweens = measures.map(m => m.betweenInertiaRatio)
//        inertiaBetweenPlot.setDistribution(timestampsValues, inertiaBetweens)
//    
//        val inertiaBetweens = measures.map(m => m.betweenInertiaRatio)
//        inertiaBetweenPlot.setDistribution(timestampsValues, inertiaBetweens)
    
        
        ProgressBar.startTask("Measures - Avg. Shortest Path")
        if(avgShortestPathPlot.plotAutoUpdate.isSelected()) {
            val avgShortestPaths = measures.map(m => m.avgShortestPath)
    		avgShortestPathPlot.setDistribution(timestampsValues, avgShortestPaths)
        } else {
            avgShortestPathPlot.setDistribution(Array(1D), Array(1D))
        }

        ProgressBar.startTask("Measures - Community Size")
        val communitiesSizesPlotFirst = measures.map(m => m.communitySizes(0).toDouble)
        val communitiesSizesPlotSecond = measures.map { m => if(m.communitySizes.length >= 2) m.communitySizes(1).toDouble else 0 }
        val communitiesSizesPlotThird = measures.map { m => if(m.communitySizes.length >= 3) m.communitySizes(2).toDouble else 0 }
        communitiesSizesPlot.setDistribution(timestampsValues, communitiesSizesPlotFirst)
        communitiesSizesPlot.setDistribution(timestampsValues, communitiesSizesPlotSecond, 2)
        communitiesSizesPlot.setDistribution(timestampsValues, communitiesSizesPlotThird, 3)
        
        ProgressBar.startTask("Measures - Nb. Connected components")
        val nbComponents = measures.map(m => m.connectedComponentsSizes.length.toDouble)
        nbComponentsPlot.setDistribution(timestampsValues, nbComponents)
        
        ProgressBar.startTask("Measures - Connected components Size")
        val componentSizesPlotFirst = measures.map(m => m.connectedComponentsSizes(0).toDouble)
        val componentSizesPlotSecond = measures.map { m => if(m.connectedComponentsSizes.length >= 2) m.connectedComponentsSizes(1).toDouble else 0 }
        val componentSizesPlotThird = measures.map { m => if(m.connectedComponentsSizes.length >= 3) m.connectedComponentsSizes(2).toDouble else 0 }
        componentsSizesPlot.setDistribution(timestampsValues, componentSizesPlotFirst)
        componentsSizesPlot.setDistribution(timestampsValues, componentSizesPlotSecond, 2)
        componentsSizesPlot.setDistribution(timestampsValues, componentSizesPlotThird, 3)
    }
}