/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package guidyn.components

import java.awt.BorderLayout
import java.io.File
import scala.Array.canBuildFrom
import com.alee.extended.label.WebHotkeyLabel
import com.alee.laf.button.WebButton
import com.alee.laf.combobox.WebComboBox
import algo.structures.DynamicGraph
import algo.GeneratorDyn
import algo.GeneratorParameters
import algo.MicroProba
import gui.components.RoundedSwitch
import gui.components.Savable
import gui.components.SavableManager
import gui.components.SliderDescriptor
import gui.components.SpinSlider
import gui.components.TitledSeparator
import guidyn.Main
import javax.swing.BoxLayout
import javax.swing.JButton
import javax.swing.JLabel
import javax.swing.JPanel
import javax.swing.JTextField
import utils.Icon
import utils.SwingEvents._
import utils.Timer
import javax.swing.JSpinner
import javax.swing.SpinnerNumberModel
import com.alee.laf.spinner.WebSpinner
import javax.swing.DefaultComboBoxModel
import java.util.Vector
import java.awt.FlowLayout
import com.alee.managers.popup.WebButtonPopup
import com.alee.managers.popup.PopupWay
import algo.structures.GraphBuilder
import com.alee.managers.notification.WebNotificationPopup
import com.alee.managers.notification.NotificationManager
import java.io.PrintWriter
import java.util.Date

object RunPanel extends JPanel(new BorderLayout()) with Runnable with Savable {
    val seedField = new JTextField(12)
    var runButton = new WebButton("Generate", Icon.get("resultset_next"))
    var runOptions = new WebButton("Options", Icon.get("cog"))
    val useSeed = new RoundedSwitch()
    useSeed.onAction { e => 
        seedField.setEditable(useSeed.isSelected())
        seedField.setEnabled(useSeed.isSelected())
    }
    
    val popup = new WebButtonPopup(runOptions, PopupWay.downRight)
    val runOptionsPanel = new JPanel(new VerticalLayout(0, VerticalLayout.BOTH, VerticalLayout.TOP))
    popup.setContent(runOptionsPanel)
    
    /**
     * Load graph panel
     */
    val isLoadGraphPanel = new JPanel(new FlowLayout(FlowLayout.LEFT))
    isLoadGraphPanel.add(new JLabel("Load existing graph"))
    val isLoadGraph = new RoundedSwitch()
    isLoadGraphPanel.add(isLoadGraph)
    isLoadGraph.setSelected(false)
    val timestampSpinnerModel = new SpinnerNumberModel(1, 1, 1, 1)
    val timestampSpinner = new WebSpinner(timestampSpinnerModel)
    timestampSpinner.setVisible(false)
    val timestampPanel = new JPanel(new FlowLayout(FlowLayout.LEFT))
    timestampPanel.add(new JLabel("Starting timestamp: "))
    timestampPanel.add(timestampSpinner)
    val comboBoxModel = new DefaultComboBoxModel(new java.util.Vector[String]())
    val datasetListComboBox = new WebComboBox(comboBoxModel)
    val datasetLoadPanel = new JPanel(new FlowLayout(FlowLayout.LEFT))
    datasetLoadPanel.add(new JLabel("Dataset:"))
    datasetLoadPanel.add(datasetListComboBox, BorderLayout.CENTER)
    
    runOptionsPanel.add(isLoadGraphPanel)
    runOptionsPanel.add(datasetLoadPanel)
    runOptionsPanel.add(timestampPanel)
    
    isLoadGraph.onAction { e =>
        val isSelected = isLoadGraph.isSelected()
        datasetLoadPanel.setVisible(isSelected)
        timestampSpinner.setVisible(isSelected)
        Parameters.setParametersEditable(isSelected)
    	if(isSelected) {
    	    comboBoxModel.removeAllElements()
    	    for(n <- new File("out/").listFiles().filter(_.isDirectory()).map(a => (a.getName(), a.lastModified())).sortBy(_._2).reverse.map(_._1)) { 
    	        comboBoxModel.addElement(n)
	        }
    	}
        popup.updateBounds()
    }
    
    datasetListComboBox.onAction { e =>
        if(datasetListComboBox.getItemCount() > 0) {
            val dataset = datasetListComboBox.getSelectedItem().toString
	        val parameters = GeneratorParameters.read(s"out/${dataset}/parameters")
	        Parameters.loadStaticFromParamters(parameters)
	        timestampSpinnerModel.setMaximum(parameters.nbTimestamps)
	        timestampSpinnerModel.setValue(1)
        }
    }

    
    val leftPanel = new JPanel(new FlowLayout(FlowLayout.LEFT))
    leftPanel.add(runOptions)
    leftPanel.add(runButton)
    leftPanel.add(new JLabel("Seed:"))
    leftPanel.add(seedField)
    leftPanel.add(useSeed)
    
    add(leftPanel, BorderLayout.WEST)
    add(new SavePanel(), BorderLayout.EAST)

    var curThread : Option[Thread] = None
    runButton.onAction { e =>
        val shouldStop = if(curThread.isDefined) {
            curThread.get.isAlive()
        } else { false }
        if(shouldStop) {
            runButton.setIcon(Icon.get("resultset_next"))
            runButton.setText("Run")
            ProgressBar.progressAction.setText("")
		    ProgressBar.timestamp.setText("Generation Stopped")
		    ProgressBar.progressBar.setIndeterminate(false)
		    ProgressBar.progressBar.setMaximum(1)
		    ProgressBar.progressBar.setValue(1)
            for(th <- curThread) { th.stop() }
        } else {
            runButton.setIcon(Icon.get("cross"))
            runButton.setText("Stop")
            curThread = Option(new Thread(this))
            curThread.get.start()
        }
    }
    
    /**
     * Run the generator
     */
    var currentGenerator : Option[GeneratorDyn] = None
    var currentGraph : Option[DynamicGraph] = None
    override def run()  = {
        val seed = if(useSeed.isSelected() && !seedField.getText().isEmpty()) Option(seedField.getText().toLong) else None
        val devs = Parameters.attributesParams.take(Parameters.nbAttributesField.value.toInt).map(_.value)
        val microProba = MicroProba(
                probaMicro = Parameters.pMicroField.value,
                addVertices = Parameters.pAddVertexField.value,
                removeVertices = Parameters.pRemoveVertexField.value,
                updateAttributes = Parameters.pUpdateAttributeField.value,
                addExternalEdges = Parameters.pAddBtwEdgeField.value,
                removeExternalEdges = Parameters.pRemoveBtwEdgeField.value,
                addInternalEdges = Parameters.pAddWthEdgeField.value,
                removeInternalEdges = Parameters.pRemoveWthEdgeField.value,
                probaMerge = Parameters.pMergeField.value,
                probaSplit = Parameters.pSplitField.value,
                probaChange = Parameters.pChangeField.value)
                
        ProgressBar.generateNewGraph(Parameters.nbTimestampsField.value.toInt)
        
        try {
            val gen = new GeneratorDyn(k = Parameters.kField.value.toInt, nbVertices = Parameters.nbVerticesField.value.toInt,
            dev = devs, theta = Parameters.thetaField.value,
            nbRep = Parameters.nbRepresententsField.value.toInt,
            avgWthDegree = Parameters.nbMaxInternalEdgesField.value.toInt, avgBtwDegree = Parameters.nbMaxExternalEdgesField.value.toInt,
            minimumTotalEdges = Parameters.minimalTotalEdgesField.value.toInt, microProba = microProba, nbTimestamps = Parameters.nbTimestampsField.value.toInt, 
            seed = seed, useAverageDegree = Parameters.averageDegree.isSelected(),
            /*probaCloseTriangle = Parameters.closingTriangleField.value.toDouble / 100*/0)
        
	        seedField.setText(gen.selectedSeed.toString)
	        val startingGraph = if(isLoadGraph.isSelected()) {
	            val graphPath = "out/" + datasetListComboBox.getSelectedItem().toString() + "/t" +  (timestampSpinnerModel.getValue().toString().toInt - 1) + ".graph"
	            Option(algo.structures.GraphBuilder.readGraph2(graphPath, Parameters.nbRepresententsField.value.toInt))
	        } else { None }
	        val dynamicGraph = gen.generateDynamicGraph(startingGraph)
	        ProgressBar.startTask("Updating the GUI...")
	        for(e <- Main.graphUpdateListeners) { e.graphUpdated(dynamicGraph) }
	        ProgressBar.allDone()
	        runButton.setIcon(Icon.get("resultset_next"))
	        runButton.setText("Generate")
	
	        currentGenerator = Option(gen)
	        currentGraph = Option(dynamicGraph)
        } catch {
            case e : Exception => 
                val errorNotification = new WebNotificationPopup()
                errorNotification.setContent("<html>Sorry, an exception has occured<br>" + e)
                errorNotification.setIcon(Icon.get("error_icon"))
                NotificationManager.showNotification (errorNotification)
                val errorLog = new PrintWriter("errorLog")
                errorLog.println(new Date())
                errorLog.println(e)
                errorLog.flush()
                errorLog.close()
                Parameters.getParameters.save("errorLog.parameters")
                e.printStackTrace()
        }
    }
    
    // Automatically load and save the parameters
    def load() = { 
        seedField.setText(SavableManager.load("seed", "1")) 
        useSeed.setSelected(SavableManager.load("useSeed", true))
        Parameters.averageDegree.setSelected(SavableManager.load("averageDegree", true))
    }
    def save() = {
        SavableManager.save("seed", seedField.getText())
        SavableManager.save("useSeed", useSeed.isSelected())
        SavableManager.save("averageDegree", Parameters.averageDegree.isSelected())
    }
}

