/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package guidyn.components.graph

import guidyn.components.CollapsablePanel
import com.alee.laf.label.WebLabel
import javax.swing.JCheckBox
import edu.uci.ics.jung.algorithms.layout.StaticLayout
import com.alee.laf.combobox.WebComboBox
import edu.uci.ics.jung.algorithms.util.IterativeContext
import com.alee.managers.tooltip.TooltipManager
import com.alee.laf.button.WebButton
import edu.uci.ics.jung.algorithms.layout.Layout
import edu.uci.ics.jung.graph.Graph
import edu.uci.ics.jung.algorithms.layout.FRLayout
import org.apache.commons.collections15.Transformer
import com.alee.managers.tooltip.TooltipWay
import java.awt.geom.Point2D
import utils.Icon
import utils.SwingEvents._
import edu.uci.ics.jung.algorithms.layout.KKLayout
import guidyn.layout.PlotLayout
import edu.uci.ics.jung.algorithms.layout.ISOMLayout
import gui.components.SavableManager
import java.awt.FlowLayout
import javax.swing.JPanel
import guidyn.components.listeners.GraphUpdateListener
import algo.structures.DynamicGraph
import edu.uci.ics.jung.algorithms.layout.FRLayout2
import guidyn.components.graph.layouts.PlotLayoutManager
import guidyn.components.listeners.TimestampUpdateListener
import gui.components.Savable

object GraphLayoutOption extends CollapsablePanel("Layout") with GraphUpdateListener with TimestampUpdateListener with Savable {
    
    // List of layouts
    val plotXY = GraphLayout("PlotXY", classOf[PlotLayout])
    val fruchtermanReingold = GraphLayout("Fruchterman-Reingold", classOf[FRLayout[Int, Int]])
    val kamadaKawai = GraphLayout("Kamada-Kawai", classOf[KKLayout[Int, Int]])
    val selfOrganizingMap = GraphLayout("Self Organizing Map", classOf[ISOMLayout[Int, Int]])
    val plotLayout = new PlotLayoutOption()
    val allLayouts = Array[GraphLayout](fruchtermanReingold, kamadaKawai, selfOrganizingMap, plotLayout)
    
    // Components
    val layouts = new WebComboBox(allLayouts.asInstanceOf[Array[Object]])
    layouts.onAction { e => updateLayout() }
    val stopLayout = new WebButton(Icon.get("cross"))
    TooltipManager.addTooltip(stopLayout, "Stop computing the layout", TooltipWay.down, 0)
    stopLayout.onAction { e => 
        val curLayout = getSelectedLayout()
        val transformer = new Transformer[Int,Point2D] { def transform(v : Int) = curLayout.transform(v) }
        val staticLayout = new StaticLayout(GraphView.jungGraph, transformer, curLayout.getSize())
        GraphView.viewer.setGraphLayout(staticLayout)
    }
    
    val layoutSelectionPanel = new JPanel(new FlowLayout(FlowLayout.LEFT))
    layoutSelectionPanel.add(new WebLabel("Layout:"))
    layoutSelectionPanel.add(layouts)
    layoutSelectionPanel.add(stopLayout)
    addComponent(layoutSelectionPanel)
    
    /**
     * Return the current layout
     */
    def getSelectedLayout() = {
        layouts.getSelectedItem() match {
            case x : GraphLayout => x.getInstance()
            case _ => plotXY.getInstance()
        }
    }
    
    /** 
     * Require to build a new layout when the graph has been updated
     */
    def graphUpdated(g : DynamicGraph) { for(l <- allLayouts) { l.setGraphUpdated } }
    
    def timestampUpdated(t : Int, g : DynamicGraph) = {
        if(layouts.getSelectedItem() == plotLayout) {
            GraphView.viewer.setGraphLayout(plotLayout.plotLayoutManager.getLayoutForTimestamp(t))
        }
    }
    
    /**
     * Update the graph layout
     */
//    var doComputeLayout = true
    def updateLayout() = {
        if(GraphView.jungGraph != null) { 
	        new Thread(new Runnable { def run() = { GraphView.viewer.setGraphLayout(getSelectedLayout()) } }).start
	        GraphView.repaint()
        }
    }
    
    override def save() {
        super.save()
        SavableManager.save("selectedLayoutIdx", layouts.getSelectedIndex())
    }
    override def load() {
        super.load()
        layouts.setSelectedIndex(SavableManager.load("selectedLayoutIdx", 0))
    }
}

case class GraphLayout(label : String, layoutClass : Class[_]) {
    var instance : Option[Layout[Int, Int]] = None
    def setGraphUpdated() { instance = None }
    def getInstance() : Layout[Int, Int] = {
    	instance = Option(instance.getOrElse {
    	    val constructor = layoutClass.getDeclaredConstructor(Array(classOf[Graph[Int, Int]]) : _*)
	        val l = constructor.newInstance(Array(GraphView.jungGraph) : _*).asInstanceOf[Layout[Int, Int]]
	        if(l.isInstanceOf[FRLayout[Int, Int]]) {
	            l.asInstanceOf[FRLayout[Int, Int]].setAttractionMultiplier(0.08)
	            l.asInstanceOf[FRLayout[Int, Int]].setRepulsionMultiplier(0.3)
	        }
    	    if(l.isInstanceOf[ISOMLayout[Int, Int]]) {
	        }
    	    
	        l
    	})
    	instance.get
    }
    override def toString() = label
}

class PlotLayoutOption extends GraphLayout("Plot Layout", classOf[PlotLayout]) {
    var plotLayoutManager = new PlotLayoutManager()
    override def getInstance() : Layout[Int, Int] = { plotLayoutManager.getLayoutForTimestamp() }
}
