/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package guidyn.components.graph

import java.awt.BasicStroke
import java.awt.BorderLayout
import java.awt.Color
import java.awt.Dimension
import java.awt.Paint
import java.awt.Stroke
import java.awt.event.InputEvent
import java.awt.event.MouseAdapter
import java.awt.event.MouseEvent
import java.awt.geom.Ellipse2D
import scala.collection.JavaConversions._
import scala.collection.mutable.HashMap
import scala.collection.mutable.HashSet
import org.apache.commons.collections15.Predicate
import org.apache.commons.collections15.Transformer
import algo.structures.DynamicGraph
import edu.uci.ics.jung.algorithms.layout.Layout
import edu.uci.ics.jung.graph.Graph
import edu.uci.ics.jung.graph.UndirectedSparseGraph
import edu.uci.ics.jung.graph.util.Context
import edu.uci.ics.jung.visualization.GraphZoomScrollPane
import edu.uci.ics.jung.visualization.Layer
import edu.uci.ics.jung.visualization.RenderContext
import edu.uci.ics.jung.visualization.VisualizationViewer
import edu.uci.ics.jung.visualization.control.DefaultModalGraphMouse
import edu.uci.ics.jung.visualization.control.ModalGraphMouse
import edu.uci.ics.jung.visualization.control.RotatingGraphMousePlugin
import edu.uci.ics.jung.visualization.control.ShearingGraphMousePlugin
import edu.uci.ics.jung.visualization.control.TranslatingGraphMousePlugin
import edu.uci.ics.jung.visualization.decorators.EdgeShape
import edu.uci.ics.jung.visualization.renderers.Renderer
import guidyn.components.listeners.GraphUpdateListener
import javax.swing.JPanel
import java.awt.Shape
import edu.uci.ics.jung.visualization.VisualizationServer
import java.awt.Graphics
import edu.uci.ics.jung.visualization.decorators.ToStringLabeller
import javax.swing.ToolTipManager
import utils.Stat
import utils.PdfOutput
import edu.uci.ics.jung.visualization.control.GraphMouseListener

object GraphView extends JPanel(new BorderLayout()) with GraphUpdateListener {
    var viewer: VisualizationViewer[Int, Int] = _
	var jungGraph : UndirectedSparseGraph[Int, Int] = _
	var communitiesGraph = new Array[UndirectedSparseGraph[Int, Int]](0)
	setBackground(Color.white)
	
	var curTimestamp = 0
	var verticesAtTimestamp = HashMap[Int, HashSet[Int]]()
	var edgeAtTimestamp = HashMap[Int, HashSet[Int]]()
	var timestampToVertexToColor = HashMap[Int, HashMap[Int, Color]]()
	
	def lastTimestamps = verticesAtTimestamp.keySet.max
	var curDynamicGraph : Option[DynamicGraph] = None
	
	def graphUpdated(g: DynamicGraph) = {
        // Build the jung graph
        curDynamicGraph = Option(g)
        curTimestamp = 0
        var edgeId = 1
        jungGraph = new UndirectedSparseGraph[Int, Int]()
        verticesAtTimestamp.clear
        edgeAtTimestamp.clear
        for((graph, t) <- g.graphs.zipWithIndex) {
            verticesAtTimestamp(t) = HashSet[Int]()
            edgeAtTimestamp(t) = HashSet[Int]()
            timestampToVertexToColor(t) = HashMap[Int, Color]()
            for(v <- graph.vertices) {
                jungGraph.addVertex(v)
                verticesAtTimestamp(t).add(v)
                timestampToVertexToColor(t)(v) = g.communityColors(graph.vertexCommunities(v))
                
                for(v2 <- graph(v)) {
                    val newEdge = jungGraph.findEdge(v, v2)
                    if(newEdge == 0) {
                        jungGraph.addEdge(edgeId, v, v2)
                        edgeAtTimestamp(t).add(edgeId)
                        edgeId += 1
                    } else {
                        edgeAtTimestamp(t).add(newEdge)
                    }
                }
            }
        }
        
        val vertexPaint = new Transformer[Int, Paint]() {
            def transform(v: Int): Paint = {
                if (viewer.getPickedVertexState().getPicked().contains(v)) { Color.red } else {
                    timestampToVertexToColor(curTimestamp)(v)
                }
            }
        }
        val edgePaint = new Transformer[Int, Paint]() {
            val noColor = new Color(0,0,0,0f)
            def transform(edge: Int): Paint = {
                if(edgeAtTimestamp(curTimestamp).contains(edge)) Color.black else noColor
            }
        }
        
        val gm = new DefaultModalGraphMouse[Int, Int]()
        gm.setMode(ModalGraphMouse.Mode.PICKING)
        val translatingPlugin = new TranslatingGraphMousePlugin(InputEvent.BUTTON3_MASK)

        val rotatingPlugin = new RotatingGraphMousePlugin()
        val shearingPlugin = new ShearingGraphMousePlugin()
        gm.add(rotatingPlugin)
        gm.add(shearingPlugin)
        gm.add(translatingPlugin)

        val dim = new Dimension(getWidth(), getHeight())
        viewer = new VisualizationViewer(GraphLayoutOption.getSelectedLayout(), dim)
        viewer.setGraphMouse(gm)
        viewer.setBackground(Color.white)
        viewer.getRenderContext().setVertexFillPaintTransformer(vertexPaint)
//        viewer.getRenderContext().setEdgeDrawPaintTransformer(edgePaint)
        viewer.setVertexToolTipTransformer(new VertexTooltip())
        viewer.setEdgeToolTipTransformer(new EdgeTooltip())
        viewer.getRenderContext().setEdgeShapeTransformer(new EdgeShape.Line())
        viewer.getRenderContext().setVertexIncludePredicate(new VertexDisplayPredicate())
        viewer.getRenderContext().setEdgeIncludePredicate(new EdgeDisplayPredicate())
        viewer.getRenderContext().setEdgeStrokeTransformer(new EdgeStrokeTransformer)
        viewer.getRenderer().setVertexRenderer(new VertexRenderer())
        viewer.getRenderContext().setVertexShapeTransformer(new Transformer[Int,Shape]() {
            def transform(v : Int) : Shape = {
                val diameter = VertexSizeOption.getVertexSize(v)
                new Ellipse2D.Double(-diameter / 2, -diameter / 2, diameter, diameter)
            }
        })
        
        viewer.addPostRenderPaintable(new VisualizationServer.Paintable { 
            def paint(g : Graphics ) {
                
                if(GraphViewOption.displayLegend.isSelected()) {
                    val fm = g.getFontMetrics()
                    val width = (for(property <- VertexEvents.allProperties) yield { fm.stringWidth(property.name) }).max + 35
                    val height = fm.getHeight() * VertexEvents.allProperties.length + 15
                    g.setColor(Color.white)
                    g.fillRoundRect(10, 10, width, height, 10, 10)
                    g.setColor(Color.black)
                    g.drawRoundRect(10, 10, width, height, 10, 10)
                    var curY = 20
                    for(property <- VertexEvents.allProperties) {
                        g.setColor(property.color)
                        g.fillRect(20, curY, 10, 10)
                        g.setColor(Color.black)
                        g.drawRect(20, curY, 10, 10)
                        g.drawString(property.name, 35, curY + 10)
                        curY += fm.getHeight()
                    }
                }
                
            }
            
            def useTransform() = false
        })
        
        viewer.addGraphMouseListener(new GraphMouseListener[Int]() {
            def graphClicked(v : Int, e: MouseEvent) {
                if (e.getButton() == MouseEvent.BUTTON1 && e.getClickCount() >= 2) {
                    val curGraph = g.graphs(curTimestamp)
                    val curCom = curGraph.vertexCommunities(v)
                    for (v2 <- curGraph.partition(curCom)) { viewer.getPickedVertexState().pick(v2, true) }
                }
                if (e.getButton() == MouseEvent.BUTTON3 && e.getClickCount() >= 2) {
                    val curGraph = g.graphs(curTimestamp)
                    viewer.getPickedVertexState().pick(v, true)
                    for (v2 <- curGraph(v)) { viewer.getPickedVertexState().pick(v2, true) }
                }
            }
            def graphPressed(v : Int, e: MouseEvent) { }
            def graphReleased(v : Int, e: MouseEvent) { }
        })
        ToolTipManager.sharedInstance().setInitialDelay(0)
        val zoomScroll = new GraphZoomScrollPane(viewer)
        removeAll()
        add(zoomScroll, BorderLayout.CENTER)
        repaint()
    }
    
    val hiddenVertices = HashSet[Int]()
    def setHiddenVertices(f : Int => Boolean) = {
        hiddenVertices.clear
        for(v <- jungGraph.getVertices(); if(f(v))) { hiddenVertices.add(v) }
        repaint()
    }
    
    def setSelectedVertices(f : Int => Boolean) = {
        val pickedVertexState = viewer.getPickedVertexState()
        for(v <- jungGraph.getVertices()) { pickedVertexState.pick(v, f(v)) }
        repaint()
    }
    
    class VertexTooltip extends Transformer[Int, String] {
        def transform(vertex : Int) = {
            val curGraph = curDynamicGraph.get.graphs(curTimestamp)
            val attributeValues = (for((attr, i) <- curGraph.attr(vertex).zipWithIndex) yield { f"Attr ${i + 1}: ${attr}%.3f" }).toList
            (List(
                "<b>Id: " + vertex + "</b>", 
                "Degree: " + curGraph.degree(vertex),
                "Wth. Degree: " + curGraph.withinDegree(vertex),
                "Btw. Degree: " + curGraph.betweenDegree(vertex)) ::: attributeValues).mkString("<html>", "<br>", "</html>")
        }
    }
    
    class EdgeTooltip extends Transformer[Int, String] {
        def transform(edgeId : Int) = {
            val v1 = GraphView.jungGraph.getEndpoints(edgeId).getFirst()
		    val v2 = GraphView.jungGraph.getEndpoints(edgeId).getSecond()
            val curGraph = curDynamicGraph.get.graphs(curTimestamp)
            val distance = curGraph.distance(v1, v2)
            val kind = if(curGraph.sameCommunity(v1, v2)) { "Within edge" } else { "Between edge" }
            List(
                "<b>" + kind + "</b>", 
                "From " + v1 + " to " + v2,
                f"Distance: ${distance}%.3f").mkString("<html>", "<br>", "</html>")
        }
    }
    
    class VertexDisplayPredicate[V, E] extends Predicate[Context[Graph[V, E], V]] {
		def evaluate(context: Context[Graph[V, E], V]) : Boolean = {
			val curVertex = context.element.asInstanceOf[Int]
			verticesAtTimestamp(curTimestamp).contains(curVertex) && !hiddenVertices.contains(curVertex)
//			if(verticesAtTimestamp(curTimestamp).contains(curVertex)) {
//			    !hiddenVertices.contains(curVertex)
//			    SelectVerticesOption.isVertexInSelection(curVertex, curDynamicGraph.get.graphs(curTimestamp))
//			} else { false }
		}
	}
    
    class EdgeDisplayPredicate[V, E] extends Predicate[Context[Graph[V, E], E]] {
		def evaluate(context: Context[Graph[V, E], E]) : Boolean = {
		    val edgeId = context.element.asInstanceOf[Int]
		    if(edgeAtTimestamp(curTimestamp).contains(edgeId)) {
			    val v1 = GraphView.jungGraph.getEndpoints(edgeId).getFirst()
			    val v2 = GraphView.jungGraph.getEndpoints(edgeId).getSecond()
			    val curGraph = curDynamicGraph.get.graphs(curTimestamp)
			    var isBetweenEdge = curGraph.vertexCommunities(v1) != curGraph.vertexCommunities(v2)
			    val displayAllEdges = GraphViewOption.displayAllEdges.isSelected()
			    val displayNoEdges = GraphViewOption.displayNoEdges.isSelected()
			    val displayWithinEdges = GraphViewOption.displayWithinEdges.isSelected()
			    val displayBetweenEdges = GraphViewOption.displayBetweenEdges.isSelected()
		        displayAllEdges || (displayWithinEdges && !isBetweenEdge) || (displayBetweenEdges && isBetweenEdge)
		    } else { false }
		        
//		    val displayEdges = GraphViewOption.displayAllEdges.isSelected()
//		    displayEdges && edgeAtTimestamp(curTimestamp).contains(context.element.asInstanceOf[Int])
		}
	}
    
    class EdgeStrokeTransformer[V, E] extends Transformer[E, Stroke] {
        val stroke = new BasicStroke(2f)
		def transform(edge : E) = {
            stroke
        }
	}
    
    class VertexRenderer extends Renderer.Vertex[Int, Int] {
        def paintVertex(rc : RenderContext[Int, Int], layout : Layout[Int, Int] , vertex : Int) = {
            val curGraph = curDynamicGraph.get.graphs(curTimestamp)
            val displayVertex = verticesAtTimestamp(curTimestamp).contains(vertex) && !hiddenVertices.contains(vertex)
//                if(verticesAtTimestamp(curTimestamp).contains(vertex)) {
//			    SelectVerticesOption.isVertexInSelection(vertex, curGraph)
//			} else { false }
            
		    if(displayVertex) {
		        val graphicsContext = rc.getGraphicsContext()
        		val center = rc.getMultiLayerTransformer().transform(Layer.LAYOUT, layout.transform(vertex))
        		val (cx, cy) = (center.getX(), center.getY())
        		val diameter = VertexSizeOption.getVertexSize(vertex)
		        if(GraphViewOption.displayLegend.isSelected()) {
		            val events = curGraph.vertexEvents(vertex)
		            var eventCircleDiameter = diameter + 6 * events.size
		            for(e <- events) {
		                graphicsContext.setPaint(e.color)
		        		graphicsContext.fill(new Ellipse2D.Double(cx-eventCircleDiameter / 2, cy-eventCircleDiameter / 2, eventCircleDiameter, eventCircleDiameter))
				        graphicsContext.setPaint(Color.black)
		        		graphicsContext.draw(new Ellipse2D.Double(cx-eventCircleDiameter / 2, cy-eventCircleDiameter / 2, eventCircleDiameter, eventCircleDiameter))
		                eventCircleDiameter -= 6
		            }
		        } 
        		
        		val paint = if (viewer.getPickedVertexState().getPicked().contains(vertex)) { Color.red } else { timestampToVertexToColor(curTimestamp)(vertex) }
		        graphicsContext.setPaint(paint)
        		graphicsContext.fill(new Ellipse2D.Double(cx-diameter / 2, cy-diameter / 2, diameter, diameter))
        		graphicsContext.setPaint(Color.black)
        		graphicsContext.draw(new Ellipse2D.Double(cx-diameter / 2, cy-diameter / 2, diameter, diameter))
		    }
        }
	}
    
    def saveAsPdf(path : String) = {
        val prevTimestamp = curTimestamp
        for(t <- 0 to lastTimestamps) {
            curTimestamp = t
            PdfOutput.createPdf(viewer, s"out/${path}/graph_${t}.pdf", Option(viewer.getSize()))
        }
        curTimestamp = prevTimestamp
    }
}