/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package guidyn.components.graph

import guidyn.components.CollapsablePanel
import java.awt.BorderLayout
import javax.swing.JPanel
import com.alee.extended.label.WebHotkeyLabel
import gui.components.SpinSlider
import com.alee.laf.combobox.WebComboBox
import guidyn.components.listeners.TimestampUpdateListener
import algo.structures.Graph
import utils.SwingEvents._
import guidyn.components.listeners.GraphUpdateListener
import algo.structures.DynamicGraph
import javax.swing.JLabel
import javax.swing.JComboBox
import javax.swing.JCheckBox
import scala.collection.mutable.HashMap


object VertexSizeOption extends CollapsablePanel("Vertex Size") with TimestampUpdateListener with GraphUpdateListener {
    // List of measures to consider
    val degreeVertexMeasure = new VertexMeasure("Degree", (v, g, dg) => g.degree(v))
	val constantVertexMeasure = new VertexMeasure("Constant", (v, g, dg) => 1D)
	val wthDegreeVertexMeasure = new VertexMeasure("Wth. Degree", (v, g, dg) => g.withinDegree(v))
	val btwDegreeVertexMeasure = new VertexMeasure("Btw. Degree", (v, g, dg) => g.betweenDegree(v))
    //Add a mesure for community (or community size)
    val comDegreeVertexMeasure = new VertexMeasure("Community", (v, g, dg) => g.vertexCommunities(v))
	val ageMeasure = new VertexMeasure("Age", (v, g, dg) => {
	    var age = 0
	    var timestamp = curTimestamp
	    while(timestamp != -1 && dg.graphs(timestamp).vertices.contains(v)) {
        	timestamp -= 1
        	age += 1
    	}
	    age.toDouble
	})
	val vertexMeasures = Array[VertexMeasure](constantVertexMeasure, degreeVertexMeasure, wthDegreeVertexMeasure, btwDegreeVertexMeasure, ageMeasure, comDegreeVertexMeasure)

    var curTimestamp = 0
    var curGraph : Option[Graph] = None
    var curDynamicGraphGraph : Option[DynamicGraph] = None
    def timestampUpdated(timestamp : Int, g : DynamicGraph) {
        curTimestamp = timestamp
        curDynamicGraphGraph = Option(g)
        curGraph = Option(g.graphs(curTimestamp))
        updateMinMaxVertexMeasure()
    }
    def graphUpdated(g : DynamicGraph) = timestampUpdated(0, g)
    
    // Store the minimum and maximum of the measures
    var minVertexMeasure = Double.MaxValue
    var maxVertexMeasure = Double.MinValue
    def updateMinMaxVertexMeasure() {
        minVertexMeasure = Double.MaxValue
        maxVertexMeasure = Double.MinValue
        curVertexMeasure = relativeToComponent.getSelectedItem().asInstanceOf[VertexMeasure]
        for(g <- curGraph; v <- g.vertices) {
            val m = curVertexMeasure.f(v, g, curDynamicGraphGraph.get)
            minVertexMeasure = math.min(minVertexMeasure, m)
            maxVertexMeasure = math.max(maxVertexMeasure, m)
        }
    }
    
    def getVertexSize(v : Int) : Int = {
        if(curGraph.isDefined) {
            val m = curVertexMeasure.f(v, curGraph.get, curDynamicGraphGraph.get)
            if(maxVertexMeasure - minVertexMeasure == 0) { minVertexSize.value.toInt }
            else { (minVertexSize.value + (maxVertexSize.value - minVertexSize.value) * (m - minVertexMeasure) / (maxVertexMeasure - minVertexMeasure)).toInt }
        } else { 10 }
    }
    
    // Components
	val minVertexSize = new SpinSlider("Min Size: ", 0, 100, 1, 1, help = "Minimum size of the vertices")
	minVertexSize.onChange { e =>
	    if(minVertexSize.value > maxVertexSize.value) { maxVertexSize.setValue(minVertexSize.value) }
	    GraphView.repaint()
	}
	val maxVertexSize = new SpinSlider("Max Size: ", 0, 100, 1, 1, help = "Maximum size of the vertices")
	maxVertexSize.onChange { e =>
	    if(maxVertexSize.value < minVertexSize.value) { minVertexSize.setValue(maxVertexSize.value) }
	    GraphView.repaint()
	}
	val relativeToComponent = new WebComboBox(vertexMeasures.asInstanceOf[Array[Object]])
	var curVertexMeasure = relativeToComponent.getSelectedItem().asInstanceOf[VertexMeasure]
	relativeToComponent.onAction { e => updateMinMaxVertexMeasure(); GraphView.repaint() }
	
	addComponent(relativeToComponent)
	addComponent(minVertexSize)
	addComponent(maxVertexSize)
}

class VertexMeasure(name : String, val f : (Int, Graph, DynamicGraph) => Double) {
    override def toString() = name
}
