/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package layout.yifanHu

import utils.Timer

class YifanHuLayout(var optimalDistance : Float, var relativeStrength : Float, var initialStep : Float, 
        var stepRatio : Float, var quadTreeMaxLevel : Int, var barnesHutTheta : Float, var convergenceThreshold : Float, var adaptiveCooling : Boolean,
        var displacement : Displacement) {
    var progress = 0f 
    var step = 0f
    var energy0 = 0D
    var energy = 0D
    var isConverged = false
    
    def postAlgo() {
        updateStep()
        if (Math.abs((energy - energy0) / energy) < convergenceThreshold) {
            isConverged = true
        }
    }
    
    def getDisplacement() : Displacement = {
        displacement.setStep(step)
        displacement
    }
    
    def updateStep() {
        if (adaptiveCooling) {
            if (energy < energy0) {
                progress += 1
                if (progress >= 5) {
                    progress = 0;
                    step /= stepRatio
                }
            } else {
                progress = 0;
                step *= stepRatio
            }
        } else {
            step *= stepRatio
        }
    }
    
    
    def initAlgo(graph : Graph) {
        energy = Float.MaxValue
//        for (Node n : graph.getNodes()) {
//            n.setLayoutData(new ForceVector());
//        }
        progress = 0;
        isConverged = false
        step = initialStep
    }

    def goAlgo(graph : Graph) {
    	for (n <- graph.nodes; if (n.layoutData == null)) {
            n.layoutData = new ForceVector();
        }

        // Evaluates n^2 inter node forces using BarnesHut.
    	Timer.startTimer("build quad tree")
    	val tree = QuadTreeBuilder.buildTree(graph, quadTreeMaxLevel)
    	Timer.stopTimer("build quad tree")
        val barnes = new BarnesHut(new ElectricalForce(relativeStrength, optimalDistance), barnesHutTheta)
    	
    	Timer.startTimer("init nodes")
        for (node <- graph.nodes) {
            val layoutData = node.layoutData
            val f = barnes.calculateForce(node, tree)
            layoutData.add(f)
        }
    	Timer.stopTimer("init nodes")

    	Timer.startTimer("init edges")
        for (e <- graph.edges) {
            if (!e.source.equals(e.target)) {
                val n1 = e.source
                val n2 = e.target
                val f1 = n1.layoutData
                val f2 = n2.layoutData
                val f = new SpringForce(optimalDistance).calculateForce(n1, n2);
                f1.add(f)
                f2.subtract(f)
            }
        }
    	Timer.stopTimer("init edges")

        // Calculate energy and max force.
    	Timer.startTimer("energy and max force")
        energy0 = energy;
        energy = 0;
        var maxForce = 1D
        for (n <- graph.nodes) {
            val force = n.layoutData
            energy += force.getNorm();
            maxForce = Math.max(maxForce, force.getNorm());
        }
        Timer.stopTimer("energy and max force")

        // Apply displacements on nodes.
        Timer.startTimer("displacement")
        for (n <- graph.nodes) {
            val force = n.layoutData
            force.multiply((1.0 / maxForce).toFloat);
            getDisplacement().moveNode(n, force);
        }
        Timer.stopTimer("displacement")
        postAlgo();
    }

    
    def resetPropertiesValues(graph : Graph) {
        stepRatio = 0.95f
        relativeStrength = 0.2f
        if (graph != null) {
            optimalDistance = Math.pow(relativeStrength, 1f / 3).toFloat * getAverageEdgeLength(graph)
        } else {
            optimalDistance = 100
        }

        initialStep = optimalDistance / 5
        step = initialStep
        quadTreeMaxLevel = 10
        barnesHutTheta = 1.2f
        adaptiveCooling = true
        convergenceThreshold = 1e-4f
    }

    def getAverageEdgeLength(graph : Graph) : Float = {
        var edgeLength = 0f;
        var count = 1;
        for (e <- graph.edges) {
            edgeLength += Math.hypot(e.source.x - e.target.x, e.source.y - e.target.y).toFloat
            count += 1
        }
        edgeLength / count;
    }

//    protected void setConverged(boolean isConverged) {
//    	this.isConverged = isConverged;
//    }
    
//    public YifanHuLayout(Displacement displacement) {
//        this.displacement = displacement;
//    }
    
//    public boolean getConverged() {
//        return isConverged;
//    }
    

    /*
    def getEdgeForce() : AbstractForce = {
        new SpringForce(getOptimalDistance());
    }

    def getNodeForce() : AbstractForce = {
        new ElectricalForce(getRelativeStrength(), getOptimalDistance());
        
    }
    */

    /*
    /* Maximum level for Barnes-Hut's quadtree */
    public Integer getQuadTreeMaxLevel() {
        return quadTreeMaxLevel;
    }

    public void setQuadTreeMaxLevel(Integer quadTreeMaxLevel) {
        this.quadTreeMaxLevel = quadTreeMaxLevel;
    }

    /* theta is the parameter for Barnes-Hut opening criteria */
    public Float getBarnesHutTheta() {
        return barnesHutTheta;
    }

    public void setBarnesHutTheta(Float barnesHutTheta) {
        this.barnesHutTheta = barnesHutTheta;
    }

    /**
     * @return the optimalDistance
     */
    public Float getOptimalDistance() {
        return optimalDistance;
    }

    /**
     * @param optimalDistance the optimalDistance to set
     */
    public void setOptimalDistance(Float optimalDistance) {
        this.optimalDistance = optimalDistance;
    }

    /**
     * @return the relativeStrength
     */
    public Float getRelativeStrength() {
        return relativeStrength;
    }

    /**
     * @param relativeStrength the relativeStrength to set
     */
    public void setRelativeStrength(Float relativeStrength) {
        this.relativeStrength = relativeStrength;
    }

    /**
     * @param step the step to set
     */
    public void setStep(Float step) {
        this.step = step;
    }

    /**
     * @return the adaptiveCooling
     */
    public Boolean isAdaptiveCooling() {
        return adaptiveCooling;
    }

    /**
     * @param adaptiveCooling the adaptiveCooling to set
     */
    public void setAdaptiveCooling(Boolean adaptiveCooling) {
        this.adaptiveCooling = adaptiveCooling;
    }

    /**
     * @return the stepRatio
     */
    public Float getStepRatio() {
        return stepRatio;
    }

    /**
     * @param stepRatio the stepRatio to set
     */
    public void setStepRatio(Float stepRatio) {
        this.stepRatio = stepRatio;
    }

    /**
     * @return the convergenceThreshold
     */
    public Float getConvergenceThreshold() {
        return convergenceThreshold;
    }

    /**
     * @param convergenceThreshold the convergenceThreshold to set
     */
    public void setConvergenceThreshold(Float convergenceThreshold) {
        this.convergenceThreshold = convergenceThreshold;
    }

    /**
     * @return the initialStep
     */
    public Float getInitialStep() {
        return initialStep;
    }

    /**
     * @param initialStep the initialStep to set
     */
    public void setInitialStep(Float initialStep) {
        this.initialStep = initialStep;
    }
	*/
}

class ElectricalForce(relativeStrength : Float, optimalDistance : Float) extends AbstractForce {
    def calculateForce(n1 : Node , n2 : Node , distance : Float) : ForceVector = {
        val f = new ForceVector(n2.x - n1.x, n2.y - n1.y)
        var scale = -relativeStrength * optimalDistance * optimalDistance / (distance * distance);
        if (scale.isNaN() || scale.isInfinite()) { scale = -1; }
        f.multiply(scale);
        f
    }
}

class SpringForce(optimalDistance : Float) extends AbstractForce {
    def calculateForce(n1 : Node , n2 : Node , distance : Float) : ForceVector = {
        val f = new ForceVector(n2.x - n1.x, n2.y - n1.y)
        f.multiply(distance / this.optimalDistance);
        f
    }
}
