/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package measures;

import scala.Array.canBuildFrom
import scala.collection.mutable.HashMap
import scala.collection.mutable.HashSet
import scala.collection.mutable.Stack
import utils.Stat
import algo.Generator1
import utils.Timer
import java.util.Arrays
import scala.collection.mutable.BitSet
import algo.structures.Graph
import utils.MBitSet

class GraphMeasures(val g : Graph, dev : Array[Double] = Array(1D)) {
    Timer.startTimer("Init Measures")
    val graphT = g.graphTab
    val attr = g.attr
    val vertexCommunities = g.vertexCommunities
    val nbVertices = g.nbVertices

    lazy val nbEdges = g.vertices.toArray.map(v => graphT(v).size).sum.toDouble / 2
    lazy val avgDegree: Double = nbEdges * 2 / nbVertices
    lazy val maxDegree: Double = g.vertices.map(v => graphT(v).size).max
    lazy val density = nbEdges / ((nbVertices * (nbVertices - 1)) / 2)

    val nbCommunities = g.partition.size
    val communities = g.partition.values
    val dThresholds = dev.map ( d => (2 * d.toDouble / nbCommunities))
    
    val communitySizes = g.partition.values.map(_.size).toArray.sorted.reverse
    
    Timer.stopTimer("Init Measures")

    /**
     * Compute a matrix having all pair of shortest path length
     */
    lazy val shortestPathFloydWarshall = {
        Timer.startTimer("Measure - Shortest Path FloydWarshall")
        val dist : Array[Array[Byte]] = Array.fill(nbVertices, nbVertices)(Byte.MaxValue)
        for (v1 <- g.vertices) { dist(v1)(v1) = 0 }
        for (v1 <- g.vertices; v2 <- graphT(v1)) { dist(v1)(v2) = 1; dist(v2)(v1) = 1 }
        for (k <- g.vertices; i <- g.vertices; j <- g.vertices) {
            if (dist(i)(k) != Byte.MaxValue && dist(k)(j) != Byte.MaxValue && dist(i)(k) + dist(k)(j) < dist(i)(j)) {
                dist(i)(j) = (dist(i)(k).toInt + dist(k)(j).toInt).toByte
            }
        }
        Timer.stopTimer("Measure - Shortest Path FloydWarshall")
        dist
    }

    val mappings = HashMap[Int, Int]()
    def shortestPathFast(v1 : Int)(v2 : Int) = shortestPathFastCompute(mappings(v1))(mappings(v2))
    var dist : Array[Array[Byte]] = Array.fill(0, 0)(0)
    
    /**
     * This algorithm is highly optimized and much faster than usual implementation for large graphs
     * It requires slightly more memory and give a valid result up to graph with a diameter of 126 (but is there graphs with larger diameter ?
     * The return result has been tested against Gephi algorithm
     */
    lazy val shortestPathFastCompute = {
        Timer.startTimer("Measure - Shortest Path Fast")
        // Build a mapping between the vertex ids and an array id
//        Timer.startTimer("Build array graph")
        val graph2 = new Array[Array[Int]](g.nbVertices)
        for((v1, i) <- g.vertices.zipWithIndex) { mappings(v1) = i }
        for((v1, i) <- g.vertices.zipWithIndex) {
            graph2(i) = new Array[Int](g(v1).size)
            for((n, i2) <- g(v1).zipWithIndex) { graph2(i)(i2) = mappings(n) }
        }
        
        var cand = Array.fill(2 * nbVertices)(0)
        var nbCand = 0
        var v = 0
        dist = Array.fill(nbVertices, nbVertices)(0)
        
        // Main loop
        while(v != nbVertices) {
            val remainingVertices = new MBitSet(nbVertices)
            for(i <- 0 until nbVertices) { remainingVertices.add(i) }
            
            // Build the initial set of candidates and remaining vertices
            for(x <- graph2(v)) {
                cand(nbCand) = x
                remainingVertices.remove(x)
                nbCand += 1
            }
	        remainingVertices.remove(v)
        	
	        var curDistance : Byte = 1
	        while(nbCand != 0 && nbCand != nbVertices) {
	            var (newNbCand, i) = if(nbCand > nbVertices) (0, nbVertices) else (nbVertices, 0)
	            while(i != nbCand) {
	                val v2 = cand(i)
	                dist(v)(v2) = curDistance
	                dist(v2)(v) = curDistance
	                var j = 0
	                val neighbors = graph2(v2)
	                val nbNeighbors = neighbors.length
	                while(j != nbNeighbors) {
	                    val n = neighbors(j)
	                    if(remainingVertices.remove(n)) {
	                        cand(newNbCand) = n
                    		newNbCand += 1
	                    }
	                    j +=1
	                }
	                i += 1
	            }
	            nbCand = newNbCand
	            curDistance = (curDistance.toInt + 1).toByte
	        }
	        // Update distance for not reachable vertices
	        for(v2 <- remainingVertices) {
	            dist(v2)(v) = Byte.MaxValue
                dist(v)(v2) = Byte.MaxValue
            }
	        dist(v)(v) = 0
            v += 1
        }
        Timer.stopTimer("Measure - Shortest Path Fast")
        dist
    }
    
    def freeShortestPaths() = {
        dist = Array.fill(0, 0)(0)
    }
    
    def freeMemory() = {
//        for((_, v) <- shortestPathFast()) { v.clear }
//        shortestPathFast.clear
    }
    
    /**
     * The diameter of the graph, i.e., the largest shortest path between each pair of vertices
     */
    lazy val diameter: Int = {
        Timer.startTimer("Measure - Diameter")
        /*
        var maxDistances = 0
        for (i <- graph.keys; j <- graph.keys) {
            val d = if(shortestPathFast(i)(j) == Byte.MaxValue) 0 else shortestPathFast(i)(j)
            maxDistances = math.max(maxDistances, d)
        }
        */
        
        var maxDistances = 0
        val d = shortestPathFastCompute
        var i = 0
        var j = 0
        while(i != g.nbVertices) {
            var j = i + 1
            while(j != g.nbVertices) {
                if(d(i)(j) != Byte.MaxValue) { maxDistances = math.max(d(i)(j), maxDistances) }
                j += 1
            }
            i += 1
        }
        Timer.stopTimer("Measure - Diameter")
        maxDistances
    }
    
    lazy val effectiveDiameter : Int = {
        Timer.startTimer("Measure - Effective Diameter")
        var maxDistances = 0
        var nbConnectedPaths = 0
        val nbPathAtDistance = HashMap[Int, Int]()
        for (i <- g.vertices; j <- g.vertices; if(i <= j && shortestPathFast(i)(j) != Byte.MaxValue)) {
            nbConnectedPaths += 1
            nbPathAtDistance(shortestPathFast(i)(j)) = nbPathAtDistance.getOrElse(shortestPathFast(i)(j), 0) + 1
        }
        var curNbConnectedPathBelowEffectiveDiameter = 0
        val (effectiveFiameter, _) = (nbPathAtDistance.toArray.sortBy(_._1).find { case (d, nbPaths)  =>
            curNbConnectedPathBelowEffectiveDiameter += nbPaths
            curNbConnectedPathBelowEffectiveDiameter.toDouble >= nbConnectedPaths.toDouble * 0.9
        }).get
        Timer.stopTimer("Measure - Effective Diameter")
        effectiveFiameter
    }

    lazy val scaleFree = {
        val gammas = for((deg, freq) <- degreeDistribution; if(deg > 4)) yield {
            val gamma = - 1 * math.log(freq.toDouble / g.nbVertices) / math.log(deg)
            if(gamma >= 2 && gamma <= 3) 1 else 0
        }
        gammas.sum.toDouble / gammas.size
    }
    
    def meanGammaScaleFree(minDegree : Int) = {
        var nbValid = 0
        var nbTotal = 0
        val gammas = (for((deg, freq) <- degreeDistribution; if(deg > minDegree)) yield {
            - 1 * math.log(freq.toDouble / g.nbVertices) / math.log(deg)
        })
        gammas.sum / gammas.size
    }

    /**
     * The center of gravity of all the points
     */
    lazy val gravityCenter = Stat.gravityCenterOfPoints(attr)

    /**
     * The total inertia of the points
     */
    lazy val totalInertia = {
        Timer.startTimer("Measure - totalInertia")
        val gr = gravityCenter
        val ti = (for (v <- g.vertices) yield {
            val d = Stat.distance(gr, attr(v))
            d * d
        }).sum.toDouble
        Timer.stopTimer("Measure - totalInertia")
        ti
    }

    /**
     * The between inertia
     */
    lazy val betweenInertia = {
        Timer.startTimer("Measure - betweenInertia")
        val g = gravityCenter
        val bi = (for (community <- communities) yield {
            val gCommunity = Stat.gravityCenterOfPoints(community.toArray.map(attr))
            community.size * Stat.distanceSquared(gCommunity, g)
        }).sum.toDouble
        Timer.stopTimer("Measure - betweenInertia")
        bi
    }
    lazy val betweenInertiaRatio = betweenInertia.toDouble / totalInertia

    /**
     * The inter class inertia
     */
    lazy val withinInertia = {
        Timer.startTimer("Measure - withinInertia")
        val wi = (for (community <- communities) yield {
            val gCommunity = Stat.gravityCenterOfPoints(community.toArray.map(attr))
            (for (v <- community.toArray) yield { Stat.distanceSquared(gCommunity, attr(v)) }).sum
        }).sum
        Timer.stopTimer("Measure - withinInertia")
        wi
    }
    lazy val withinInertiaRatio = withinInertia.toDouble / totalInertia

    /**
     * The pairwise distance matrix according to the attributes
     */
    lazy val distanceMatrix = {
        Timer.startTimer("Measure - Distance matrix")
        val dist = Array.fill(nbVertices, nbVertices)(0D)
        for (v1 <- g.vertices; v2 <- g.vertices) {
            val d = Stat.distance(attr(v1), attr(v2))
            dist(v1)(v2) = d
            dist(v2)(v1) = d
        }
        Timer.stopTimer("Measure - Distance matrix")
        dist
    }
   
    def homophilyWithin = {
        Timer.startTimer("Measure - homophilyWithin")
        val (nbIntEdges, nbExtEdges) = nbInternalExternalEdges
        var hw = 0D 
        for (v1 <- g.vertices; v2 <- g(v1).toArray) {
            if (Stat.isDistanceOnAttrBelow(attr(v1), attr(v2), dThresholds) && vertexCommunities(v1) == vertexCommunities(v2)) { hw += 1 } 
        }
        Timer.stopTimer("Measure - homophilyWithin")
        hw / (2 * nbIntEdges)
    }

    def homophilyBetween = {
        Timer.startTimer("Measure - homophilyBetween")
        val (nbIntEdges, nbExtEdges) = nbInternalExternalEdges
        var hb = 0D
        for (v1 <- g.vertices; v2 <- g(v1).toArray) yield { 
            if (Stat.isDistanceOnAttrBelow(attr(v1), attr(v2), dThresholds) && vertexCommunities(v1) != vertexCommunities(v2)) { hb += 1 }
        }
        Timer.stopTimer("Measure - homophilyBetween")
        hb / (2 * nbExtEdges)
    }
    
    def homophilyObserved = {
        Timer.startTimer("Measure - homophilyObserved")
        var ho = 0D
        for(v1 <- g.vertices) {
            for(v2 <- g(v1)) {
                var i = 0
		        var belowThreshold = false
		        while (i != attr(v1).length && !belowThreshold) {
		            belowThreshold = math.abs(attr(v1)(i) - attr(v2)(i)) < dThresholds(i)
		            i += 1
		        }
		        if(belowThreshold) {
		            ho += 1
		        }
            }
        }
        Timer.stopTimer("Measure - homophilyObserved")
        ho / (2 * nbEdges)
    }
    
    def homophilyExpectedOld = {
        Timer.startTimer("Measure - homophilyExpected")
        val nbAttr = g.nbAttributes
        var nbSimilarPair = 0
        val verticesArray = g.vertices.toArray
        val nbVertices = verticesArray.length
        var v1 = 0
        var i = 0
        var belowThreshold = false
        var attrV1 = new Array[Double](0)
        var attrV2 = new Array[Double](0)
        while(v1 != nbVertices) {
            var v2 = v1 + 1
            attrV1 = attr(verticesArray(v1))
            while(v2 != nbVertices) {
                i = 0
                belowThreshold = false
                attrV2 = attr(verticesArray(v2))
		        while (i != nbAttr && !belowThreshold) {
		            belowThreshold = math.abs(attrV1(i) - attrV2(i)) < dThresholds(i)
		            if(belowThreshold) { nbSimilarPair += 1 }
		            i += 1
		        }
                v2 += 1
            }
            v1 += 1
        }
        /*
        println(nbSimilarPair)
        println(nbSimilarPair * 2 + nbVertices)
        
        
        Timer.startTimer("v2")
        var he = 0
        for(v1 <- g.vertices; v2 <- g.vertices) {
            var i = 0
	        var belowThreshold = false
	        while (i != nbAttr && !belowThreshold) {
	            belowThreshold = math.abs(attr(v1)(i) - attr(v2)(i)) < dThresholds(i)
	            if(belowThreshold) { he += 1 }
	            i += 1
	        }
        }
        println(he)
        Timer.stopTimer("v2")
        */
        Timer.stopTimer("Measure - homophilyExpected")
        
//        println("Expected: " + ((nbSimilarPair * 2 + nbVertices).toDouble) / (nbVertices * (nbVertices - 1)))
//        println("Expected: " + (nbSimilarPair * 2 + nbVertices))
        
        val heFound = homophilyExpected
        if(heFound != (nbSimilarPair * 2 + nbVertices)) {
            println("Error !!!")
            System.exit(0)
        }
        
        ((nbSimilarPair * 2 + nbVertices).toDouble) / (nbVertices * (nbVertices - 1))
        /*
        var v1 = 0
        while(v1 != nbVertices) {
            var v2 = v1
            while(v2 != nbVertices) {
		        var i = 0
		        var belowThreshold = false
		        while (i != nbAttr && !belowThreshold) {
		            belowThreshold = math.abs(attr(v1)(i) - attr(v2)(i)) < dThresholds(i)
		            if(belowThreshold) { he += 1 }
		            i += 1
		        }
                v2 += 1
            }
            v1 += 1
        }
        Timer.stopTimer("homoExp")
        (2 * he.toDouble) / (nbVertices * (nbVertices - 1))
        */
//        ((for(v1 <- 0 until nbVertices; v2 <- 0 until nbVertices) yield { if(Stat.minDistance(attr(v1), attr(v2)) <= threshold) 1 else 0 }).sum.toDouble) / (nbVertices * (nbVertices - 1)) 
    }
    
    def homophilyExpected = {
        Timer.startTimer("homoExpFast")
        val nbAttr = attr(0).length
        val largestVertexId = g.vertices.max + 1
        val attributesSorted = (for(i <- 0 until nbAttr) yield {
            val attrSorted = (for((v, a) <- g.attr) yield { (a(i), v) }).toArray.sortBy(_._1)
	        (attrSorted.map(_._1), attrSorted.map(_._2))
        }).toArray
        val arraySize = attributesSorted(0)._1.length
        var he = 0
        for(v1 <- g.vertices) {
            val set = new MBitSet(largestVertexId)
            var i = 0
            while(i != nbAttr) {
                val matchingVertices = attributesSorted(i)._2
	            val idx1 = Arrays.binarySearch(attributesSorted(i)._1, attr(v1)(i) - dThresholds(i))
	            val idx2 = Arrays.binarySearch(attributesSorted(i)._1, attr(v1)(i) + dThresholds(i))
	            val minIdx = if(idx1 < 0) math.min(arraySize, -1 * idx1) - 1 else idx1
	        	val maxIdx = if(idx2 < 0) math.max(0, math.min(arraySize, -1 * idx2 - 1) - 1) else idx2
	        	
	        	var j = minIdx
	        	while(j <= maxIdx) {
        	        set.add(matchingVertices(j))
        	        j += 1
	        	}
                i += 1
            }
            he += set.size
        }
        Timer.stopTimer("homoExpFast")
        (he.toDouble) / (nbVertices * (nbVertices - 1))
    }
    
    /*
    def homophilyExpectedFast = {
        Timer.startTimer("homoExpFast")
        val nbAttr = attr(0).length
        
        val attributesSorted = (for(i <- 0 until nbAttr) yield {
            val attrSorted = (for((a, v) <- attr.zipWithIndex) yield { (a(i), v) }).toArray.sortBy(_._1)
	        (attrSorted.map(_._1), attrSorted.map(_._2))
        }).toArray
        
        var he = 0
        var v1 = 0
        while(v1 != nbVertices) {
            val set = new BitSet(nbVertices)
            for(i <- 0 until nbAttr) {
                val minAttr1 = attr(v1)(i) - dThresholds(i) 
	            val maxAttr1 = attr(v1)(i) + dThresholds(i)
	            val idx1 = Arrays.binarySearch(attributesSorted(i)._1, minAttr1)
	            val idx2 = Arrays.binarySearch(attributesSorted(i)._1, maxAttr1)
	            val minIdx = if(idx1 < 0) math.min(data.length, -1 * idx1) - 1 else idx1
	        	val maxIdx = if(idx2 < 0) math.max(0, math.min(data.length, -1 * idx2 - 1) - 1) else idx2
	        	var j = minIdx
	        	while(j <= maxIdx) {
        	        set.add(attributesSorted(i)._2(j))
        	        j += 1
	        	}
            }
            v1 += 1
            he += set.size
            
            /*
            val minAttr1 = attr(v1)(0) - dThresholds(0) 
            val maxAttr1 = attr(v1)(0) + dThresholds(0)
            val idx1 = Arrays.binarySearch(attr1ValuesSorted, minAttr1)
            val idx2 = Arrays.binarySearch(attr1ValuesSorted, maxAttr1)
            val minIdx = if(idx1 < 0) math.min(data.length, -1 * idx1) - 1 else idx1
        	val maxIdx = if(idx2 < 0) math.max(0, math.min(data.length, -1 * idx2 - 1) - 1) else idx2
            var x = minIdx
            while(x <= maxIdx + 1) {
                val v2 = attr1VerticesSorted(x)
                x += 1
		        var i = 1
		        var belowThreshold = false
		        while (i != nbAttr && !belowThreshold) {
		            belowThreshold = math.abs(attr(v1)(i) - attr(v2)(i)) < dThresholds(i)
		            if(belowThreshold) { he += 1 }
		            i += 1
		        }
            }
            v1 += 1
            */
        }
        Timer.stopTimer("homoExpFast")
        (he.toDouble) / (nbVertices * (nbVertices - 1))
//        ((for(v1 <- 0 until nbVertices; v2 <- 0 until nbVertices) yield { if(Stat.minDistance(attr(v1), attr(v2)) <= threshold) 1 else 0 }).sum.toDouble) / (nbVertices * (nbVertices - 1)) 
    }
    */
    
    /*
    def minMaxAttrValues : Array[(Double, Double)] = {
        attr
    }
    */
    
    /**
     * Compute the average shortest path between each pair of vertices
     * The geodesic distance between vertices in different connected components
     * are not taken into account
     */
    lazy val avgShortestPath: Double = {
        var sumDistances = 0
        var nbPair = 0
        val d = shortestPathFastCompute
        var i = 0
        var j = 0
        while(i != g.nbVertices) {
            var j = i + 1
            while(j != g.nbVertices) {
                if(d(i)(j) != Byte.MaxValue) {
                    sumDistances += d(i)(j)
            		nbPair += 1
                }
                j += 1
            }
            i += 1
        }
        sumDistances.toDouble / nbPair
    }

    /**
     * Return the number of edges within a communities and the number of edges
     * between two different communities
     */
    lazy val nbInternalExternalEdges: (Int, Int) = {
        Timer.startTimer("btwWthEdges")
        var nbWithinEdges = 0
        var nbBetweenEdges = 0
        for (v1 <- g.vertices; v2 <- g(v1)) {
            if (vertexCommunities(v1) == vertexCommunities(v2)) { nbWithinEdges += 1 } else { nbBetweenEdges += 1 }
        }
        Timer.stopTimer("btwWthEdges")
        (nbWithinEdges / 2, nbBetweenEdges / 2)
    }
    
    lazy val nbWithinEdges = nbInternalExternalEdges._1
    lazy val nbBetweenEdges = nbInternalExternalEdges._2

    /**
     * Return the ratio between the number of internal edges and the number of external edges
     */
    lazy val internalExternalEdgesRatio = {
        val (int, ext) = nbInternalExternalEdges
        int.toDouble / ext
    }

    /**
     * graph average clustering coefficient and number of triangles
     */
    lazy val (clusteringCoefficient, nbTriangles) = {
        Timer.startTimer("Measure - Clust. Coeff.")
        var cc = 0D
        var nbVert = 0
        var nbTriangles = 0
        for (v <- g.vertices; if (g.degree(v) > 1)) {
            val n = g(v)
            val neighbors = new Stack[Int]()
            n.foreach { x => neighbors.push(x) }
            var localCC = 0
            while (!neighbors.isEmpty) {
                val n1 = neighbors.pop()
                for (v <- neighbors; if (g(v).contains(n1))) { localCC += 1 }
            }
            nbTriangles += localCC
            cc += (2 * localCC.toDouble / (n.size * (n.size - 1)))
            nbVert += 1
        }
        Timer.stopTimer("Measure - Clust. Coeff.")
        (cc / (nbVert), nbTriangles / 3)
    }
    
    /**
     * Clustering coefficient in an Erdos-Renyi random graph
     * having the same number of edges and vertices
     */
    lazy val randomClusteringCoefficient = (2 * g.nbEdges.toDouble) / (g.nbVertices * (g.nbVertices - 1))

    /**
     * The graph degree distribution
     */
    lazy val degreeDistribution = {
        val degDistMap = HashMap[Int, Int]()
        val degrees = for (v1 <- g.vertices) {
            degDistMap(g.degree(v1)) = degDistMap.getOrElse(g.degree(v1), 0) + 1
        }
        degDistMap.toArray.sortBy(a => a._1)
    }

    /**
     * Modularity of the graph using the given partition
     * The modularity measure is the one by Newman and Girvan
     */
    def modularity2: Double = {
        (for (v1 <- g.vertices.toArray; v2 <- g.vertices.toArray; if (g.vertexCommunities(v1) == g.vertexCommunities(v2))) yield {
            (if (g.edgeExists(v1, v2)) 1D else 0D) - (g.degree(v1) * g.degree(v2) / (2 * nbEdges))
        }).sum / (2 * nbEdges)
    }
        
    def modularity : Double = {
        Timer.startTimer("Measure - Modularity")
//        val nbVertices = graph.size
        var nbInternalEdges = 0
        for (v1 <- g.vertices; v2 <- g(v1)) {
            if (vertexCommunities(v1) == vertexCommunities(v2)) { nbInternalEdges += 1 } 
        }
        var sumPos = nbInternalEdges.toDouble
        var sumNeg = 0D
        for(vertices <- communities) {
            var comDegSum = 0D
            for(v <- vertices) comDegSum += g.degree(v)
            sumNeg += comDegSum * comDegSum
        }
        sumNeg = sumNeg / (2 * nbEdges)
        Timer.stopTimer("Measure - Modularity")
	    math.min(1D, (sumPos - sumNeg) / (2 * nbEdges))
	}
    
    lazy val connectedComponents : Array[HashSet[Int]] = {
        Timer.startTimer("Measure - connected components")
        val connectedComponents = new Stack[HashSet[Int]]()
        val subgraph = HashSet[Int](g.vertices.toArray : _*)
        val verticesToConsider = new Stack[Int]()

        while (!subgraph.isEmpty) {
            // Build a new component
            val component = HashSet[Int]()
            val seed = subgraph.find(_ => true).get
            subgraph.remove(seed)
            verticesToConsider.push(seed)

            while (!verticesToConsider.isEmpty) {
                val vertexToAdd = verticesToConsider.pop()
                component.add(vertexToAdd)
                g(vertexToAdd).foreach { n => if (subgraph.remove(n)) { verticesToConsider.push(n) } }
            }
            connectedComponents.push(component)
        }
        Timer.stopTimer("Measure - connected components")
        connectedComponents.toArray
    }
    
    lazy val connectedComponentsSizes : Array[Int] = connectedComponents.map(_.size).toArray.sorted.reverse
    
    /**
     * Compute the size of the communities
     */
    lazy val communitiesSize: Array[Int] = g.partition.values.toArray.map(_.size)
    lazy val minCommunitySize = communitiesSize.min
    lazy val maxCommunitySize = communitiesSize.max
    lazy val avgCommunitySize = communitiesSize.sum.toDouble / communitiesSize.length
}

object TestMeasures {
    val stubGraph = new Array[HashSet[Int]](0)
    val stubAttr = Array.fill(1)(Array[Double](0))
    val stubCommunities = new Array[Int](0)

    def main(args: Array[String]) = {
        // testGravityCenter()
        testInertia()
    }

    def computeModularity() = {
        val g = Array[HashSet[Int]](
            HashSet(1, 2),
            HashSet(0, 2, 3, 5),
            HashSet(0, 1, 3, 4),
            HashSet(1, 2, 6),
            HashSet(2, 6),
            HashSet(1),
            HashSet(3, 4))
        val vToC = Array(1, 1, 1, 1, 2, 2, 2)

        // val m = new GraphMeasures(g, stubAttr, vToC)
        // println(m.modularity)
    }

    def testInertia() = {
        val points = Array(
            Array(4D, 2D, 3D),
            Array(7D, 3D, 8D),
            Array(3D, 5D, 8D),
            Array(6D, 2D, 5D),
            Array(7D, 4D, 12D),
            Array(9D, 8D, 6D))
        val comm = Array(0, 0, 0, 0, 1, 1)
        val stubGraph = new Array[HashSet[Int]](points.length)
        /*
        val measures = new GraphMeasures(stubGraph, points, comm)
        println(measures.betweenInertia)
        println(measures.withinInertia)
        */
        //        val g = Stat.gravityCenterOfPoints(points)
        //        println("")
        //        println(g.mkString(" "))
    }

    def testGravityCenter() = {
        val points = Array(
            Array(4D, 2D, 3D),
            Array(7D, 3D, 8D),
            Array(3D, 5D, 8D),
            Array(6D, 2D, 5D))
        // val measures = new GraphMeasures(stubGraph, stubAttr, stubCommunities)
        // val g = Stat.gravityCenterOfPoints(points)
        // println(g.mkString(" "))
        computeModularity()
    }
}
