/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package scalargs

import scalargs.argtypes.{StringArg, IntArg, ArgType}
import collection.mutable.{LinkedHashSet, ListBuffer, LinkedHashMap}

class UserInputException() extends Exception {

}

class Scalargs() {
    var args = new LinkedHashMap[String, ArgType]()
    var argsName = new LinkedHashMap[String, ArgType]()
    var argsGroups = new LinkedHashMap[String, ListBuffer[ArgType]]()
    var positionalArgs = new ListBuffer[ArgType]()
    var nonPositionalArgs = new ListBuffer[ArgType]()
    val description = ""
    val progName = ""
    var argsFound = false
    var lineSize = 80
    var leftMargin = 2
    var helpMargin = 25

    def getHelpMessage() : String = {
        var helpStr = "" 
        for((groupName, args) <- argsGroups) {
            helpStr += "\n" + groupName + ":" + "\n"
            for(arg <- args) {
                var argNameHelp = (" " * leftMargin) + arg.allNames.mkString(",") + arg.getArgParams()
                /*
                if (arg.showDefault) {
                    var defaultValue = getValue(arg.field.getName)
                    if (defaultValue != null) {
                        argNameHelp += " = " + defaultValue
                    }
                }
                */
                helpStr += argNameHelp
                helpStr += (if(argNameHelp.size + 1 > helpMargin) "\n" + " " * helpMargin else " " * (helpMargin - argNameHelp.size))
                helpStr += formatString(arg.help, "\n" + " " * helpMargin, lineSize - helpMargin) + "\n"
            }
        }
        return helpStr
    }

    def getUsage() : String = {
        if(!argsFound) {
            getOptions()
        }

        var usageStr = "usage: " + progName + " "
        var usageBeginLen = usageStr.size
        
        var curLineLen = usageBeginLen
        for(arg <- nonPositionalArgs) {
            var argUsage = arg.getArgUsage()

            // format optionUsage nicely
            if (curLineLen + argUsage.size > lineSize) {
                usageStr += addNewLine(usageBeginLen) + argUsage
                curLineLen = usageBeginLen + argUsage.size
            } else {
                usageStr += argUsage
                curLineLen += argUsage.size
            }
        }
        if (nonPositionalArgs.size > 0) {
            usageStr += addNewLine(usageBeginLen)
            curLineLen = usageBeginLen
        }
        for(arg <- positionalArgs) {
            var argUsage = arg.getArgUsage()
            // format optionUsage nicely
            if (curLineLen + argUsage.size > lineSize) {
                usageStr += addNewLine(usageBeginLen) + argUsage
                curLineLen = usageBeginLen + argUsage.size
            } else {
                usageStr += argUsage
                curLineLen += argUsage.size
            }
        }

        return usageStr
    }

    def getDescription() : String = {
        return formatString(description, "\n", lineSize)
    }

    def formatString(string : String, lineSeparator : String, lineSize : Int) : String = {
        if(string.size == 0) {
            return string
        }
        var curLineLen = 0
        var formatedStr = ""
        string.split(" ").foreach { word =>
            if(curLineLen + word.size > lineSize) {
                formatedStr += lineSeparator
                curLineLen = 0
            }
            formatedStr += word + " "
            curLineLen += word.size + 1
        }
        return formatedStr
    }

    def addNewLine(nbSpaces : Int) : String = {
        return "\n" + " " * nbSpaces
    }

    def getFullHelpMessage() : String = {
        return Array[String](getUsage(), getDescription(), getHelpMessage()).filter(_.size > 0).mkString("\n")
    }

    /**
     * Return a list of all the field of type T
     */
    def getFieldOfType() : ListBuffer[ArgType] = {
        var allFields = new ListBuffer[ArgType]()
        getClass.getDeclaredFields.foreach { field =>
            var getter = this.getClass.getMethods.find(_.getName == field.getName())
            if(getter != None) {
                var curField = getter.get.invoke(this)
                if (curField.isInstanceOf[ArgType]) {
                    allFields += curField.asInstanceOf[ArgType]
                }
            }
        }
        return allFields
    }

    def getOptions() : Unit = {
        var allNames = new LinkedHashSet[String]()
        getFieldOfType.foreach { arg =>
            arg.allNames.foreach { name =>
                if(allNames.contains(name)) {
                    throw new Exception("Multiple arguments have name '" + name + "'")
                }
                if(!arg.isPositional) {
                    argsName(name) = arg
                }
                allNames.add(name)
            }

            if(arg.isPositional) {
                positionalArgs += arg
            } else {
                nonPositionalArgs += arg
            }
            argsGroups.getOrElseUpdate(arg.group, new ListBuffer()) += arg
            args(arg.name) = arg
        }
        argsFound = true
    }

    def parseArgs(argValues : Array[String]) : Boolean = {
        if(!argsFound) {
            getOptions()
        }

        // Assign value to each arg
        var curArg : ArgType = null
        var lookForPositionalArg = true
        var positionalIdx = 0
        argValues.foreach { argValue =>
            if (argsName.contains(argValue)) {
                curArg = argsName(argValue)
                curArg.isSet = true
                lookForPositionalArg = false
            } else {
                if (!lookForPositionalArg) {
                    if (curArg.canAddAnotherParam) {
                        curArg.addParam(argValue)
                    } else {
                        lookForPositionalArg = true
                    }
                }
                if(lookForPositionalArg) {
                    if (positionalIdx >= positionalArgs.size) {
                        Console.err.println("error: unable to use argument: " + argValue)
                        return false
                    } else {
                        positionalArgs(positionalIdx).addParam(argValue)
                        positionalArgs(positionalIdx).isSet = true
                        positionalIdx += 1
                    }
                }
            }
        }

        // check that the args have a correct number of params
        // and all required params are set
        for((fieldName, arg) <- args) {
            if (!arg.isSet && arg.isRequired) {
                Console.err.println("error: missing required arg " + arg.name)
                return false
            }
            if (arg.isSet && !arg.hasValidNumberOfParams) {
                Console.err.println("error: missing params for " + arg.name)
                return false
            }
        }

        return true
    }
}