/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package utils

import scala.collection.mutable.HashMap
import java.io.File
import java.io.FileWriter
import sys.process._
import java.awt.Color

/**
 * Simple interface with GnuPlot
 */
object Gnuplot {
    val tmpDataFileName = "/tmp/data.dat"
    
	def plot(legend : Vector[String],
	        xAxis : Axis, yAxis : Axis,
	        output : String,
	        dataFile : Option[String] = None,
	        style : GnuplotStyle = new GnuplotStyle(),
	        grid : Grid.Grid = Grid.Horizontal,
	        legendPosition : LegendPosition.LegendPosition = LegendPosition.TopLeft,
	        options : List[LegendOption] = List()) =  {
	    
	    // Select output type
	    val extension = (output.split("\\.")(output.split("\\.").size - 1)).toLowerCase()
	    val terminal = extension match {
	        case "eps" => "postscript eps enhanced solid font 'Helvetica,18' size 10cm, 7cm"
	        case "png" => "png"
            case "tex" => "epslatex size 3.5,2.62"
	        case "svg" => "svg"
	        case _ => "eps"
	    }
	    // Build script
	    val dataFileName = dataFile.getOrElse(tmpDataFileName)
	    val plotCommand = legend.zipWithIndex.map({ case (label, idx) => 
	        "'" + dataFileName + "' using 1:" + (idx + 2) + " title '" + label + "' ls " + (idx + 1)
	    }).mkString(", ")
	    
	    val gridCmd = grid match {
	        case Grid.Both => "set grid"
	        case Grid.Horizontal => "set grid ytics"
            case Grid.Vertical => "set grid xtics"
            case Grid.None => ""
	    }
	    val legendPositionCmd = "set key " + (legendPosition match {
	        case LegendPosition.BottomLeft => "bottom left"
            case LegendPosition.BottomRight => "bottom right"
            case LegendPosition.TopLeft => "top left"
            case LegendPosition.TopRight => "top right"
	    })
	    
	    
	    val logScaleX = if(xAxis.isLog) "set logscale x" else ""
        val logScaleY = if(yAxis.isLog) "set logscale y" else ""
	    val xRangeCmd = if(xAxis.range.isDefined) {
	        val xRange = xAxis.range.get
	        "set xrange [" + xRange.start + ":" + xRange.end + "]\n" +
	        (if(!xAxis.isLog) "set xtics " + xRange.step + "\n" else "")
	    } else ""
        val yRangeCmd = if(yAxis.range.isDefined) {
	        val yRange = yAxis.range.get
	        "set yrange [" + yRange.start + ":" + yRange.end + "]\n" +
	        (if(!yAxis.isLog) "set ytics " + yRange.step + "\n" else "")
	    } else ""
        val yLabelCmd = if(yAxis.legend.size == 0) "" else "set ylabel '" + yAxis.legend + "' font ',28'"
        val xLabelCmd = if(xAxis.legend.size == 0) "" else "set xlabel '" + xAxis.legend + "' font ',28'"
	        
	    var commands = List(
	        "set terminal " + terminal,
	        "set termoption dashed",
	        "set output '" + output + "'",
	        gridCmd,
	        legendPositionCmd,
	        xLabelCmd,
	        yLabelCmd,
	        xRangeCmd,
	        yRangeCmd,
	        logScaleX,
	        logScaleY,
	        "set key horiz",
	        "set key center top",
	        "set style data linespoints", 
	        style.getStyleCommand(legend.size),
	        "plot " + plotCommand
	    )
	    options.foreach { option =>
            commands = option.getCommands ::: commands 
        }
	    val script = commands.mkString("\n")
        // println(script)
        var scriptFile = new FileWriter("/tmp/script.gp")
	    scriptFile.write(script)
	    scriptFile.flush()
	    
	    // Run script
	    "gnuplot /tmp/script.gp"!
	}
	
	def createDataFile(data : Array[Vector[Double]], xValues : Option[Range] = None, dataFileName : Option[String] = None) = {
	    var dataFile = if(dataFileName.isDefined) {
	        new FileWriter(dataFileName.get)
	    } else {
	        new FileWriter(tmpDataFileName)
	    }
	    val xRange = xValues.getOrElse(0 to data.map(_.size).max)
	    xRange.zipWithIndex.foreach { case (x, idx) =>
	        dataFile.write(data.map(values => {
	            if(idx >= values.length) Double.NaN else values(idx)
	        }).mkString(x + "\t", "\t", "\n"))
	    }
	    dataFile.flush()
	}
	
	def main(args : Array[String]) : Unit = {
	    val data = Array[Vector[Double]](Vector(12, 13, 42), Vector(125, 46, 120))
	    val legend = Vector("$\\mathfrak{s} = 50$", "$\\mathfrak{s} = 100$")
	    var style = new GnuplotStyle(lineWidth = 3)
	    style.lineWidthIncrease = 3
	    
	    createDataFile(data, Option(4 to 8 by 2))
	    plot(legend, Axis("Nb vert", Option(4 to 8 by 2)), Axis("Runtime (sec)"), "/tmp/plot2.tex", grid = Grid.Horizontal, style = style, 
	            options = List(LegendOption(2)))
	}
}

abstract class GnuPlotOption {
    def getCommands() : Seq[String]
}

case class LegendOption(maxRows : Int, size : Option[Double] = None) extends GnuPlotOption {
    def getCommands() : List[String] = {
        List("set key maxrow " + maxRows)
    }
}

class GnuplotStyle(var lineWidth : Double = 4, var pointSize : Double = 1, var baseColor : Color = new Color(51, 51, 153)) {
    var lineWidthIncrease : Double = 0
    var pointSizeIncrease : Double  = 0
    var isDefaultPointTypes = false
    
    var colorScale = Vector[String]()
    
    var pointTypes = Array(8, 4, 6, 12, 2, 10, 1, 3, 5, 7, 9, 11, 13)
    
    def getStyleCommand(nbLines : Int) : String = {
        buildColorScale(nbLines)
        (0 until nbLines).map({ i =>
            val pointType = if(isDefaultPointTypes) "" + i + 1 else pointTypes(i % pointTypes.size)
            "set style line " + (i + 1) + 
            " linetype " + (i + 1) + 
            " pointsize " + (pointSize + pointSizeIncrease * i) + 
            " linewidth " + (lineWidth + lineWidthIncrease * i) + 
            " linecolor " + colorScale(i) + 
            " pointtype " + pointType  
        }).mkString("\n")
    }
    
    def buildColorScale(nbLines : Int) = {
        colorScale = Vector[String]()
        val hsbColorArray = Color.RGBtoHSB(baseColor.getRed(), baseColor.getGreen(), baseColor.getBlue(), new Array[Float](3))
        var hue = hsbColorArray(0)
        val saturation = hsbColorArray(1)
        val value = hsbColorArray(2)
        val deltaHue = 1f / nbLines
        0 until nbLines foreach { i =>
            val newColor = "#" + Integer.toHexString(Color.getHSBColor(hue, saturation, value).getRGB()).substring(2)
            colorScale = colorScale :+ "rgb '" + newColor + "'"
            hue += (deltaHue % 1f)
        }
    }
}

object LegendPosition extends Enumeration {
    type LegendPosition = Value
    val TopLeft, TopRight, BottomLeft, BottomRight = Value
}

object Grid extends Enumeration {
    type Grid = Value
    val None, Both, Horizontal, Vertical = Value
}

case class Axis(legend : String, range : Option[Range] = None, isLog : Boolean = false)