/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package utils

import java.io.PrintWriter
import scala.sys.process._
import java.io.File

case class LatexHeader(title : String, packages : List[LatexPackage] = List(), author : String = "")
case class LatexPackage(name : String, options : String = "")

class Latex(fileName : String, header : Option[LatexHeader] = None) {
    val file = new File(fileName)
    val path = new File(fileName).getParentFile().getPath()
    val out = new PrintWriter(fileName)
    for(h <- header) { printHeader(h) }
    
	def printPackage(packageName : String) = { print(s"\\usepackage{${packageName}}") }
    def printPackages(packages : List[LatexPackage]) = { 
        print((packages.map(p => s"\\usepackage[${p.options}]{${p.name}}")).mkString("\n")) 
    }
    
    def print(s : String) = { out.println(s) }

    def printTableRow(row : Seq[String]) = { print(row.mkString("", " & ", "\\\\")) }
    
    def printSection(section : String) = { print(s"\\section{${section}}") }
    def printSubsection(section : String) = { print(s"\\subsection{${section}}") }
    def newLine() = { print("\\\\") }
    
    def printGraphics(imageFile : String, width : Double) = { 
        print(s"\\includegraphics[width=${width}\\linewidth]{${path}/${imageFile}}") 
    }
    
    
    
    def printHeader(header : LatexHeader) {
        print("\\documentclass[a4paper,10pt]{article}") 
	    print("\\usepackage[utf8]{inputenc}")
	    printPackages(header.packages)
	    print(s"\\title{${header.title}}")
	    print(s"\\author{${header.author}}")
	    print("\\begin{document}")
	    print("\\maketitle")
    }
    
    def printEnding(doNotClose : Boolean = false) { 
        print("\\end{document}")
        if(!doNotClose) close()
    }
    
    def compile() {
        s"pdflatex  -output-directory=${path} ${fileName}" !!
        
        new File(fileName.replaceAllLiterally(".tex", ".log")).delete()
        new File(fileName.replaceAllLiterally(".tex", ".aux")).delete()
    }
    
    def printFigure(caption : String, label : String)(f : Unit => Unit) = {
        print("\\begin{figure}")
        f()
        print(s"\\caption{${caption}}")
        print(s"\\label{${label}}")
        print("\\end{figure}")
    }
    
    def printItemize(items : Seq[String]) = {
        print(items.map(s => "\\item " + s).mkString("\\begin{itemize}", "\n", "\\end{itemize}"))
    }
    
    def close() = {
        out.flush()
        out.close()
    }
}

object Latex {
    def use(fileName : String, close : Boolean = true)(f : Latex => Unit) = {
        val latexOutput = new Latex(fileName)
        f(latexOutput)
        if(close) latexOutput.close
    }
    
    def main(args : Array[String]) = {
        val w = new Latex("jar/test.tex")
        w.printHeader(LatexHeader("Test", List(LatexPackage("graphicx"))))
        w.printGraphics("bla/Density.pdf", .5)
        w.printEnding()
        w.close()
        w.compile()
    }
}