/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package utils

import scala.collection.mutable.BitSet
import utils.Timer
import java.io.DataOutputStream
import java.util.Arrays

object BitSetToMBitSet {
    def build(bitset: BitSet): MBitSet = {
        var initSize = bitset.max
        var array = new Array[Long]((initSize + 63) >> 6 max 1)
        var copy = new MBitSet(array)
        bitset.foreach { v =>
            copy.add(v)
        }
        copy
    }
}

/**
 * Efficient implementation of the BitSet structure using assembly operation to traverse the collection
 * 
 * This structure should be used when the set is dense and need to be efficiently traversed
 * It has also very fast intersection and union operation
 */
class MBitSet(val elems: Array[Long]) {
    private val serialVersionUID = 12545687215435L
    
    def this(initSize: Int) = this(new Array[Long]((initSize + 63) >> 6 max 1))

    val nwords = elems.length
    
    def unionUpdate(other : MBitSet) = {
        var i = 0
        while(i != nwords) {
            elems(i) = elems(i) | other.elems(i)
            i += 1
        }
    }
    
    def clear() = {
        Arrays.fill(elems, 0L)
    }
    
    def mkString(sep : String) : String = {
        var isFirst = true
        var result = ""
        foreach { v =>
            if(isFirst) {
                isFirst = false
            } else {
                result += sep
            }
            result += v
        }
        result
    }
    
    def isEmpty : Boolean = {
        var i = 0
        while(i != nwords && elems(i) == 0L) {
            i += 1
        }
        i == nwords
    }
    
    def find(f : Int => Boolean) : Option[Int] = {
        var l = 0L
		var iter2 = nwords
		var res : Option[Int] = None
		
		while(iter2 >= 0 && !res.isDefined) {
		    if(l == 0) {
	            do { iter2 -= 1 } while(iter2 >= 0 && elems(iter2) == 0L)
                if(iter2 >= 0) {
                	l = elems(iter2)
                	val b = java.lang.Long.numberOfTrailingZeros(l)
			        l &= ~(1L << b)
			        if(f((iter2 << LogWL) + b)) {
			            res = Option((iter2 << LogWL) + b)
			        }
                }
	        } else {
	            val b = java.lang.Long.numberOfTrailingZeros(l)
		        l &= ~(1L << b)
		        if(f((iter2 << LogWL) + b)) {
		            res = Option((iter2 << LogWL) + b)
		        }
	        }
		}
        res
    }
    
    def union(other : MBitSet) : MBitSet = {
        val words = elems.clone
        var idx = nwords - 1
        while (idx >= 0) {
            words(idx) |= other.elems(idx)
            idx -= 1
        }
        new MBitSet(words)
    }
    
    def unionSize(other : MBitSet) : Int = {
        var idx = 0
        var unionSize = 0
        while (idx != nwords) {
            unionSize += java.lang.Long.bitCount((elems(idx) | other.elems(idx)))
            idx += 1
        }
        unionSize
    }
    
    def save(output : DataOutputStream) = {
        output.writeInt(nwords)
        var i = 0
        while(i != nwords) {
            output.writeLong(elems(i))
            i += 1
        }
    }
    
    val LogWL = 6
    def size(): Int = {
        var s = 0
        var i = nwords
        while (i > 0) {
            i -= 1
            s += java.lang.Long.bitCount(elems(i))
        }
        s
    }
    def add(elem: Int) = {
        elems(elem >> LogWL) |= (1L << elem)
    }
    
    var iter = 0
    def iterNextAndRemove2() : Int = {
        while(elems(iter) == 0L) {
          iter += 1
        }
        val value = (iter << LogWL) + (java.lang.Long.numberOfTrailingZeros(elems(iter)))
        elems(iter) = elems(iter) & (elems(iter) - 1)
        value
    }
    
    def resetIter() = { 
        iter = 0
        iter2 = -1
        l = 0L
    } 

    
    var l = 0L
    var iter2 = nwords

    def iterNext() : Int = {
        if(l == 0) {
            do { iter2 -= 1 } while(iter2 >= 0 && elems(iter2) == 0L)
            if(iter2 < 0) {
                -1
            } else {
                l = elems(iter2)
                val b = java.lang.Long.numberOfTrailingZeros(l)
		        l &= ~(1L << b)
		        (iter2 << LogWL) + b
            }
        } else {
            val b = java.lang.Long.numberOfTrailingZeros(l)
	        l &= ~(1L << b)
	        (iter2 << LogWL) + b
        }
    }
    
    def foreach(f : Int => Unit) = {
        var i = 0
        while (i != nwords) {
            // Only iterate when the current value is different than 0
            if (elems(i) != 0L) {
                var l = elems(i)
                do {
                    f((i << LogWL) + java.lang.Long.numberOfTrailingZeros(l))
                    // Locally unset the last enumerated bit 
                    l &= l - 1
                } while (l != 0L)
            }
            i += 1
        }
        /*
        var l = 0L
		var iter2 = nwords
		
		while(iter2 >= 0) {
		    if(l == 0) {
	            do { iter2 -= 1 } while(iter2 >= 0 && elems(iter2) == 0L)
                if(iter2 >= 0) {
                	l = elems(iter2)
                	val b = java.lang.Long.numberOfTrailingZeros(l)
			        l &= ~(1L << b)
			        f((iter2 << LogWL) + b)
                }
	        } else {
	            val b = java.lang.Long.numberOfTrailingZeros(l)
		        l &= ~(1L << b)
		        f((iter2 << LogWL) + b)
	        }
		}
		*/
    }
    
    override def clone(): MBitSet = {
        val elems1 = new Array[Long](elems.length)
        Array.copy(elems, 0, elems1, 0, elems.length)
        new MBitSet(elems1)
    }

    def subsetOf(other: MBitSet): Boolean = {
        var idx = 0
        var isSubset = true
        while(idx != nwords && isSubset) {
            isSubset = (elems(idx) & ~other.elems(idx)) == 0L
            idx += 1
        }
        isSubset
        // val len = this.nwords min other.nwords
        // (0 until len) forall (idx => (elems(idx) & ~other.elems(idx)) == 0L)
    }

    def contains(elem: Int) = (elems(elem >> LogWL) & (1L << elem)) != 0L

    def & (other: MBitSet): MBitSet = {
        // val words = new Array[Long](len)
        val words = elems.clone
        var idx = nwords - 1
        while (idx >= 0) {
            words(idx) &= other.elems(idx)
            idx -= 1
        }
        new MBitSet(words)
    }
    
    def intersectUpdate(other : MBitSet) = {
        var i = 0
        while(i != nwords) {
            elems(i) = elems(i) & other.elems(i)
            i += 1
        }
    }
    
    def intersect(other : MBitSet) : MBitSet = {
        val words = other.elems.clone
        var i = 0
        while(i != nwords) {
            words(i) &= elems(i)
            i += 1
        }
        new MBitSet(words)
    }
    
    def intersectSize(other : MBitSet) : Int = {
        var intersectionSize = 0
        var idx = 0
        while(idx != nwords) {
            intersectionSize += java.lang.Long.bitCount((elems(idx) & other.elems(idx)))
            idx += 1
        }
        intersectionSize
    }
    
    def hasMinIntersect(other: MBitSet, minSize : Int) : Boolean = {
        val len = this.nwords min other.nwords
        var idx = 0
        var curSize = 0
        while (idx != len && curSize < minSize) {
            curSize += java.lang.Long.bitCount((elems(idx) & other.elems(idx)))
            idx += 1
        }
        curSize >= minSize
    }
    
    def hasMinPonderatedIntersection(other: MBitSet, minSize : Int, weights : Array[Int]) : Boolean = {
        val len = this.nwords
        var idx = 0
        var curSize = 0
        while (idx != len && curSize < minSize) {
            var localIntersect = (elems(idx) & other.elems(idx))
            while(localIntersect != 0L) {
            	curSize += weights((idx << LogWL) + java.lang.Long.numberOfTrailingZeros(localIntersect))
	            localIntersect = localIntersect & (localIntersect - 1)
	        }
            idx += 1
        }
        curSize >= minSize
    }
    
    /**
     * It is required to define the maximum value as the number of bits might be higher 
     * than the number of values
     */
    /*
    def setAllBitsUntil(maxValue : Int) = {
        var idx = 0
        while (idx < nwords) {
            elems(idx) = 0xffffffff
            idx += 1
        }
    }
    */
    override def toString() : String = {
        var values = List[Int]()
        foreach { v =>
            values = v :: values
        }
        values.mkString(" ")
    }
    
    def remove(elem: Int): Boolean = {
        val idx = elem >> LogWL
    	if ((elems(idx) & (1L << elem)) != 0L) {
	      elems(idx) = elems(idx) & ~(1L << elem)
	      true
	    } else false
    }
    
    def removeNoCheck(elem: Int) = {
	      val idx = elem >> LogWL
	      elems(idx) = elems(idx) & ~(1L << elem)
    }
}

object TestBitSet {
    def main(args : Array[String]) = {
        var z = 4876212154572201L
		println(z.toBinaryString)
        while(z != 0L) {
            println(java.lang.Long.numberOfTrailingZeros(z))
            z = z & (z - 1)
        }
        System.exit(0)
        
        val len = 1000 * 1000 * 100
        // val len = 1000
        val density = 0.9
        var x = new MBitSet(len)
        var y = new BitSet(len)
        0 until (len * density).toInt foreach { i =>
            val v = (math.random * len).toInt
            x.add(v)
            y.add(v)
        }
        
        var count = 0
        val size = x.size
        Timer.startTimer("Iter Fast")
        var v = x.iterNext
        while(v != -1) {
            v = x.iterNext
            count += 1
        }
        x.resetIter()
        Timer.stopTimer("Iter Fast")
        println(count)
        
        count = 0
        Timer.startTimer("Iter Fast foreach")
        x.foreach { v =>
            count += 1
        }
        Timer.stopTimer("Iter Fast foreach")
        println(count)
        
        count = 0
        Timer.startTimer("Iter BitSet")
        val iter = y.iterator
        while(iter.hasNext) {
            var v = iter.next
            count += 1
        }
        Timer.stopTimer("Iter BitSet")
        println(count)
        
        count = 0
        Timer.startTimer("Iter Last")
        y.foreach { v =>
            count += 1
        }
        Timer.stopTimer("Iter Last")
        println(count)
        
        
        Timer.printTimers
        
    }
}
