/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package utils

import scala.collection.mutable.ArrayBuffer
import scala.collection.mutable.HashSet
import scala.util.Random
import scala.collection.mutable.HashMap
import scala.collection.IndexedSeq

object Stat {
    /**
     * Return an integer taken uniformly at random between min and max.
     * max is required to be superior to min
     * 	@param min the minimum value of the returned integer
     *  @param max the maximum value of the returned integer
     */
    def randInt(min: Int, max: Int) = Random.nextInt(max - min + 1) + min

    /**
     * Return a sample of sampleSize elements of items selected uniformly at random
     */
    def sample(items: Array[Int], sampleSize: Int, doNotPick: Int = -1): HashSet[Int] = {
        var i = 0
        val sampleIdx = HashSet[Int]()
        val sample = HashSet[Int]()
        val maxItemsSize = if (doNotPick == -1) items.size else items.size - 1
        while (sampleIdx.size != sampleSize && sampleIdx.size != maxItemsSize) {
            sampleIdx.add((Random.nextDouble * maxItemsSize).toInt)
        }
        for (e <- items) {
            if (e == doNotPick) { i -= 1 }
            else if (sampleIdx.contains(i)) { sample.add(e) }
            i += 1
        }
        sample
    }
    def sample(items: HashSet[Int], sampleSize: Int): HashSet[Int] = sample(items.toArray, sampleSize, -1)
	def sample(items: ArrayBuffer[Int], sampleSize: Int): HashSet[Int] = sample(items.toArray, sampleSize, -1)
	def sample(items: HashSet[Int], sampleSize: Int, doNotPick: Int): HashSet[Int] = sample(items.toArray, sampleSize, doNotPick)
	def sample(items: ArrayBuffer[Int], sampleSize: Int, doNotPick: Int): HashSet[Int] = sample(items.toArray, sampleSize, doNotPick)

    def randIntPowerLaw(maxValue: Int, exp: Double) = {
        weightedProbability[Int](1 to maxValue, x => math.pow(x, -exp))
    }

    def pickAtRandom[I](seq : IndexedSeq[I]) : I  = {
        seq(Random.nextInt(seq.length))
    }
    def pickAtRandom(seq : HashSet[Int]) : Int = pickAtRandom[Int](seq.toArray)
    def pickAtRandom(seq : Set[Int]) : Int = pickAtRandom[Int](seq.toArray)
    
    /**
     * Pick an item from items according to the almost distribution function f
     */
    def weightedProbability[A](items: Iterable[A], f: A => Double): A = {
        var sum = 0D
        for(v <- items) { sum += f(v)}
        val idx = Random.nextDouble * sum
        var acc = 0D
        (items.find { i => acc += f(i); acc >= idx }).get
    }

    def inverseWeightedProbabilityOld[A](items: Iterable[A], f: A => Double): A = {
        var sumNorm = 0D
        for(v <- items) { sumNorm += 1D / f(v)}
        
//        val sumNorm = items.map(i => 1D / f(i)).sum
        val idx = Random.nextDouble * sumNorm
        var acc = 0D
        (items.find { i => acc += (1D / f(i)); acc >= idx }).get
    }
    
    def inverseWeightedProbability[A](items: Iterable[A], f: A => Double): A = {
        Timer.startTimer("Inverse probability")
        var sumNorm = 0D
        var i = 0
        var cacheValues = new Array[Double](items.size)
        for(v <- items) {
            cacheValues(i) = 1D / f(v)
            sumNorm += cacheValues(i)
            i += 1
        }
        
//        val sumNorm = items.map(i => 1D / f(i)).sum
        val idx = Random.nextDouble * sumNorm
        var acc = 0D
        var j = 0
        val x = (items.find { i =>
            acc += cacheValues(j)
            j += 1
            acc >= idx
        }).get
        Timer.stopTimer("Inverse probability")
        x
    }
    
    def weightedProbabilityWithSum[A](items: ArrayBuffer[A], f: A => Int, sum : Double) : A = {
        val idx = Random.nextDouble * sum
        var acc = 0D
        var i = -1
        while(acc < idx) {
            i += 1
            acc += f(items(i))
        }
        items(i)
//         (items.find { i => acc += f(i); acc >= idx }).get
    }
    
    def weightedProbabilityWithSum(items: HashSet[Int], f: Int => Int, sum : Double) : Int = {
        /*
        val exp = (for(x <- items.toArray) yield (f(x))).sum
        if(sum != exp) {
            println("Wrong expected sum !!")
            println("found: " + sum)
            println("exp: " + exp)
        }
        */
        val idx = Random.nextDouble * sum
        var acc = 0D
        items.find { i => 
            acc += f(i)
            acc >= idx
        }.get
    }
    
    def weightedProbabilityWithSum(items: Seq[Int], f: Int => Int, sum : Double) : Int = {
        val idx = Random.nextDouble * sum
        var acc = 0D
        items.find { i => 
            acc += f(i)
            acc >= idx
        }.get
    }
    
    def sampleWeightedProbabilityWithSum[A](items: ArrayBuffer[A], f: A => Double, sum : Double, nbItems : Int) : HashSet[A] = {
        println("Ensure that the sum is correct !!")
        
        
        val idxs = Array.fill(nbItems)(Random.nextDouble * sum).sorted
        var ii = 0
        var idx = idxs(ii)
        var acc = 0D
        var i = -1
        val res = HashSet[A]()
        while(ii < nbItems) {
            i += 1
            acc += f(items(i))
            while(acc >= idx && ii != nbItems) {
                ii += 1
                if(ii != nbItems) {
                    idx = idxs(ii)
                    res.add(items(i))
                }
            }
        }
        res
//         (items.find { i => acc += f(i); acc >= idx }).get
    }

    
    /**
     * Compute the Euclidean distance between two points
     */
    def distance(a: Array[Double], b: Array[Double]): Double = {
        var d = 0D
        var i = 0
        while (i != a.length) {
            val delta = a(i) - b(i)
            d += delta * delta
            i += 1
        }
        math.sqrt(d)
    }

    /**
     * Compute the minimal difference
     */
    def minDistance(a: Array[Double], b: Array[Double]): Double = {
        var minD = Double.MaxValue
        var i = 0
        while (i != a.length) {
            minD = math.min(math.abs(a(i) - b(i)), minD)
            i += 1
        }
        minD
    }

    def isDistanceOnAttrBelow(a: Array[Double], b: Array[Double], thresholds: Array[Double]): Boolean = {
        var i = 0
        var belowThreshold = false
        while (i != a.length && !belowThreshold) {
            belowThreshold = math.abs(a(i) - b(i)) < thresholds(i)
            i += 1
        }
        belowThreshold
    }

    def distanceSquared(a: Array[Double], b: Array[Double]): Double = {
        var d = 0D
        var i = 0
        while (i != a.length) {
            val delta = a(i) - b(i)
            d += delta * delta
            i += 1
        }
        d
    }

    def doWithProbability(p : Double, f : () => Unit) : Boolean = {
        val testPassed = Random.nextDouble <= p
        if(testPassed) { f() }
        testPassed
    }
    
    /**
     * Compute the center of gravity of a set of points
     */
    def gravityCenterOfPoints(points: Array[Array[Double]]) = {
        val nbAttr = points(0).size
        var sumPoints = new Array[Double](nbAttr)
        (for(a <- points; i <- 0 until nbAttr) { sumPoints(i) += a(i) })
        sumPoints.map(a => a / points.size)
//        (points.reduceLeft((A, B) => for ((a, i) <- B.zipWithIndex) yield { a + A(i) })).map(a => a / points.size)
    }
    def gravityCenterOfPoints(points: HashMap[Int, Array[Double]]) = {
        val nbAttr = points.find(a => true).get._2.size
        var sumPoints = new Array[Double](nbAttr)
        (for((_, a) <- points; i <- 0 until nbAttr) { sumPoints(i) += a(i) })
        sumPoints.map(a => a / points.size)
    }

    def meanStdDev(v: Seq[Double]): (Double, Double) = {
        val mean = v.sum / v.length
        val stdDev = math.sqrt(v.foldLeft(0D)((B, A) => B + ((A - mean) * (A - mean))) / (v.length - 1))
        (mean, stdDev)
    }
    
    def main(args : Array[String]) : Unit = {
        val l = Array(70D,50D,20D,90D,1D,30D)
        val i = 0 until l.length
        println(inverseWeightedProbability(i, l))
    }
}
