/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package xps

import measures.GraphMeasures
import algo.Generator1
import scala.collection.mutable.HashMap
import scala.collection.mutable.LinkedHashMap
import utils.Timer

object Measure extends Enumeration {
    type Measure = Value
    val ClusteringCoefficient = Value("Clust. Coeff.")
    val RandomClusteringCoefficient = Value("Rand. Clust. Coeff.")
    val NbTriangles = Value("Nb. Triangles")
    val Density = Value("Density")
    val Modularity = Value("Modularity")

    val AvgDegree = Value("Avg. degree")
    val MaxDegree = Value("Max. degree")
    val AvgShortestPathLength = Value("Short. path")
    val Diameter = Value("Diameter")

    val AvgCommunitySize = Value("Avg. com. size")
    val MaxCommunitySize = Value("Max. com. size")
    val MinCommunitySize = Value("Min. com. size")
    val DiffCommunitySize = Value("Dif. com. size")
    val Runtime = Value("Runtime")

    val NbInternalEdges = Value("Wth. edges")
    val NbExternalEdges = Value("Btw. edges")
    val NbEdges = Value("Nb. edges")
    val InternalExternalEdgesRatio = Value("Wth.-Btw. R.")

    val BetweenInertia = Value("Btw inertia")
    val BetweenInertiaRatio = Value("Btw. inertia")
    val WithinInertia = Value("Wth inertia")
    val WithinInertiaRatio = Value("Wth. inertia")
    val RelativeConcentration = Value("Rel. concentration")

    val FractionOfHomophileEdges = Value("Obs. homo.")
    val ExpectedHomophileEdges = Value("Exp. homo.")

    val Homophily = Value("Homo.")
    val HomophilyBetween = Value("Homo. btw.")
    val HomophilyWithin = Value("Homo. wth.")
}

object Measures {
    import Measure._

     val distanceThreshold: Double = 3D

    def appendMeasures(measures: HashMap[Measure, List[Double]], m: GraphMeasures, g: Generator1): HashMap[Measure, List[Double]] = {
        
        for ((k, v) <- measures) yield { k -> (getValue(k, m, g) :: v) }
    }

    def getValue(measureType: Measure, m: GraphMeasures, g: Generator1) = {
        val value = measureType match {
            case Measure.AvgDegree => m.avgDegree
            case Measure.MaxDegree => m.maxDegree
            case Measure.AvgShortestPathLength => m.avgShortestPath

            case Measure.AvgCommunitySize => m.avgCommunitySize
            case Measure.MaxCommunitySize => m.maxCommunitySize
            case Measure.MinCommunitySize => m.minCommunitySize
            case Measure.DiffCommunitySize => (m.maxCommunitySize - m.minCommunitySize)

            case Measure.BetweenInertia => m.betweenInertia
            case Measure.BetweenInertiaRatio => m.betweenInertiaRatio
            case Measure.WithinInertia => m.withinInertia
            case Measure.WithinInertiaRatio => m.withinInertiaRatio
//            case Measure.RelativeConcentration => m.relativeConcentrationWithin

            case Measure.NbEdges => m.nbEdges
            case Measure.NbInternalEdges => m.nbInternalExternalEdges._1
            case Measure.NbExternalEdges => m.nbInternalExternalEdges._2
            case Measure.InternalExternalEdgesRatio => m.internalExternalEdgesRatio

            case Measure.Runtime => g.runTimeSec
            case Measure.Density => m.density
            case Measure.Diameter => m.diameter

            case Measure.Modularity => m.modularity
            case Measure.ClusteringCoefficient => m.clusteringCoefficient
            case Measure.RandomClusteringCoefficient => m.randomClusteringCoefficient
            case Measure.NbTriangles => m.nbTriangles
            
            
            
            // case Measure.Homophily => m.probaDifferentLink(distanceThreshold)

            case Measure.FractionOfHomophileEdges => m.homophilyObserved
            case Measure.ExpectedHomophileEdges => m.homophilyExpected

            case Measure.HomophilyBetween => m.homophilyBetween
            case Measure.HomophilyWithin => m.homophilyWithin

            case _ => {
                println("WARNING: Measure " + measureType + " not found")
                0D
            }
        }
        value
    }

    val measuresByType = LinkedHashMap(
            "Other Measures" -> List(
                Measure.AvgCommunitySize,
                Measure.MaxCommunitySize,
                Measure.MinCommunitySize,
                Measure.DiffCommunitySize,
                Measure.Runtime
                ),
            "Attribute Measures" -> List(
                Measure.BetweenInertia,
                Measure.BetweenInertiaRatio,
                Measure.WithinInertia,
                Measure.WithinInertiaRatio,
                Measure.RelativeConcentration,
                Measure.Homophily,
                Measure.HomophilyBetween,
                Measure.HomophilyWithin,
        		Measure.ExpectedHomophileEdges,
                Measure.FractionOfHomophileEdges
                ),
            "Structural Measures" -> List(
                Measure.AvgDegree,
                Measure.AvgShortestPathLength,
                Measure.ClusteringCoefficient,
                Measure.RandomClusteringCoefficient,
                Measure.Density,
                Measure.Diameter,
                Measure.InternalExternalEdgesRatio,
                Measure.Modularity,
                Measure.NbExternalEdges,
                Measure.NbInternalEdges,
                Measure.NbEdges,
                Measure.NbTriangles
                )
            )
            
    def meanStdDev(a: HashMap[Measure, List[Double]]): HashMap[Measure, (Double, Double)] = {
        for ((k, v) <- a) yield {
            val mean = v.sum / v.length
            val stdDev = math.sqrt(v.foldLeft(0D)((B, A) => B + ((A - mean) * (A - mean))) / (v.length - 1))
            k -> (mean, stdDev)
        }
    }
}