/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package xps

import java.io.PrintWriter
import scala.sys.process._
import java.awt.Color
import algo.Generator1
import algo.structures.CommunitiesInformation
import utils.Colors
import scala.collection.mutable.HashMap

object PlotDegreeDistrib {

    def plot(distribution : Array[(Int, Int)], fileName : String, useLogScale : Boolean = true) = {
        
        // Build the data file
        val dataFile = "/tmp/dd.data"
        val nbVertices = distribution.map(_._2).sum
        val out = new PrintWriter(dataFile)
        for((deg, nbVert) <- distribution) {
            out.println(deg + "\t" + nbVert.toDouble / nbVertices)
        }
        out.flush()
        out.close()
        
        val termCmd = s"set terminal pdf size 5,4 font 'Helvetica,20'"
        
//        val colors = Colors.getMatchingColors(measuresToPlot.size).map(c => f"#${c.getRed()}%02x${c.getGreen()}%02x${c.getBlue()}%02x")
        val gridCmd = "set grid y"
        val outCmd = s"set output '${fileName}'"
        val xLabelCmd = s"set xlabel 'Degree'"
        val yLabelCmd = s"set ylabel 'Frequency'"
        val plotCmd = s"plot '${dataFile}' using 1:2 t '' lt 7 lw 2 pt 7 ps 0.5 with points"
        val logScale = if(useLogScale) "set logscale xy" else "set nologscale xy"
        
        val script = List(termCmd, outCmd, gridCmd, xLabelCmd, yLabelCmd, logScale, plotCmd).mkString("\n")
        val outTmpFile = new PrintWriter("/tmp/plotDegDist.gp")
        outTmpFile.println(script)
        outTmpFile.flush
        outTmpFile.close()
        
        "gnuplot /tmp/plotDegDist.gp".!!
    }
    
    case class Distribution(name : String, values : Array[(Int, Int)], color : Color)
    
    def plot2(distributions : Seq[Distribution], fileName : String, useLogScale : Boolean = true) = {
        var maxDegree = 0
        // Build the data file
        for((d, i) <- distributions.zipWithIndex) {
	        val nbVertices = d.values.map(_._2).sum
	        val out = new PrintWriter(s"/tmp/dd${i}.data")
	        for((deg, nbVert) <- d.values) {
	            maxDegree = math.max(maxDegree, deg)
	            if(useLogScale) {
	                out.println(deg + "\t" + nbVert.toDouble / nbVertices)
	            } else {
	                out.println(math.log(deg) + "\t" + math.log(nbVert.toDouble / nbVertices))
	            }
	            
	        }
	        out.flush()
	        out.close()
        }
        val termCmd = s"set terminal postscript eps enhanced size 5,4 font 'Helvetica,20'"
        
        val gridCmd = "set grid ytics"
        val outCmd = s"set output '${fileName}'"
        val xRange = if(useLogScale) { s"set xrange [1:${maxDegree}]" } else { s"set xrange [0:${math.log(maxDegree)}]" }
        val yLabelLog = if(useLogScale) { "set format y '10^{%L}'" } else { "" }
        val xLabelLog = if(useLogScale) { "set format x '10^{%L}'" } else { "" }
        val yLabelCmd = if(useLogScale) { s"set ylabel 'Frequency'" } else { "set ylabel 'log(Frequency)'" }
        val xLabelCmd = if(useLogScale) { s"set xlabel 'Degree'" } else { "set xlabel 'log(Degree)'" }
        val logScale = if(useLogScale) "set logscale xy" else "set nologscale xy"
    	val lineStyle = "lt 7 lw 2 pt 7 ps 0.5"
        
	    val plotCmd = (for((d, i) <- distributions.zipWithIndex) yield {
	        val color = f"#${d.color.getRed()}%02x${d.color.getGreen()}%02x${d.color.getBlue()}%02x"
	        s"'/tmp/dd${i}.data' using 1:2 t '${d.name}' lt 7 lw 2 lc rgb '${color}' pt 7 ps 1 with points"
	    }).mkString("plot ", ", \\\n", "")
        
        val script = List(termCmd, outCmd, gridCmd, xLabelCmd, yLabelCmd, xRange, logScale, yLabelLog, xLabelLog, plotCmd).mkString("\n")
        val outTmpFile = new PrintWriter("/tmp/plotDegDist.gp")
        outTmpFile.println(script)
        outTmpFile.flush
        outTmpFile.close()
        
        "gnuplot /tmp/plotDegDist.gp".!!
    }
    
    
    def plotGen(gen : Generator1, fileName : String, useLogScale : Boolean = true) {
        val comInfo = new CommunitiesInformation(gen)
        val colors = Colors.getMatchingColors(gen.k)
        
        val degDistMap = HashMap[Int, Int]()
        val degrees = for (v1 <- 0 until gen.nbVertices) {
            degDistMap(gen.graph.degree(v1)) = degDistMap.getOrElse(gen.graph.degree(v1), 0) + 1
        }
        val wholeGraphDistrib = Distribution("Graph", degDistMap.toArray.sortBy(a => a._1), Color.black)
        
//        val data = wholeGraphDistrib :: (for(i <- 0 until gen.k) yield {
//            Distribution("Community " + (i + 1), comInfo.communitiesInformation(i).degreeDistribution, colors(i))
//        }).toList
//        val data = (for(i <- 0 until gen.k) yield {
//            Distribution("Community " + (i + 1), comInfo.communitiesInformation(i).degreeDistribution, colors(i))
//        }).toList

        val data = List(wholeGraphDistrib)
        plot2(data, fileName, useLogScale)
    }
    
    def main(args: Array[String]): Unit = {
        plot(Array((1,10),(4,3), (7, 8)), "test.pdf")
        
        val d1 = Distribution("Graph", Array((1,10),(4,3), (7, 8)), Color.black)
        val d2 = Distribution("Com1", Array((1,5),(5,4), (7, 6)), Color.red)
        
        plot2(List(d1, d2), "test2.pdf")
    }

}