/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package xps

import scala.collection.mutable.HashMap
import algo.Generator1
import measures.GraphMeasures
import scala.compat.Platform
import java.text.SimpleDateFormat
import java.io.File
import java.io.PrintWriter
import scala.sys.process._
import scala.collection.mutable.LinkedHashMap
import utils.PdfOutput
import algo.Output
import utils.Latex
import utils.LatexHeader
import utils.LatexPackage
import utils.Colors

object Parameters {
    var nbVertices = RangeWithDefault(1000 to 10000 by 1000, 1000, "nbVertices", "")
    var nbVerticesMTE = RangeWithDefault(1000 to 10000 by 1000, 1000, "nbVerticesMTE", "")
    // var nbVertices = RangeWithDefault(100 to 100 by 100, 100, "nbVertices", "Number of vertices")
    var nbAttributes = RangeWithDefault(3 to 9 by 1, 2, "nbAttributes", "Number of attributes")
//    var mean = RangeWithDefault(30 to 30 by 1, 30, "mean", "Attribute mean")
//    var stdDev = RangeWithDefault(1 to 61 by 6, 7, "stdDev", "Standard deviation")
    var stdDevX = RangeWithDefault(1 to 15 by 2, 7, "stdDevX", "Standard deviation X")
    var stdDevY = RangeWithDefault(1 to 15 by 2, 1, "stdDevY", "Standard deviation Y")
    // var theta = RangeWithDefault(0 to 100 by 5, 5, "theta")
    var theta = RangeWithDefault(0 to 100 by 10, 0, "theta", "")
    var nbRepresentents = RangeWithDefault(10 to 100 by 10, 100, "nbRepresentatives", "Number of representatives")
    var k = RangeWithDefault(2 to 10 by 2, 3, "k", "K")
    var maxInternalEdges = RangeWithDefault(2 to 30 by 2, 6, "EmaxWth", "")
    var maxExternalEdges = RangeWithDefault(0 to 20 by 2, 3, "EmaxBtw", "")
    var minTotalEdges = RangeWithDefault(0 to 10000 by 1000, 0, "MTE", "")
    
    var nbRepPerc = RangeWithDefault(10 to 100 by 10, 0, "nbRepPerc", "Percentage of representatives")
}

object Xp  {
    val nbSeeds = 10
    
    val measures = List(
//                Measure.Diameter,
//                Measure.AvgShortestPathLength,
//                Measure.AvgDegree,
//                Measure.BetweenInertiaRatio,
                Measure.ClusteringCoefficient,
                Measure.RandomClusteringCoefficient,
//                Measure.HomophilyBetween,
//                Measure.HomophilyWithin,
                Measure.Modularity
//                Measure.NbExternalEdges,
//                Measure.NbInternalEdges,
//                Measure.NbEdges,
//                Measure.WithinInertiaRatio,
//                Measure.ExpectedHomophileEdges,
//                Measure.FractionOfHomophileEdges
                )
    val ranges = List(
            	Parameters.nbVertices,
            	Parameters.nbVerticesMTE
//            	Parameters.stdDevY,
//            	Parameters.minTotalEdges
//            	Parameters.nbRepresentents,
//            	Parameters.nbRepPerc,
//            	Parameters.maxExternalEdges
//            	Parameters.maxInternalEdges
//            	Parameters.theta
//            	Parameters.k
    )
    
    val allRanges = List(Parameters.nbVertices, Parameters.nbRepresentents, Parameters.k, Parameters.maxExternalEdges, Parameters.maxInternalEdges, 
            Parameters.nbAttributes, Parameters.stdDevX, Parameters.stdDevY, Parameters.theta, Parameters.minTotalEdges)
    
    // Out Path
    val dateFormat = new SimpleDateFormat("yyyy-MM-dd_HH:mm:ss:S")
    // val baseDir = s"${new File(".").getCanonicalPath()}/xp/${dateFormat.format(Platform.currentTime)}/"
    val baseDir = new File(".").getCanonicalPath() + "/xp/varying/"
    new File(baseDir).mkdirs()
    val allResultsOut = new PrintWriter(baseDir + "complete")
    
    // val latexFile = new PrintWriter(s"${baseDir}results.tex")
    val latex = new Latex(s"${baseDir}results.tex", Option(LatexHeader("Results", List(
            LatexPackage("graphics"), LatexPackage("graphicx"), LatexPackage("caption"), LatexPackage("booktabs"), LatexPackage("subfigure")))))
    
    /**
     * Entry point
     */
    def main(args: Array[String]) : Unit = {
        latex.printSection("Default values")
        latex.printItemize(for(range <- allRanges) yield { range.label + ": " + range.default })
        // runDefaultParameters()
        latex.print("Moyenne sur " + nbSeeds + " graphes")
        for(range <- ranges) { runWithinRange(range) }
        latex.printEnding()
        latex.compile
    }
    
    
    def formatDouble(x : Double) = if(x == Math.floor(x)) { "" + x.toInt } else { f"${x}%.2f" }
    
    def runDefaultParameters() = {
        var measuresAggregator = HashMap(measures.map(m => (m -> List[Double]())) : _*)
        
        for(seed <- 0 until nbSeeds) {
            val k = Parameters.k.default
            val nbVertices = Parameters.nbVertices.default 
            val nbAttributes = Parameters.nbAttributes.default 
            val stdDevX = Parameters.stdDevX.default 
            val stdDevY = Parameters.stdDevY.default
            val theta = Parameters.theta.default
            val nbRepresentents = Parameters.nbRepresentents.default
            val maxInternalEdges = Parameters.maxInternalEdges.default
            val maxExternalEdges = Parameters.maxExternalEdges.default
            val minTotalEdges = Parameters.minTotalEdges.default
	        val devs = Array(stdDevX.toDouble, stdDevY.toDouble)
	        val gen = new Generator1(k, nbVertices, devs, theta.toDouble / 100, nbRepresentents, maxInternalEdges, maxExternalEdges, minTotalEdges, seed = Option(seed), probaCloseTriangle = 0)
            val m = new GraphMeasures(gen.graph, devs)
            measuresAggregator = Measures.appendMeasures(measuresAggregator, m, gen)
        }
        val aggregatedMeasures = Measures.meanStdDev(measuresAggregator)
        
        
        // Output the results
        latex.printSection("Results for default values")
        latex.print("\\begin{center}")
        val nbColumns = ((nbSeeds - 1) / 10) + 1
        latex.print((for(_ <- 0 until nbColumns) yield { "c" }).mkString("\\begin{tabular}{l", "", "cc}"))
        latex.print("\\hline{}")
        latex.print((for(i <- 0 until nbColumns) yield { i * 10 }).mkString("Seed & ", " & ", " & \\textbf{Mean} & Dev \\\\"))
        latex.print("\\hline{}")
        for((k, v) <- measuresAggregator.toList.sortBy(_._1)) {
            val row = (for(i <- 0 until nbSeeds; if(i % 10 == 0)) yield { formatDouble(v(i)) }).toList ::: List(
                    "\\textbf{" + formatDouble(aggregatedMeasures(k)._1) + "}", formatDouble(aggregatedMeasures(k)._2))
            latex.print(row.mkString(k.toString + " & ", " & " , "\\\\"))
        }
        latex.print("\\hline{}")
        latex.print("\\end{tabular}")
        latex.print("\\end{center}")
    }
    
    def runWithinRange(range : RangeWithDefault) = {
        val rangeName = range.name
        val header = measures.map(a => List(a.toString + "_mean", a.toString + "_dev")).flatten
        println((range.name :: header).mkString("# ", "\t", ""))
        val path = s"${baseDir}${range.name}_${range.range.start}_${range.range.end}_${range.range.step}/"
        new File(path).mkdirs()
        new File(s"${path}eps").mkdirs()
        new File(s"${path}graph").mkdirs()
        
        // Build data file
        val dataFile = new PrintWriter(path + "data.dat")
        range.range.foreach { v =>
            val aggregatedMeasures = range match {
                case x if(x == Parameters.nbVertices) => runXp(v, path, nbVertices = v)
                case x if(x == Parameters.nbAttributes) => runXp(v, path, nbAttributes = v)
                case x if(x == Parameters.stdDevX) => runXp(v, path, stdDevX = v)
                case x if(x == Parameters.stdDevY) => runXp(v, path, stdDevY = v)
                case x if(x == Parameters.theta) => runXp(v, path, theta = v)
                case x if(x == Parameters.nbRepresentents) => runXp(v, path, nbRepresentents = v)
                case x if(x == Parameters.nbRepPerc) => {
                    val maxNbRep = Parameters.nbVertices.default / Parameters.k.default
                    val nbRep = (maxNbRep * v.toDouble / 100).toInt
                    runXp(v, path, nbRepresentents = nbRep)
                }
                case x if(x == Parameters.k) => runXp(v, path, k = v, stdDevX = v, stdDevY = v)
                case x if(x == Parameters.minTotalEdges) => runXp(v, path, maxTotalEdges = v)
                case x if(x == Parameters.maxInternalEdges) => runXp(v, path, maxInternalEdges = v)
                case x if(x == Parameters.maxExternalEdges) => {
                    runXp(v, path, maxExternalEdges = v, maxInternalEdges = 30)
                }
                case x if(x == Parameters.nbVerticesMTE) => { runXp(v, path, nbVertices = v, maxTotalEdges = v * 10) }
            }
            if(range == Parameters.theta) {
                dataFile.println(((v.toDouble / 100) :: measures.map(a => List(aggregatedMeasures(a)._1, aggregatedMeasures(a)._2)).flatten).mkString("\t"))
            } else {
                dataFile.println((v :: measures.map(a => List(aggregatedMeasures(a)._1, aggregatedMeasures(a)._2)).flatten).mkString("\t"))
            }
                
            
        }
        dataFile.flush()
        dataFile.close()
        
        buildGnuPlotScript2(range.label, "Inertia", List(Measure.BetweenInertiaRatio, Measure.WithinInertiaRatio), path, Option((0,1)))
        buildGnuPlotScript2(range.label, "Avg. Clustering Coefficient", List(Measure.ClusteringCoefficient, Measure.RandomClusteringCoefficient), path, Option((0,0.2)))
        // buildGnuPlotScript2(range.label, "Homophily", List(Measure.ExpectedHomophileEdges, Measure.FractionOfHomophileEdges, Measure.HomophilyBetween, Measure.HomophilyWithin), path, Option((0,1)))
        buildGnuPlotScript2(range.label, "Homophily", List(Measure.ExpectedHomophileEdges, Measure.FractionOfHomophileEdges), path, Option((0,1)))
        buildGnuPlotScript(range.label, path, HashMap(Measure.Modularity -> (0.5, 0.67)))
//        buildLatex(range, path)
    }
    
    def runXp(
            paramValue : Int,
            path : String, 
            k : Int = Parameters.k.default,
            nbVertices : Int = Parameters.nbVertices.default, 
            nbAttributes : Int = Parameters.nbAttributes.default, 
            stdDevX: Int = Parameters.stdDevX.default, 
            stdDevY: Int = Parameters.stdDevY.default,
            theta : Int = Parameters.theta.default,
            nbRepresentents : Int = Parameters.nbRepresentents.default,
            maxInternalEdges : Int = Parameters.maxInternalEdges.default,
            maxExternalEdges : Int = Parameters.maxExternalEdges.default,
            maxTotalEdges : Int = Parameters.minTotalEdges.default) = {
        val devs = Array[Double](stdDevX.toDouble, stdDevY.toDouble)
        
        val genParams = List("# Seed", "K", "n", "nbAttr", "mean", "stdDev", "theta", "nbRep", "maxInt", "maxExt")
        // val measureNames = List("CC.", "AvgDeg", "AvgPath", "Diam.", "Mod.", "Time")
        
        var measuresAggregator = HashMap(measures.map(m => (m -> List[Double]())) : _*)
//        allResultsOut.println((genParams ::: measures.map(_.toString)).mkString("\t"))
        
        for(seed <- 0 until nbSeeds) {
            val gen = new Generator1(k, nbVertices, devs, theta.toDouble / 100, nbRepresentents, maxInternalEdges, maxExternalEdges, maxTotalEdges, seed = Option(seed), probaCloseTriangle = 0)
            val m = new GraphMeasures(gen.graph, devs)
            measuresAggregator = Measures.appendMeasures(measuresAggregator, m, gen)
            
            
//            val genParamsValues = List(seed, k, nbVertices, nbAttributes, mean, stdDev, theta, nbRepresentents, maxInternalEdges, maxExternalEdges)
            // val measuresValues = List(f"${m.clusteringCoefficient}%.4f", m.avgDegree, f"${m.avgShortestPath}%.4f", m.diameter, f"${m.modularity}%.4f", gen.runTimeSec)
            // allResultsOut.println((genParamsValues ::: measuresValues).mkString("\t"))
            
            // Pdf output
            /*
			if(seed < 3) {
			    val panel = new GraphViewPanel()
				panel.setGraph(gen)
				PdfOutput.createPdf(panel, s"${path}graph/attr-v-${paramValue}-seed-${seed}.pdf")
				
				val outP = new Output(gen)
			    outP.toJungGraph(s"${path}graph/layout-v-${paramValue}-seed-${seed}.pdf")
			}
			*/
        }
        println("Pour le shortest path en fonction du nombre de sommets, ajouter courbes alpha * log(nbVertices) en esperant que cela fit bien !!")
        val aggregatedMeasures = Measures.meanStdDev(measuresAggregator)
        println((paramValue :: measures.map(a => List(aggregatedMeasures(a)._1, aggregatedMeasures(a)._2)).flatten).mkString("\t"))
        aggregatedMeasures
    }
    
    def buildGnuPlotScript(xLabel : String, path : String, yAxisRanges : HashMap[Measure.Value, (Double, Double)] = HashMap()) = {
        for((m, i) <- measures.zipWithIndex) {
            val measureName = m.toString().replaceAllLiterally(" ", "_").replaceAllLiterally(".", "")
            val gnuPlotScriptPath = s"${path}${measureName}.gp"
            val dataFile = s"${path}data.dat"
            val termCmd = s"set terminal pdf monochrome size 5,4 font 'Helvetica,25'"
            
            val gridCmd = "set grid y"
            val outCmd = s"set output '${path}eps/${measureName}.pdf'"
            val xLabelCmd = s"set xlabel '${xLabel}'"
            val yLabelCmd = s"set ylabel '${m.toString()}'"
//            val offsetCmd = "set offsets 1, 1, 0, 0"
//            val scaleCmd = "set autoscale fix"
            val scaleYCmd = if(yAxisRanges.contains(m)) { s"set yrange [${yAxisRanges(m)._1}:${yAxisRanges(m)._2}]" } else { "set yrange [*:*]" }
	        val scaleXCmd = "set xrange [*:*]"
	        val plotCmd = s"plot '${dataFile}' using 1:${i * 2 + 2}:${i * 2 + 3} t '' lt -1 lw 2 with errorbars, " + 
            		s"'' using 1:${i * 2 + 2} t '' ls -1 lw 4 with linespoint"
            val script = List(termCmd, outCmd, gridCmd, xLabelCmd, yLabelCmd, scaleYCmd, scaleXCmd, plotCmd).mkString("\n")
            val outTmpFile = new PrintWriter(gnuPlotScriptPath)
            outTmpFile.println(script)
            outTmpFile.flush
            outTmpFile.close()
            
            s"gnuplot ${gnuPlotScriptPath}".!!
            
        }
    }
    
    def buildGnuPlotScript2(xLabel : String, yLabel : String, measuresToPlot : List[Measure.Value], path : String, yAxisRange : Option[(Double, Double)] = None) = {
        val measureName = yLabel.replaceAllLiterally(" ", "_").replaceAllLiterally(".", "")
        val gnuPlotScriptPath = s"${path}${measureName}.gp"
        val dataFile = s"${path}data.dat"
        val termCmd = s"set terminal pdf monochrome dashed size 5,4 font 'Helvetica,25'"
//        val termCmd = s"set terminal postscript eps enhanced solid"
        
        val colors = Colors.getMatchingColors(measuresToPlot.size).map(c => f"#${c.getRed()}%02x${c.getGreen()}%02x${c.getBlue()}%02x")
        val gridCmd = "set grid y"
        val outCmd = s"set output '${path}eps/${measureName}.pdf'"
        val xLabelCmd = s"set xlabel '${xLabel}'"
        val yLabelCmd = s"set ylabel '${yLabel}'"
        // val offsetCmd = "set offsets 1, 1, 0, 0"
        val offsetCmd = ""
        val legendPosition = List("set key out horiz", "set key center top", "set key box").mkString("\n")
//        val offsetCmd = ""
        val scaleYCmd = if(yAxisRange.isDefined) { s"set yrange [${yAxisRange.get._1}:${yAxisRange.get._2}]" } else { "set autoscale fix" }
        val scaleXCmd = "set xrange [*:*]"
        val plotCmd = (for((m, ii) <- measuresToPlot.zipWithIndex) yield {
            val i = measures.indexOf(m)
            s"'${dataFile}' using 1:${i * 2 + 2}:${i * 2 + 3} lt -1 lw 2 with errorbars t '' , \\\n" + 
        		s"'' using 1:${i * 2 + 2} lt ${ii + 1} lw 4 with lines t '${m.toString()}'"
        }).mkString("plot ", ", \\\n", "")
        
        val script = List(termCmd, outCmd, gridCmd, xLabelCmd, yLabelCmd, legendPosition, scaleYCmd, scaleXCmd, offsetCmd, plotCmd).mkString("\n")
        val outTmpFile = new PrintWriter(gnuPlotScriptPath)
        outTmpFile.println(script)
        outTmpFile.flush
        outTmpFile.close()
        
        s"gnuplot ${gnuPlotScriptPath}".!!
    }
    
    /*
    def buildLatex(range : RangeWithDefault, path : String) = {
//        val outLatex = new PrintWriter()
        Latex.use(s"${path}include.tex", true) { outLatex =>
            outLatex.print(s"\\newpage")
	        outLatex.printSection("Parameter " + range.label)
	        outLatex.print(s"${range.label} varying between ${range.range.start} and ${range.range.end} by ${range.range.step}")
	        for((k, v) <- Measures.measuresByType) {
	            outLatex.printSubsection(k)
	            var i = 0
	            if(k == "Attribute Measures") {
	                outLatex.print(s"\\includegraphics[width=.5\\linewidth]{${path}eps/Homophily.pdf}")
	                outLatex.print(s"\\includegraphics[width=.5\\linewidth]{${path}eps/Inertia.pdf}")
	            }
	            if(k == "Structural Measures") {
	                outLatex.print(s"\\includegraphics[width=.5\\linewidth]{${path}eps/Clustering_Coeff.pdf}")
	                for(m <- v; if measures.contains(m)) {
		                val measureName = m.toString().replaceAllLiterally(" ", "_").replaceAllLiterally(".", "")
			            outLatex.print(s"\\includegraphics[width=.5\\linewidth]{${path}eps/${measureName}.pdf}")
			            if(i % 2 == 1) { outLatex.newLine }
			            i += 1
			        }
	            }
	            
	        }
        }
        
//        outLatex.println(s"\\subsection{Graph}")
//        for(paramValue <- range.range) {
//            outLatex.println(s"${range.label} = ${paramValue}")
//            outLatex.println("\\\\")
//            for(seed <- 0 until 1) {
//                outLatex.println(s"\\includegraphics[width=.45\\linewidth]{${path}graph/attr-v-${paramValue}-seed-${seed}.pdf}")
//                outLatex.println(s"\\includegraphics[width=.45\\linewidth]{${path}graph/layout-v-${paramValue}-seed-${seed}.pdf}")
//            }
//            outLatex.println("\\\\")
//        }
        latex.print(s"\\input{${path}include.tex}")
    }
    */
}

case class RangeWithDefault(var range : Range, var default : Int, var name : String, val label : String)

