/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package xps

import java.io.File
import java.io.PrintWriter
import algo.Generator1
import utils.FileUtils
import utils.Stat
import algo.GeneratorDyn
import algo.MicroProba
import measures.GraphMeasures
import algo.structures.DynamicGraph
import utils.Timer
import scala.collection.mutable.Stack
import algo.structures.Graph


object XpDynRuntime {
    case class RuntimeRange(label : String, name : String, ls : Int)
    
    val addVertices = RuntimeRange("addVertices", "R_{AddVert}", 1)
    val addWthEdges = RuntimeRange("addWthEdges", "R_{AddWthEdge}", 2)
    val addBtwEdges = RuntimeRange("addBtwEdges", "R_{AddBtwEdge}", 3)
    val updateAttributes = RuntimeRange("updateAttributes", "R_{UpdAttr}", 4)
    val removeVertices = RuntimeRange("removeVertices", "R_{RemoveVert}", 5)
    val removeWthEdges = RuntimeRange("removeWthEdges", "R_{RemoveWthEdge}", 6)
    val removeBtwEdges = RuntimeRange("removeBtwEdges", "R_{RemoveBtwEdge}", 7)
    
    val probaMerge = RuntimeRange("probaMerge", "{/Symbol q}_6 (merge)", 1)
    val probaSplit = RuntimeRange("probaSplit", "{/Symbol q}_7 (split)", 2)
    val probaChange = RuntimeRange("probaChange", "{/Symbol q}_8 (migrate)", 3)
    
     val microOperations = List(addVertices, addWthEdges, addBtwEdges, updateAttributes, removeVertices, removeWthEdges, removeBtwEdges)
     val macroOperations = List(probaMerge, probaSplit, probaChange)
    val ranges = microOperations ::: macroOperations
    val inertiaMeasure = MeasureXp("inertiaWithin", "Inertia Within", "", measures => Stat.meanStdDev(measures.map(m => m.withinInertiaRatio.toDouble)), Option("0:1.2"))
    val clusteringCoeffMeasure = MeasureXp("clusteringCoeff", "Avg. Clustering Coefficient", "", measures => Stat.meanStdDev(measures.map(m => m.clusteringCoefficient.toDouble)), Option("0:1.1"))
    val modularityMeasure = MeasureXp("modularity", "Modularity", "", measures => Stat.meanStdDev(measures.map(m => m.modularity.toDouble)), Option("0:1.4"))
    val homophilyMeasure = MeasureXp("homophily", "Observed Homophily", "", measures => Stat.meanStdDev(measures.map(m => m.homophilyObserved.toDouble)), Option("0:1.2"))
    val homophilyExpectedMeasure = MeasureXp("homophilyExpected", "Expected Homophily", "", measures => Stat.meanStdDev(measures.map(m => m.homophilyExpected.toDouble)), Option("0:1"))
    val diameterMeasure = MeasureXp("diameter", "Diameter", "", measures => Stat.meanStdDev(measures.map(m => m.diameter.toDouble)), Option("0:15"))
    val shortestPathMeasure = MeasureXp("shortestPath", "Avg. Shortest Path Length", "", measures => Stat.meanStdDev(measures.map(m => m.avgShortestPath.toDouble)), Option("0:8"))
    
    val nbCCMeasure = MeasureXp("nbCC", "Nb. Connected Components", "", measures => Stat.meanStdDev(measures.map(m => m.connectedComponents.length.toDouble)), Option("0:14"))
    val nbVerticesMeasure = MeasureXp("nbVertices", "Nb. Vertices", "", measures => Stat.meanStdDev(measures.map(m => m.nbVertices.toDouble)), Option("0:*"))
    val nbEdgesMeasure = MeasureXp("nbEdges", "Nb. Edges", "", measures => Stat.meanStdDev(measures.map(m => m.nbEdges.toDouble)), Option("0:*"))
    val nbWthEdgesMeasure = MeasureXp("nbWthEdges", "Nb. Wth. Edges", "", measures => Stat.meanStdDev(measures.map(m => m.nbWithinEdges.toDouble)), Option("0:*"))
    val nbBtwEdgesMeasure = MeasureXp("nbBtwEdges", "Nb. Btw. Edges", "", measures => Stat.meanStdDev(measures.map(m => m.nbBetweenEdges.toDouble)), Option("0:*"))
    
    val maxComponentSize = MeasureXp("maxCompSize", "Size of the largest component", "set logscale x", measures => Stat.meanStdDev(measures.map(m => m.connectedComponents.map(_.size).max.toDouble)), Option("0:300000"), "set nologscale x")
    val homogeneityDifference = MeasureXp("homogeneityDiff", "Homogeneity", "", measures => Stat.meanStdDev(measures.map(m => m.homophilyObserved - m.homophilyExpected)), Option("-1:1"))
    
    val mesuresToConsider = List(clusteringCoeffMeasure, inertiaMeasure, modularityMeasure, homophilyMeasure, homophilyExpectedMeasure, diameterMeasure, shortestPathMeasure, 
            nbCCMeasure, nbVerticesMeasure, nbEdgesMeasure, nbWthEdgesMeasure, nbBtwEdgesMeasure, maxComponentSize, homogeneityDifference)
    
    // Nb timestamps = 10 throws a OutOfMemory exception with 8Gb of memory
    val nbTimestamps = 10
    val nbRun = 5

    def main(args: Array[String]) = {
        val folder = "runtimeDyn"
        new File(s"xp/${folder}/").mkdirs()
        
        FileUtils.using(s"xp/${folder}/buildPlots.gp") { gp =>
            gp.println(headerGnuPlot)
            gp.println((for((m, i) <- mesuresToConsider.zipWithIndex) yield { m.toGnuplot(i * 2 + 4) }).mkString(""))
        }
        System.exit(0)
        // JVM startup
        jvmStartup()
        
        // Range changes updates
        for(range <- ranges) {
            println(range.label)
            FileUtils.using(s"xp/${folder}/" + range.label + "Results.tex") { latex =>
            	FileUtils.using(s"xp/${folder}/" + range.label) { w =>
	                FileUtils.using(s"xp/${folder}/" + range.label + "_overTime") { w2 =>
	                    latex.println("\\begin{table}")
	                    latex.println("\\begin{tabular}{ccccccccc}")
	                    latex.println("\\toprule")
	                    latex.println(" & Runtime & CC. & W.Inertia & Mod. & Homo. & Diam. & ShortPath \\\\")
	                    latex.println("\\midrule")
			            for (valueInt <- 0 to 10) {
			                val v = valueInt.toDouble / 10 
			                val microProba = MicroProba(probaMicro = 1D, addVertices = 0.0D, removeVertices = 0D, updateAttributes = 0D, removeInternalEdges = 0D, addInternalEdges = 0D,
			                		removeExternalEdges = 0D, addExternalEdges = 0D, probaMerge = 0D, probaSplit = 0D, probaChange = 0D)
			                range match {
			                    case `addVertices` => microProba.addVertices = v
			                    case `removeVertices` => microProba.removeVertices = v
			                    case `updateAttributes` => microProba.updateAttributes = v
			                    
			                    case `removeWthEdges` => microProba.removeInternalEdges = v
			                    case `addWthEdges` => microProba.addInternalEdges = v
			                    case `removeBtwEdges` => microProba.removeExternalEdges = v
			                    case `addBtwEdges` => microProba.addExternalEdges = v
			                    
			                    case `probaMerge` => microProba.probaMerge = v
			                    case `probaSplit` => microProba.probaSplit = v
			                    case `probaChange` => microProba.probaChange = v
			                }
			                
			                val nbVerticesAtStart = 500
			                val firstTimestampGraphs = Stack[Graph]()
			                
			                
					        val results : Seq[(Double, DynamicGraph)] = (for (seed <- 0 until nbRun) yield {
					            val startTime = System.currentTimeMillis()
					            val gen = new GeneratorDyn(k = 10, nbVertices = nbVerticesAtStart, dev = Array(1D, 1D), theta = 0, probaCloseTriangle = 0, nbRep = 10,
					                avgWthDegree = 6, avgBtwDegree = 3, minimumTotalEdges = 5000, seed = Option(seed), nbTimestamps = nbTimestamps, useAverageDegree = false, 
					                microProba = microProba)
					            
					            val dynGraph = gen.generateDynamicGraph(computeDynamicMeasure = false)
					            val runTime = ((System.currentTimeMillis() - startTime).toDouble / 1000)
					            (runTime, dynGraph)
					        })
					        val runtimes = results.map(_._1)
					        val measures = results.map( x => new GraphMeasures(x._2.graphs(x._2.nbTimestamps - 1), Array(1D, 1D)))
					        val measuresTimestamp5 = results.map( x => new GraphMeasures(x._2.graphs(4), Array(1D, 1D)))
					        
					        val measuresAtTimestamp5 = Set(nbBtwEdgesMeasure, nbWthEdgesMeasure, shortestPathMeasure, diameterMeasure, nbCCMeasure, nbEdgesMeasure,
					                homophilyMeasure, homophilyExpectedMeasure)
					        var measuresValues = List[Double]()
//					        Timer.clearAllTimers
					        for((m, i) <- mesuresToConsider.zipWithIndex) {
//					            println("Computing " + m.label)
					            val (mean, dev) = if(measuresAtTimestamp5.contains(m) && range == addVertices) {m.f(measuresTimestamp5)} else {m.f(measures)}
					            measuresValues = measuresValues :+ mean 
					            measuresValues = measuresValues :+ dev 
					        }
//			                Timer.printTimers
					        
					        val (meanRuntime, devRuntime) = Stat.meanStdDev(runtimes)
					        val res = List(v, meanRuntime, devRuntime) ::: measuresValues
					        
					        w.println(res.mkString("\t"))
					        println(res.mkString("\t"))
					        
					        // Compute measure evolution when the parameter value is 0.5
					        if(valueInt == 5) {
					            val measuresAllTimestamps = results.map(_._2).map(_.graphs.map(g => new GraphMeasures(g, Array(1D, 1D))))
					            for(t <- 0 until nbTimestamps) yield {
					                val m2 = measuresAllTimestamps.map(m => m(t))
					                var measuresValues = List[Double]()
						            for((m, i) <- mesuresToConsider.zipWithIndex) {
							            val (mean, dev) = m.f(m2)
							            measuresValues = measuresValues :+ mean 
							            measuresValues = measuresValues :+ dev 
							        }
					                val res2 = List(t + 1, 0, 0) ::: measuresValues
					                w2.println(res2.mkString("\t"))
					                println(res2.mkString("\t"))
					            }
					            
					        }
			                
					        //////////////////////////////////////////////////////////////////////////////////////////////////////
					        // Measures for the first timestamp
			                // Only do it for the first range, the remaining is identical
					        /*
			                if(range == ranges(0)) {
			                    FileUtils.using(s"xp/${folder}/firstTimestamp.tex") { ftTex =>
				                    ftTex.println("\\begin{table}")
				                    ftTex.println("\\begin{tabular}{lccccccccc}")
				                    ftTex.println("\\toprule")
				                    ftTex.println("Seed & CC. & W.Inertia & Mod. & Homo. & Diam. & ShortPath & NbEdges & Avg.Deg \\\\")
				                    ftTex.println("\\midrule")
							        for((res, seed) <- results.zipWithIndex) {
							            val mm = new GraphMeasures(res._2.graphs(0), Array(1D, 1D))
							            val row = seed.toString :: List(mm.clusteringCoefficient, mm.withinInertiaRatio, mm.modularity, mm.homophilyObserved, mm.diameter, mm.avgShortestPath, mm.nbEdges, mm.avgDegree).map(x => f"${x}%.3f".replaceAllLiterally(",", "."))
							            ftTex.println(row.mkString("", " & ", "\\\\"))
							        }
				                    ftTex.println("\\midrule")
							        val measuresFT = results.map( x => new GraphMeasures(x._2.graphs(0), Array(1D, 1D)))
							        val (clusteringCoefficientFT, clusteringCoefficientDevFT) = Stat.meanStdDev(measuresFT.map(m => m.clusteringCoefficient))
							        val (inertiaWithinFT, inertiaWithinDevFT) = Stat.meanStdDev(measuresFT.map(m => m.withinInertiaRatio))
							        val (modularityFT, modularityDevFT) = Stat.meanStdDev(measuresFT.map(m => m.modularity))
							        val (homophilyObservedFT, homophilyObservedDevFT) = Stat.meanStdDev(measuresFT.map(m => m.homophilyObserved))
							        val (avgShortestPathFT, avgShortestPathDevFT) = Stat.meanStdDev(measuresFT.map(m => m.avgShortestPath))
							        val (diameterFT, diameterDevFT) = Stat.meanStdDev(measuresFT.map(m => m.diameter.toDouble))
							        val (nbEdgesFT, nbEdgesDevFT) = Stat.meanStdDev(measuresFT.map(m => m.nbEdges.toDouble))
							        val (avgDegreeFT, avgDegreeDevFT) = Stat.meanStdDev(measuresFT.map(m => m.avgDegree.toDouble))
							        
							        val (meanRuntimeFT, devRuntimeFT) = Stat.meanStdDev(runtimes)
							        
							        val rowMean = "Mean" :: List(clusteringCoefficientFT, inertiaWithinFT, modularityFT, homophilyObservedFT, diameterFT, avgShortestPathFT, nbEdgesFT, avgDegreeFT).map(x => f"${x}%.3f".replaceAllLiterally(",", "."))
							        val rowDev = "Dev." :: List(clusteringCoefficientDevFT, inertiaWithinDevFT, modularityDevFT, homophilyObservedDevFT, diameterDevFT, avgShortestPathDevFT, nbEdgesDevFT, avgDegreeDevFT).map(x => f"${x}%.3f".replaceAllLiterally(",", "."))
							        ftTex.println("\\midrule")
							        ftTex.println(rowMean.mkString("", " & ", "\\\\"))
							        ftTex.println(rowDev.mkString("", " & ", "\\\\"))
							        ftTex.println("\\bottomrule")
				                    ftTex.println("\\end{tabular}")
				                    ftTex.println("\\caption{Measures for the initial graphs}")
				                    ftTex.println("\\end{table}")
							        
				                }
			                }
			                */
					        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
					        
			            }
	                    latex.println("\\bottomrule")
	                    latex.println("\\end{tabular}")
	                    latex.println("\\caption{" + range.label + "}")
	                    latex.println("\\end{table}")
		            }
            	}
            }
        }
    }
    
    
    def jvmStartup() = {
        for(s <- 0 until 10) {
	        val microProbaTmp = MicroProba(probaMicro = 1D, addVertices = 0.1D, removeVertices = 0.1D, updateAttributes = 0.1D, removeInternalEdges = 0.1D, 
                				addInternalEdges = 0.1D,
		                		removeExternalEdges = 0.1D, addExternalEdges = 0.1D, probaMerge = 0.1D, probaSplit = 0.1D, probaChange = 0.1D)
	        val genTmp = new GeneratorDyn(k = 10, nbVertices = 500, dev = Array(1D, 1D), theta = 0, probaCloseTriangle = 0, nbRep = 10,
				                avgWthDegree = 6, avgBtwDegree = 3, minimumTotalEdges = 5000, seed = Option(s), nbTimestamps = nbTimestamps, useAverageDegree = false, 
				                microProba = microProbaTmp)
	        genTmp.generateDynamicGraph(computeDynamicMeasure = false)
        }
    }
    
    val headerGnuPlot = """
set terminal eps enhanced size 5,4 font 'Helvetica,25'
set grid y
set xlabel 'Ratio'
set ylabel 'Runtime (sec)'
set yrange [0:*]
set xrange [*:*]
set key center top horizontal
set key font "Helvetica,14"
set key spacing 0.6 samplen 0

set style line 1 lt 1 pt 11 lw 4 linecolor rgb "brown"
set style line 2 lt 2 pt 6 lw 4 linecolor rgb "royalblue"
set style line 3 lt 3 pt 3 lw 4 linecolor rgb "forest-green"
set style line 4 lt 4 pt 12 lw 4 linecolor rgb "salmon"
set style line 5 lt 5 pt 13 lw 4  linecolor rgb "purple"
set style line 6 lt 6 pt 8 lw 4 linecolor rgb "orange"
set style line 7 lt 7 pt 7 lw 4 linecolor rgb "dark-blue"
        
set output 'runtime1.eps'
plot 'removeVertices' using 1:2 t 'R_{RemoveVert}' ls 1 with linespoint, \
     'removeBtwEdges' using 1:2 t 'R_{RemoveBtwEdge}' ls 2 with linespoint, \
     'updateAttributes' using 1:2 t 'R_{UpdAttr}' ls 3 with linespoint, \
     'removeWthEdges' using 1:2 t 'R_{RemoveWthEdge}' ls 4 with linespoint

set output 'runtime3.eps'
plot 'addVertices' using 1:2 t 'R_{AddVert}' ls 1  with linespoint

set output 'runtime4.eps'
plot 'addWthEdges' using 1:2 t 'R_{AddWthEdge}' ls 1  with linespoint, \
     'addBtwEdges' using 1:2 t 'R_{AddBtwEdge}' ls 2  with linespoint

set output 'runtime2.eps'
set xlabel 'Probability'
set yrange [0:*]
plot 'probaChange' using 1:2 t '{/Symbol q}_8 (migrate)' ls 1 with linespoint, \
     'probaMerge' using 1:2 t '{/Symbol q}_6 (merge)' ls 2 with linespoint, \
     'probaSplit' using 1:2 t '{/Symbol q}_7 (split)' ls 3 with linespoint

        
        """

    case class MeasureXp(fileName : String, label : String, gnuplotCmd : String, f : Seq[GraphMeasures] => (Double, Double), yrange : Option[String] = None, endingPlotCmd : String = "") {
	    def toGnuplot(idx : Int) = {
	        val yRangeStr = yrange.getOrElse("*:*")
	        val macroCmd = (for(o <- macroOperations) yield {
	            s"'${o.label}' using 1:${idx} t '${o.name}' ls ${o.ls} with linespoint"
	        }).mkString("plot ", ", \\\n", "")
	        val microCmd = (for(o <- microOperations) yield {
	            s"'${o.label}' using 1:${idx} t '${o.name}' ls ${o.ls} with linespoint"
	        }).mkString("plot ", ", \\\n", "")
	        val macroCmdTimestamp = (for(o <- macroOperations) yield {
	            s"'${o.label}_overTime' using 1:${idx} t '${o.name}' ls ${o.ls} with linespoint"
	        }).mkString("plot ", ", \\\n", "")
	        val microCmdTimestamp = (for(o <- microOperations) yield {
	            s"'${o.label}_overTime' using 1:${idx} t '${o.name}' ls ${o.ls} with linespoint"
	        }).mkString("plot ", ", \\\n", "")
	        
s"""${gnuplotCmd}

set output '${fileName}Macro.eps'
set ylabel '${label}'
set xlabel 'Parameter value for the macro updates'
set yrange [${yRangeStr}]
${macroCmd}

set xlabel 'Parameter value for the micro updates'
set output '${fileName}Micro.eps'
${microCmd}

set xlabel 'Timestamp'

set output '${fileName}TimestampMacro.eps'
${macroCmdTimestamp}

set output '${fileName}TimestampMicro.eps'
${microCmdTimestamp}

${endingPlotCmd}
 """
	    }
    }
        
}
