/**********************************************************************************************************************************************************
Copyright 2014, Christine Largeron, Universit Jean Monnet.
Licensed under the GNU General Public License V3
 
This file is part of DANC_Generator.
http://perso.univ-st-etienne.fr/largeron/DANC_Generator/.

DANC_Generator is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DANC_Generator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DANC_Generator.  If not, see <http://www.gnu.org/licenses/>.

If you publish material based on DANC_Generator, then, please note the assistance you received by using this project.
Use the following reference for referring to this project:
Generating Attributed Networks with Communities by Christine Largeron, Pierre-Nicolas Mougel, Reihaneh Rabbany and Osmar R. Zaane published in PLOS ONE.
**********************************************************************************************************************************************************/

package xps

import scala.collection.mutable.HashMap
import algo.Generator1
import measures.GraphMeasures
import scala.compat.Platform
import java.text.SimpleDateFormat
import java.io.File
import java.io.PrintWriter
import scala.sys.process._
import scala.collection.mutable.LinkedHashMap
import utils.PdfOutput
import algo.Output
import utils.Latex
import utils.LatexHeader
import utils.LatexPackage
import utils.Colors
import utils.Timer

object Parameters2 {
    var nbVertices = RangeWithDefault(500 to 2000 by 500, 1000, "nbVertices", "Number of vertices")
    // var nbVertices = RangeWithDefault(100 to 100 by 100, 100, "nbVertices", "Number of vertices")
    var nbAttributes = RangeWithDefault(3 to 9 by 1, 2, "nbAttributes", "Number of attributes")
//    var stdDev = RangeWithDefault(1 to 61 by 6, 7, "stdDev", "Standard deviation")
    var stdDevX = RangeWithDefault(1 to 15 by 2, 7, "stdDevX", "Standard deviation X")
    var stdDevY = RangeWithDefault(1 to 15 by 2, 1, "stdDevY", "Standard deviation Y")
    // var theta = RangeWithDefault(0 to 100 by 5, 5, "theta")
    var theta = RangeWithDefault(0 to 100 by 10, 0, "theta", "Theta")
    var nbRepresentents = RangeWithDefault(10 to 100 by 10, 100, "nbRepresentatives", "Number of representatives")
    var k = RangeWithDefault(2 to 10 by 2, 3, "k", "K")
    var maxInternalEdges = RangeWithDefault(2 to 30 by 2, 6, "maxIntEdges", "Max Internal Edges")
    var maxExternalEdges = RangeWithDefault(0 to 20 by 2, 3, "maxEntEdges", "Max External Edges")
    var minTotalEdges = RangeWithDefault(0 to 10000 by 1000, 0, "minTotEdges", "Min Total Edges")
    
    var nbRepPerc = RangeWithDefault(10 to 100 by 10, 0, "nbRepPerc", "Percentage of representatives")
}

object XpRep  {
    val nbSeeds = 10
    
    val measures = List(
//                Measure.Diameter,
//                Measure.AvgShortestPathLength,
                Measure.AvgDegree,
                Measure.BetweenInertiaRatio,
                Measure.ClusteringCoefficient,
                Measure.RandomClusteringCoefficient,
                Measure.HomophilyBetween,
                Measure.HomophilyWithin,
                Measure.Modularity,
                Measure.NbExternalEdges,
                Measure.NbInternalEdges,
                Measure.NbEdges,
                Measure.WithinInertiaRatio,
                Measure.ExpectedHomophileEdges,
                Measure.FractionOfHomophileEdges
                )
    val ranges = List(
//            	Parameters.nbVertices,
//            	Parameters.stdDevY,
//            	Parameters.minTotalEdges,
//            	Parameters.nbRepresentents,
            	Parameters.nbRepPerc
//            	Parameters.maxExternalEdges,
//            	Parameters.maxInternalEdges,
//            	Parameters.theta,
//            	Parameters.k
    )
    
    val allRanges = List(Parameters2.k, Parameters2.maxExternalEdges, Parameters2.maxInternalEdges, 
            Parameters2.nbAttributes, Parameters2.stdDevX, Parameters2.stdDevY, Parameters2.theta, Parameters2.minTotalEdges)
    
    // Out Path
    val dateFormat = new SimpleDateFormat("yyyy-MM-dd_HH:mm:ss:S")
    // val baseDir = s"${new File(".").getCanonicalPath()}/xp/${dateFormat.format(Platform.currentTime)}/"
    val baseDir = new File(".").getCanonicalPath() + "/xp/"
    new File(baseDir).mkdirs()
    val allResultsOut = new PrintWriter(baseDir + "complete")
    
    // val latexFile = new PrintWriter(s"${baseDir}results.tex")
    val latex = new Latex(s"${baseDir}results.tex", Option(LatexHeader("Results", List(
            LatexPackage("graphics"), LatexPackage("graphicx"), LatexPackage("caption"), LatexPackage("booktabs"), LatexPackage("subfigure")))))
    
    /**
     * Entry point
     */
    def main(args: Array[String]) : Unit = {
        latex.printSection("Default values")
        latex.printItemize(for(range <- allRanges) yield { range.label + ": " + range.default })
        // runDefaultParameters()
        latex.print("Moyenne sur " + nbSeeds + " graphes")
        
        for(nbVertX <- List(30000)) {
            for(range <- ranges) { runWithinRange(range, nbVertX) }
            Timer.printTimers
        }
        
        latex.printEnding()
        latex.compile
    }
    
    
    def formatDouble(x : Double) = if(x == Math.floor(x)) { "" + x.toInt } else { f"${x}%.2f" }
    
    def runWithinRange(range : RangeWithDefault, nbVertX : Int) = {
        val rangeName = range.name
        val header = measures.map(a => List(a.toString + "_mean", a.toString + "_dev")).flatten
        println((range.name :: header).mkString("# ", "\t", ""))
        val path = s"${baseDir}_repVarying_nVert_${nbVertX}/"
        new File(path).mkdirs()
        new File(s"${path}eps").mkdirs()
        new File(s"${path}graph").mkdirs()
        
        // Build data file
        val dataFile = new PrintWriter(path + "data.dat")
        range.range.foreach { v =>
            val aggregatedMeasures = range match {
                case x if(x == Parameters2.nbRepPerc) => {
                    val maxNbRep = Parameters2.nbVertices.default / Parameters2.k.default
                    val nbRep = (maxNbRep * v.toDouble / 100).toInt
                    runXp(v, path, nbRepresentents = nbRep, nbVertices = nbVertX)
                }
            }
            dataFile.println((v :: measures.map(a => List(aggregatedMeasures(a)._1, aggregatedMeasures(a)._2)).flatten).mkString("\t"))
        }
        dataFile.flush()
        dataFile.close()
        
        buildGnuPlotScript2(range.label, "Inertia", List(Measure.BetweenInertiaRatio, Measure.WithinInertiaRatio), path, Option((0,1)))
        buildGnuPlotScript2(range.label, "Clustering Coeff.", List(Measure.ClusteringCoefficient, Measure.RandomClusteringCoefficient), path)
        buildGnuPlotScript2(range.label, "Homophily", List(Measure.ExpectedHomophileEdges, Measure.FractionOfHomophileEdges, Measure.HomophilyBetween, Measure.HomophilyWithin), path, Option((0,1)))
        buildGnuPlotScript(range.label, path)
        buildLatex(range, path, nbVertX)
    }
    
    def runXp(
            paramValue : Int,
            path : String, 
            k : Int = Parameters2.k.default,
            nbVertices : Int = Parameters2.nbVertices.default, 
            nbAttributes : Int = Parameters2.nbAttributes.default, 
            stdDevX: Int = Parameters2.stdDevX.default, 
            stdDevY: Int = Parameters2.stdDevY.default,
            theta : Int = Parameters2.theta.default,
            nbRepresentents : Int = Parameters2.nbRepresentents.default,
            maxInternalEdges : Int = Parameters2.maxInternalEdges.default,
            maxExternalEdges : Int = Parameters2.maxExternalEdges.default,
            maxTotalEdges : Int = Parameters2.minTotalEdges.default) = {
        val devs = Array[Double](stdDevX.toDouble, stdDevY.toDouble)
        
        val genParams = List("# Seed", "K", "n", "nbAttr", "mean", "stdDev", "theta", "nbRep", "maxInt", "maxExt")
        // val measureNames = List("CC.", "AvgDeg", "AvgPath", "Diam.", "Mod.", "Time")
        
        var measuresAggregator = HashMap(measures.map(m => (m -> List[Double]())) : _*)
//        allResultsOut.println((genParams ::: measures.map(_.toString)).mkString("\t"))
        
        
        for(seed <- 0 until nbSeeds) {
            Timer.startTimer("Gen")
            val gen = new Generator1(k, nbVertices, devs, theta.toDouble / 100, nbRepresentents, maxInternalEdges, maxExternalEdges, maxTotalEdges, seed = Option(seed), probaCloseTriangle = 0)
            Timer.stopTimer("Gen")
            val m = new GraphMeasures(gen.graph, devs)
            Timer.startTimer("Measures")
            measuresAggregator = Measures.appendMeasures(measuresAggregator, m, gen)
            Timer.stopTimer("Measures")
//              val genParamsValues = List(seed, k, nbVertices, nbAttributes, mean, stdDev, theta, nbRepresentents, maxInternalEdges, maxExternalEdges)
            // val measuresValues = List(f"${m.clusteringCoefficient}%.4f", m.avgDegree, f"${m.avgShortestPath}%.4f", m.diameter, f"${m.modularity}%.4f", gen.runTimeSec)
            // allResultsOut.println((genParamsValues ::: measuresValues).mkString("\t"))
            
            // Pdf output
            /*
			if(seed < 3) {
			    val panel = new GraphViewPanel()
				panel.setGraph(gen)
				PdfOutput.createPdf(panel, s"${path}graph/attr-v-${paramValue}-seed-${seed}.pdf")
				
				val outP = new Output(gen)
			    outP.toJungGraph(s"${path}graph/layout-v-${paramValue}-seed-${seed}.pdf")
			}
			*/
        }
//        println("Pour le shortest path en fonction du nombre de sommets, ajouter courbes alpha * log(nbVertices) en esperant que cela fit bien !!")
        val aggregatedMeasures = Measures.meanStdDev(measuresAggregator)
        println((paramValue :: measures.map(a => List(aggregatedMeasures(a)._1, aggregatedMeasures(a)._2)).flatten).mkString("\t"))
        aggregatedMeasures
    }
    
    def buildGnuPlotScript(xLabel : String, path : String) = {
        for((m, i) <- measures.zipWithIndex) {
            val dataFile = s"${path}data.dat"
            val termCmd = s"set terminal pdf font 'Helvetica,20'"
            
            val gridCmd = "set grid y"
            val measureName = m.toString().replaceAllLiterally(" ", "_").replaceAllLiterally(".", "")
            val outCmd = s"set output '${path}eps/${measureName}.pdf'"
            val xLabelCmd = s"set xlabel '${xLabel}'"
            val yLabelCmd = s"set ylabel '${m.toString()}'"
            val offsetCmd = "set offsets 1, 1, 0, 0"
            val scaleCmd = "set autoscale fix"
            val plotCmd = s"plot '${dataFile}' using 1:${i * 2 + 2}:${i * 2 + 3} t '' lt 1 lw 2 lc rgb '#CC0033'  with errorbars, " + 
            		s"'' using 1:${i * 2 + 2} t '' lt 1 lw 2 lc rgb '#CC0033' with linespoint"
            val script = List(termCmd, outCmd, gridCmd, xLabelCmd, yLabelCmd, scaleCmd, offsetCmd, plotCmd).mkString("\n")
            val outTmpFile = new PrintWriter("/tmp/plot.gp")
            outTmpFile.println(script)
            outTmpFile.flush
            outTmpFile.close()
            
            "gnuplot /tmp/plot.gp".!!
            
        }
    }
    
    def buildGnuPlotScript2(xLabel : String, yLabel : String, measuresToPlot : List[Measure.Value], path : String, yAxisRange : Option[(Double, Double)] = None) = {
        val dataFile = s"${path}data.dat"
        val termCmd = s"set terminal pdf size 5,4 font 'Helvetica,20'"
        
        val colors = Colors.getMatchingColors(measuresToPlot.size).map(c => f"#${c.getRed()}%02x${c.getGreen()}%02x${c.getBlue()}%02x")
        val gridCmd = "set grid y"
        val measureName = yLabel.replaceAllLiterally(" ", "_").replaceAllLiterally(".", "")
        val outCmd = s"set output '${path}eps/${measureName}.pdf'"
        val xLabelCmd = s"set xlabel '${xLabel}'"
        val yLabelCmd = s"set ylabel '${yLabel}'"
//        val offsetCmd = "set offsets 1, 1, 0, 0"
        val offsetCmd = ""
        val legendPosition = List("set key out horiz", "set key center top", "set key box").mkString("\n")
//        val offsetCmd = ""
        val scaleCmd = if(yAxisRange.isDefined) { s"set yrange [${yAxisRange.get._1}:${yAxisRange.get._2}]" } else { "set autoscale fix" }
        val plotCmd = (for((m, ii) <- measuresToPlot.zipWithIndex) yield {
            val i = measures.indexOf(m)
            s"'${dataFile}' using 1:${i * 2 + 2}:${i * 2 + 3} t '' lt 1 lw 2 lc rgb '${colors(ii)}' with errorbars, \\\n" + 
        		s"'' using 1:${i * 2 + 2} lt 1 lw 3 lc rgb '${colors(ii)}' with lines title '${m.toString()}'"
        }).mkString("plot ", ", \\\n", "")
        
        val script = List(termCmd, outCmd, gridCmd, xLabelCmd, yLabelCmd, legendPosition, scaleCmd, offsetCmd, plotCmd).mkString("\n")
        val outTmpFile = new PrintWriter("/tmp/plot2.gp")
        outTmpFile.println(script)
        outTmpFile.flush
        outTmpFile.close()
        
        "gnuplot /tmp/plot2.gp".!!
    }
    
    def buildLatex(range : RangeWithDefault, path : String, nbVert : Int) = {
        Latex.use(s"${path}include.tex", true) { outLatex =>
	        outLatex.print(s"\\newpage")
	        outLatex.printSection("NbVertices: " + nbVert)
	        outLatex.print(s"\\includegraphics[width=.5\\linewidth]{${path}eps/Homophily.pdf}")
	        outLatex.print(s"\\includegraphics[width=.5\\linewidth]{${path}eps/Inertia.pdf}")
	        outLatex.newLine
	        outLatex.print(s"\\includegraphics[width=.5\\linewidth]{${path}eps/Clustering_Coeff.pdf}")
	        outLatex.print(s"\\includegraphics[width=.5\\linewidth]{${path}eps/Avg_degree.pdf}")
	        outLatex.newLine
	        outLatex.print(s"\\includegraphics[width=.5\\linewidth]{${path}eps/Modularity.pdf}")
	        outLatex.print(s"\\includegraphics[width=.5\\linewidth]{${path}eps/Btw_edges.pdf}")
	        outLatex.newLine
	        outLatex.print(s"\\includegraphics[width=.5\\linewidth]{${path}eps/Wth_edges.pdf}")
	        outLatex.print(s"\\includegraphics[width=.5\\linewidth]{${path}eps/Nb_edges.pdf}")
        }
        latex.print(s"\\input{${path}include.tex}")
    }
}

